# -*- perl -*- 
#
# DO NOT MOVE THE FIRST LINE
# It identifies the rest of the file as PERL for EMACS autoformatting
# put perl options at the end of that line, e.g., -p
#
# ----------------------------------------------------------------------------
#				    X-BONE
#
#		 USC Information Sciences Institute (USC/ISI)
#		    Marina del Rey, California 90292, USA
#			   Copyright (c) 1998-2000
# ----------------------------------------------------------------------------
#
# Copyright (c) 1998-2000 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
#
# -------------------------------------------------------------------------
#
# Effort sponsored by the Defense Advanced Research Projects Agency (DARPA)
# and Air Force Research Laboratory, Air Force Materiel Command, USAF, under
# agreement number F30602-98-1-0200. The U.S. Government is authorized to 
# reproduce and distribute reprints for Governmental purposes not withstanding
# any copyright annotation therein.
# 
# The views and conclusions contained herein are those of the authors and
# should not be interpreted as necessarily representing the official policies
# or endorsements, either expressed or implied, of the Defense Advanced
# Research Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
# 
# -------------------------------------------------------------------------
# 
#
# ----------------------------------------------------------------------------
#  $RCSfile: XB_IPsec.pm,v $
# 
# $Revision: 1.22 $
#   $Author: yushunwa $
#     $Date: 2000/10/27 22:42:51 $
#    $State: Exp $
# ----------------------------------------------------------------------------
#
#
# Primary Author: Lars Eggert

package XB_IPsec;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(add delete init reset);

use strict;
use sigtrap;

use XB_Defs;
use XB_Utils;
use XB_Log;


###############################################################################
# KAME IPSEC
###############################################################################

my %kame_aalgs = ( "md5"  => "hmac-md5",
		   "sha1" => "hmac-sha1",
		   "null" => "null" );

my %kame_ealgs = ( "des"      => "des-cbc",
		   "3des"     => "3des-cbc",
		   "simple"   => "simple",
		   "blowfish" => "blowfish-cbc",
		   "cast128"  => "cast128-cbc",
		   "rc5"      => "rc5-cbc" );

# which syntax the setkey tool likes, use 1 for new, 0 for old style
my $new_kame_style;


# Description: 
#	Add IPsec SA between $src and $dst. See XB_IPsec::add.
# Arguments:
#	see XB_IPsec::add
# Returns:
#	1 on success
# Exceptions:
#	"XB_IPsec::kame_add" on error 
sub kame_add ($$$$$$$$) {
  my ($src, $dst, $dir, $spi, $aalg, $akey, $ealg, $ekey) =@_;
  # print trace line
  my $trace_str = sprintf "XB_IPsec::kame_add $src, $dst, $dir, $spi, %s, %s",
    (defined $aalg ? "$aalg, <akey>" : "undef, undef"),
    (defined $ealg ? "$ealg, <ekey>" : "undef, undef");
  XB_Log::log "info", "-> $trace_str";
  eval {
    # check if new_kame_style is defined
    if(not defined $new_kame_style) {
      XB_Log::log "err", 
        "cannot determine setkey(8) syntax, was XB_Tunnel::init() called?"
	  and die "style";
    }
    # map algorithm names to KAME
    if(defined $aalg) {
      if(not $kame_aalgs{$aalg}) {
	XB_Log::log "err", "unsupported authentication algorithm $aalg"
	  and die "aalg";
      }
      $aalg = $kame_aalgs{$aalg};
    }
    if(defined $ealg) {
      if(not $kame_ealgs{$ealg}) {
	XB_Log::log "err", "unsupported encryption algorithm $ealg"
	  and die "ealg";
      }
      $ealg = $kame_ealgs{$ealg};
    }
    # add keys and selectors to database
    my $pipe = "| setkey -c 1> /dev/null";
    local $SIG{PIPE} = 
      sub { XB_Log::log "err", "pipe $pipe broke: $!"; return };
    open PIPE, $pipe 
      or XB_Log::log "err", "cannot open pipe $pipe: $!" and die "open";
    # add SAD/SPD entries
    my $cmd = "add $src $dst ";
    # specify transport mode instead of any
    my $ext = "-m transport";
    if($new_kame_style) {
      # this works on > 9/99 KAME and FreeBSD-4.0
      if($aalg and $ealg) {
	# want both AH and ESP
	$cmd .= "ah $spi $ext -A $aalg $akey;\n" . $cmd . 
	  "esp $spi $ext -E $ealg $ekey;\nspdadd $src $dst any -P $dir ipsec ".
	    "ah/transport/$src-$dst/require " .
	      "esp/transport/$src-$dst/require;\n";
      } elsif($aalg) { 
	# only want AH
	$cmd .= "ah $spi $ext -A $aalg $akey;\nspdadd $src $dst any " .
	  "-P $dir ipsec ah/transport/$src-$dst/require ;\n"; 
      } elsif($ealg) { 
	# only want ESP
	$cmd .= "esp $spi $ext -E $ealg $ekey;\nspdadd $src $dst any " .
	  "-P $dir ipsec esp/transport/$src-$dst/require ;\n";  
      }
    } else {
      # this works on older KAME systems < 9/99
      $cmd .= "any $spi ";
      if($aalg and $ealg) {
	# want both AH and ESP
	$cmd .= "-p ah $ext -A $aalg $akey;\n" .$cmd. 
	  "-p esp $ext -E $ealg $ekey;\n" .
	  "spdadd $src $dst any -P ipsec ah/require esp/require ;\n";
      } elsif($aalg) { 
	# only want AH
	$cmd .= "-p ah $ext -A $aalg $akey;\n" .
	  "spdadd $src $dst any -P ipsec ah/require ;\n"; 
      } elsif($ealg) { 
	# only want ESP
	$cmd .= "-p esp $ext -E $ealg $ekey;\n" . 
	  "spdadd $src $dst any -P ipsec esp/require ;\n";  
      }
    }
    print PIPE $cmd;
    XB_Log::log "debug", $cmd;
    close PIPE #or not $! 
      or XB_Log::log "err", "cannot close pipe $pipe: $!" and die "close";
  };
  # print trace line
  XB_Log::log "info", "<- $trace_str";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(open|close|(e|a)alg)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", 
    "XB_IPsec::kame_add: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_IPsec::kame_add";
}


# Description: 
#	Delete IPsec SA with $spi between $src and $dst.
# Arguments:
#	$src	source for SA (dotted decimal address)
#	$dst	destination for SA (dotted decimal address)
#       $dir    direction of SA, either "in" or "out"
#	$spi	spi (in "0x" hex notation)
# Returns:
#	1 on success
# Exceptions:
#	"XB_IPsec::kame_delete" on error, nothing to clean up by caller
sub kame_delete ($$$$) {
  my ($src, $dst, $dir, $spi) = @_;
  # print trace line
  XB_Log::log "info", "-> XB_IPsec::kame_delete $src, $dst, $dir, $spi";
  eval {
    my $ext = "-m transport";
    # check if new_kame_style is defined
    if(not defined $new_kame_style) {
      XB_Log::log "err", 
        "cannot determine setkey(8) syntax, was XB_Tunnel::init() called?"
	  and die "style";
    }
    my $pipe = "| setkey -c 1> /dev/null";
    local $SIG{PIPE} = sub { XB_Log::log "err", "Pipe $pipe broke"; return };
    open PIPE, $pipe 
      or XB_Log::log "err", "cannot open pipe $pipe: $!" and die "open";
    # delete SPD entry
    my $cmd = "spddelete $src $dst ";
    if($new_kame_style) {
      # this works on > 9/99 KAME and FreeBSD-4.0
      $cmd .= "any -P $dir ipsec ;\n";
      print PIPE $cmd;
      XB_Log::log "debug", $cmd;
      # delete SAD entry
      foreach my $t ("ah", "esp") {
	$cmd = "delete $src $dst $t $spi $ext;\n";
	print PIPE $cmd;
	XB_Log::log "debug", $cmd;
      }
    } else {
      # this works on older KAME systems < 9/99
      $cmd .= "any ;\n";
      print PIPE $cmd;
      XB_Log::log "debug", $cmd;
      # delete SAD entry
      foreach my $t ("ah", "esp") {
	$cmd = "delete $src $dst any $spi -p $t ;\n";
	print PIPE $cmd;
	XB_Log::log "debug", $cmd;
      }
    }
    close PIPE #or not $! 
      or XB_Log::log "err", "cannot close pipe $pipe: $!" and die "close";
  };
  # print trace line
  XB_Log::log "info", "<- XB_IPsec::kame_delete $src, $dst, $dir, $spi";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(open|close)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", 
      "XB_IPsec::kame_delete: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_IPsec::kame_delete";
}


# Description: 
#	Reset KAME IPsec, will remove all X-Bone SAs.
# Arguments:
#       none
# Returns:
#	1 on success
# Exceptions:
#	"XB_IPsec::kame_reset" on error, nothing to clean up by caller
sub kame_reset () {
  # print trace line
  XB_Log::log "info", "-> XB_IPsec::kame_reset";
  eval {
    # XXX: At this time it simply removes ALL SAs!
    XB_Log::log "warning", "Removing ALL IPsec SAs, not only X-Bone ones.";
    foreach my $opt ("-F", "-FP") {
      my @cmd = ("setkey", $opt);
      system(@cmd) == 0 
	or XB_Log::log "err", "@cmd failed: $!" and die "system";
    }
  };
  # print trace line
  XB_Log::log "info", "<- XB_IPsec::kame_reset";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^system/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", 
      "XB_IPsec::kame_reset: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_IPsec::kame_reset";
}


# Description: 
#	Init KAME IPsec.
# Arguments:
#       none
# Returns:
#	1 on success
# Exceptions:
#	"XB_IPsec::kame_init" on error, nothing to clean up by caller
sub kame_init () {
  # print trace line
  XB_Log::log "info", "-> XB_IPsec::kame_init";
  eval {
    if(not defined $new_kame_style) {
      # warn if not root
      if($<) { 
	XB_Log::log "warning", "only root can probe setkey(8) version, " .
	  "probe result may be incorrect"; 
      }
      # figure out if we're on an old-style (< 9/99) or new (> 9/99) KAME host
      my $pipe = "| setkey -c 2> /dev/null | grep '127.0.0.1' > /dev/null";
      local $SIG{PIPE} = sub { XB_Log::log "err", "Pipe $pipe broke"; return };
      open PIPE, $pipe 
	or XB_Log::log "err", "cannot open pipe $pipe: $!" and die "open";
      my $cmd = "spdadd $XB_Defs::LOOPBACK_IP_ADDRESS " .
	  "$XB_Defs::LOOPBACK_IP_ADDRESS any -P ipsec ah/use ;\n" .
	"spddump ;\nspddelete $XB_Defs::LOOPBACK_IP_ADDRESS " .
	    "$XB_Defs::LOOPBACK_IP_ADDRESS any ;\n";
      print PIPE $cmd;
      XB_Log::log "debug", $cmd;
      close PIPE or not $!
	or XB_Log::log "err", "cannot close pipe $pipe: $!" and die "close";
      $new_kame_style = ($? ? 1 : 0);
      XB_Log::log "notice", "setkey(8) on this KAME host seems to support " .
	"the " . ($new_kame_style ? "new syntax (9/1999 and later)" : 
		  "old syntax (before 9/1999)");
    }
    # if we're on a new-style KAME host, figure out if spddelete works
    if($new_kame_style) {
      my $pipe = "| setkey -c 2> /dev/null | grep 'parse failed' > /dev/null";
      local $SIG{PIPE} = sub { XB_Log::log "err", "Pipe $pipe broke"; return };
      open PIPE, $pipe 
	or XB_Log::log "err", "cannot open pipe $pipe: $!" and die "open";
      my $cmd = "spddelete $XB_Defs::LOOPBACK_IP_ADDRESS " . 
	  "$XB_Defs::LOOPBACK_IP_ADDRESS any -P in ipsec;\n";
      print PIPE $cmd;
      XB_Log::log "debug", $cmd;
      close PIPE or not $!
	or XB_Log::log "err", "cannot close pipe $pipe: $!" and die "close";
      if(not $?) {
	XB_Log::log "err", "Your KAME host does not have a recent setkey(8) " .
	  "patch installed that enables the \"spddelete\" command." 
	    and die "spddelete";
      }
    }
  };
  # print trace line
  XB_Log::log "info", "<- XB_IPsec::kame_init";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(open|close|spddelete)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", 
      "XB_IPsec::kame_init: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_IPsec::kame_init";
}


###############################################################################
# LINUX IPSEC
###############################################################################

my %linux_aalgs = ( "md5"  => "132",  # 128-bit key
		    "sha1" => "133",  # 160-bit key
		    "null" => "0"     # 0-bit key
		  );

my %linux_ealgs = ( "des"      => "2",            # 64-bit key
		    "3des"     => "3",            # 192-bit key
		    "rc5"      => "4",            # 128-bit key
		    "blowfish" => "5",            # 128-bit key
		    "idea"     => "6",            # 128-bit key
		    "none"     => "7",            # 0-bit key
                     );

# 0: backward; 1: forward
sub direction($) {
  my $dst = shift;
  my $pipe = "ifconfig |";
  open PIPE, $pipe or XB_Log::log "err", "$!: Could not open pipe $pipe"
    and return undef;
  # see if dst is one of the local interfaces (backward)
  while(<PIPE>) {
    if(/\s+inet\s(addr:)?((\d{1,3}\.){3}\d{1,3})/) {
      if($dst eq $2) { return 0;} # backward
    }
  }    
  close PIPE
    or XB_Log::log "err", "$!: Cannot close pipe $pipe";
  return 1; # forward
}


sub linux_add ($$$$$$$) {
  my ($src, $dst, $spi, $aalg, $akey, $ealg, $ekey) =@_;
  my ($peer, $dir, $akey_len, $ekey_len, @cmd);
  my $trace_str = sprintf "XB_IPsec::linux_add $src, $dst, $spi, %s, %s",
    (defined $aalg ? "$aalg, <akey>" : "undef, undef"),
    (defined $ealg ? "$ealg, <ekey>" : "undef, undef");
  XB_Log::log "info", "-> $trace_str";
  eval {
    # map algorithm names to Linux
    if($aalg) {
      if(not $linux_aalgs{$aalg}) {
        XB_Log::log "err", "unsupported authentication algorithm $aalg"
          and die "aalg";
      } 
      else { 
        $akey_len = $XB_Defs::KEY_LENGTHS{$aalg}/8;
        $aalg = $linux_aalgs{$aalg}; 
      }
    }
    if($ealg) {  
      if(not $linux_ealgs{$ealg}) {
        XB_Log::log "err", "unsupported encryption algorithm $ealg"
          and die "ealg";
      } 
      else { 
        $ekey_len = $XB_Defs::KEY_LENGTHS{$ealg}/8;
        $ealg = $linux_ealgs{$ealg}; 
      }
    }
    # check if it's a forward or backward policy
    if(direction($dst) == 1) { $peer = $dst; $dir = "-o"; }
    elsif(direction($dst) == 0) { $peer = $src; $dir = "-i"; }
      else { XB_Log::log "err", "cannot find the direction" 
             and die "direction";
      }
    if($aalg){  # AH
      if($ealg){ # ESP+AH (linked SA)
        @cmd = ("sadb", $dir, "add", "d", $peer, "32", $spi, $peer, $spi,
                "0", "0", "1", "51", "50", $spi, $aalg, "0", $akey_len,
                substr($akey,2), "3");
      }
      else{  # AH only
        @cmd = ("sadb", $dir, "add", "d", $peer, "32", $spi, $peer, $spi,
                "0", "0", "1", "51", "0", "0", $aalg, "0", $akey_len,
                substr($akey,2), "3");
      }
      system(@cmd) == 0 or XB_Log::log "err", "@cmd failed with $?"
        and die "sadb";            
    }  
    if($ealg){   # ESP 
      if($ealg ne "7"){  # non-null encryption
        @cmd = ("sadb", $dir, "add", "d", $peer, "32", $spi, $peer, $spi,
                "0", "0", "1", "50", "0", "0", $ealg, "8", $ekey_len,
                substr($ekey,2), "0");
      }
      else {    # ESP with null encrption
        @cmd = ("sadb", $dir, "add", "d", $peer, "32", $spi, $peer, $spi,
                "0", "0", "1", "50", "0", "0", "7", "0", "0", "0");
      }
      system(@cmd) == 0 or XB_Log::log "err", "@cmd failed with $?"
        and die "sadb";
    }
  };
  # print trace line
  XB_Log::log "info", "<- $trace_str";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(direction|(e|a)alg|sadb)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", 
      "XB_IPsec::linux_add: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_IPsec::linux_add";
}


sub linux_delete ($$$) {
  my ($src, $dst, $spi) = @_;
  my ($peer, $dir);
  # print trace line
  XB_Log::log "info", "-> XB_IPsec::linux_delete $src, $dst, $spi";
  eval {
    if(direction($dst) == 1) { $peer = $dst; $dir = "-o"; }
    elsif(direction($dst) == 0) { $peer = $src; $dir = "-i"; }
      else { XB_Log::log "err", "cannot find the direction"
             and die "direction";
      }
    my @cmd =("sadb", $dir, "delete", "d", $peer, $spi, "2");
    system(@cmd) == 0 or XB_Log::log "err", "@cmd failed with $?";            
    @cmd = ("sadb", $dir, "delete", "d", $peer, $spi, "50");
    system(@cmd) == 0 or XB_Log::log "err", "@cmd failed with $?";
    @cmd = ("sadb", $dir, "delete", "d", $peer, $spi, "51");
    system(@cmd) == 0 or XB_Log::log "err", "@cmd failed with $?"
      and die "sadb";
  };
  # print trace line
  XB_Log::log "info", "<- XB_IPsec::linux_delete $src, $dst, $spi";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(direction|sadb)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", 
      "XB_IPsec::linux_delete: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_IPsec::linux_delete";
}


sub linux_init () {
  # print trace line
  XB_Log::log "info", "-> XB_IPsec::linux_init";
  my $ok = 1;
  # only install ipsec_mod if not there yet
  my $pipe = "lsmod |";
  open PIPE, $pipe
    or XB_Log::log "err","$!: Could not open pipe $pipe" and $ok = 0;
  my $found = 0;
  while(<PIPE>) { if(/ipsec_mod/) { $found = 1; last; } }
  close PIPE
    or XB_Log::log "err", "$!: Cannot close pipe $pipe" and $ok = 0; 
  unless($found) {
    # no ipsec_mod loaded yet, load it
    my @cmd = ("insmod", "ipsec_mod");
    system(@cmd) == 0 or XB_Log::log "err", "@cmd failed with $?" and $ok = 0; 
  }
  XB_Log::log "info", "<- XB_IPsec::linux_init";
  if($ok) { return 1; } else { die "XB_IPsec::linux_init"; }
}


sub linux_reset () {
  # print trace line
  XB_Log::log "info", "-> XB_IPsec::linux_reset";
  eval {
    XB_Log::log "warning", "Removing ALL IPsec SAs, not only X-Bone ones.";
    my @cmd = ("sadb", "flush");
    system(@cmd) == 0 or XB_Log::log "err", "@cmd failed with $?" 
      and die "system";
  };
  # print trace line
  XB_Log::log "info", "<- XB_IPsec::linux_reset";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^system/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", 
    "XB_IPsec::linux_reset: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_IPsec::linux_reset";
}


###############################################################################
# CAIRN IPSEC
###############################################################################

my %cairn_aalgs = ( "md5"  => "md5-hmac-96",  # /-- use these ASAP, they inter-
		    "sha1" => "sha1-hmac-96", # \-- operate with KAME & Linux
		    "null" => "none" );

my %cairn_ealgs = ( "des"  => "des-cbc",
		    "3des" => "3des-cbc",
		    "none" => "none" );


# Description: 
#	Delete spd rule $n.
# Arguments:
#	$n	number of rule to delete
# Returns:
#	1 on success
#	undef on failure
# Exceptions:
#	None.
sub cairn_spd_delete_rule ($) {
  my @cmd = ("spd", "delete", shift);
  system(@cmd) == 0 or XB_Log::log "err", "@cmd: $!" and return undef;
  return 1;
}


# Description: 
#	Add spd entry for SA between $src and $dst.
# Arguments:
#	$src	source for SA (dotted decimal address)
#	$dst	destination for SA (dotted decimal address)
#	$aalg	valid auth algorithm name or undef
#	$ealg	valid enc algorithm name or undef
# Returns:
#	1 on success
# Exceptions:
#	"XB_IPsec::cairn_spd_add" on error, nothing to clean up by caller
sub cairn_spd_add ($$$$) {
  my ($from, $to, $aalg, $ealg) = @_;
  my $number;
  # print trace line
  my $trace_str = sprintf "XB_IPsec::cairn_spd_add $from, $to, %s, %s",
    (defined $aalg ? $aalg : "undef"), (defined $ealg ? $ealg : "undef");
  XB_Log::log "info", "-> $trace_str";
  eval {
    my @cmd;
    # find an unused rule numbers
    my @rules = ($XB_Defs::CAIRN_SPD_RULE_MIN..$XB_Defs::CAIRN_SPD_RULE_MAX);
    my $pipe = "spd list |";
    open PIPE, $pipe 
      or XB_Log::log "err", "cannot open pipe $pipe: $!" and die "open";
    # get all used rules from spd
    while(<PIPE>) {
      # mark each used entry in our range
      if(/(\d+)/ and $1 >= $XB_Defs::CAIRN_SPD_RULE_MIN and 
	 $1 <= $XB_Defs::CAIRN_SPD_RULE_MAX) { 
	$rules[$1-$XB_Defs::CAIRN_SPD_RULE_MIN] = undef; 
      } 
    }
    close PIPE #or not $!
      or XB_Log::log "err", "cannot close pipe $pipe: $!" and die "close";
    # remove used rules from the list
    @rules = grep { defined $_ } @rules;
    # add the accept rule(s) using the first unused entry we found
    $number = shift @rules 
      or XB_Log::log "err", "no unused spd numbers between " .
	"$XB_Defs::CAIRN_SPD_RULE_MIN and $XB_Defs::CAIRN_SPD_RULE_MAX" 
	  and die "nounused";
    @cmd = ("spd", "-q", "add", $number, 0, "all", "from", $from, "to", $to);
    if(defined $aalg) {
      $cmd[4] = "auth";
      system(@cmd) == 0 or XB_Log::log "err", "@cmd failed: $!" and die "aalg";
    }
    if(defined $ealg) {
      $cmd[4] = "encrypt";
      system(@cmd) == 0 or XB_Log::log "err", "@cmd failed: $!" and die "ealg";
    }
  };
  # print trace line
  XB_Log::log "info", "<- $trace_str";
  return 1 unless $@; # success if no exception
  # exception handling
  if($@ =~ /^(open|close|nounused|aalg)/) {
  } elsif($@ =~ /^ealg/) {
    eval { cairn_spd_delete_rule $number };
  } else {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", 
      "XB_IPsec::cairn_spd_add: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_IPsec::cairn_spd_add";
}


# Description: 
#	Delete spd entry between $src and $dst.
# Arguments:
#	$src	source for SA (dotted decimal address)
#	$dst	destination for SA (dotted decimal address)
# Returns:
#	1 on success
# Exceptions:
#	"XB_IPsec::cairn_spd_delete" on error, nothing to clean up by caller
sub cairn_spd_delete ($$) {
  my ($src, $dst) = @_;
  # print trace line
  XB_Log::log "info", "-> XB_IPsec::cairn_spd_delete $src, $dst";
  eval {
    my @rules;
    my $pipe = "spd list |";
    open PIPE, $pipe 
      or XB_Log::log "err", "cannot open pipe $pipe: $!" and die "open";
    while(<PIPE>) {
      # if line matches endpoints, remember its entry number
      if(/(\d+)\s+.*from\s+$src\s+to\s+$dst/) { push @rules, $1; }
    }
    close PIPE #or not $! 
      or XB_Log::log "err", "cannot close pipe $pipe: $!" and die "close";
    # remove rules found (except the default one; also ignore duplicates)
    # ignoring duplicates would not be necessary if spd behaved as documented
    my $old_entry = -1;
    foreach my $entry (sort @rules) {
      if($entry >= $XB_Defs::CAIRN_SPD_RULE_MIN and 
	 $entry <= $XB_Defs::CAIRN_SPD_RULE_MAX and $entry != $old_entry) {
	$old_entry = $entry;
	cairn_spd_delete_rule $entry 
	  or XB_Log::log "err", "cannot delete spd rule $entry" 
	    and die "delrule";
      }
    }
  };
  # print trace line
  XB_Log::log "info", "<- XB_IPsec::cairn_spd_delete $src, $dst";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(open|close|delrule)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_IPsec::cairn_spd_delete: caught unexpected " .
      "exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_IPsec::cairn_spd_delete";
}


# Description: 
#	Add mkey entry for SA between $src and $dst.
# Arguments:
#	$src	source for SA (dotted decimal address)
#	$dst	destination for SA (dotted decimal address)
#	$spi	spi (in "0x" hex notation)
#	$atype	algorithm type: "auth" or "encrypt"
#	$alg	valid algorithm name for $atype
#	$key	undef or valid key for the chosen $ealg (in hex notation)
# Returns:
#	1 on success
# Exceptions:
#	"XB_IPsec::cairn_mkey_add" on error, nothing to clean up by caller
sub cairn_mkey_add ($$$$$$) {
  my ($src, $dst, $spi, $atype, $alg, $key) = @_;
  # print trace line
  XB_Log::log "info", 
    "-> XB_IPsec::cairn_mkey_add $src, $dst, $spi, $atype, $alg, <key>";
  eval {
    my $pipe = "| mkey";
    my %sa_type = ( "auth" => "ah", "encrypt" => "esp" );
    local $SIG{PIPE} = sub { warn "Pipe $pipe broke"; return };
    open PIPE, $pipe 
      or XB_Log::log "err", "cannot open pipe $pipe: $!" and die "open";
    # address objects
    my $cmd = "address src {\n\tfamily inet\n\thost $src\n}\n" .
      "address dst {\n\t family inet\n\thost $dst\n}\n";
    # key object
    $cmd .= "key key {\n\thex $key\n}\n";
    # sa object
    $cmd .= "sa sa {\n\tspi $spi\n\tstate mature\n\t$atype $alg\n";
    $cmd .= "}\nsend {\n\tmsg {\n\t\ttype add\n\t\t";
    $cmd .= "satype " . $sa_type{$atype} . "\n\t";
    # eventually, we'll may have to send another authkey with esp add messages
    # to allow esp authentication as defined by the rfc
    # at this time, we don't because Suresh's code can't handle it anyway
    $cmd .= "}\n\tsa sa\n\tsrc src\n\tdst dst\n\t" . $atype . "key key\n}\n";
    print PIPE $cmd;
    close PIPE #or not $! 
      or XB_Log::log "err", "cannot close pipe $pipe: $!" and die "close";
  };
  # print trace line
  XB_Log::log "info", 
    "<- XB_IPsec::cairn_mkey_add $src, $dst, $spi, $atype, $alg, <key>";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(open|close)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", 
      "XB_IPsec::cairn_mkey_add: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_IPsec::cairn_mkey_add";
}


# Description: 
#	Delete mkey entry with $spi between $src and $dst.
# Arguments:
#	$src	source for SA (dotted decimal address)
#	$dst	destination for SA (dotted decimal address)
#	$spi	spi (in "0x" hex notation)
# Returns:
#	1 on success
# Exceptions:
#	"XB_IPsec::cairn_mkey_delete" on error, nothing to clean up by caller
sub cairn_mkey_delete ($$$) {
  my ($src, $dst, $spi) = @_;
  # print trace line
  XB_Log::log "info", "-> XB_IPsec::cairn_mkey_delete $src, $dst, $spi";
  eval {
    my $pipe = "| mkey";
    local $SIG{PIPE} = sub { warn "Pipe $pipe broke"; return };
    open PIPE, $pipe 
      or XB_Log::log "err", "cannot open pipe $pipe: $!" and die "open";
    foreach my $satype (qw(ah esp)) {
      # address objects
      my $cmd =  "address src {\n\tfamily inet\n\thost $src\n}\n" .
	"address dst {\n\tfamily inet\n\thost $dst\n}\n";
      # sa object
      $cmd .= "sa sa {\n\tspi $spi\n\tstate mature\n}\n";
      # delete association
      $cmd .= "send {\n\tmsg {\n\t\ttype delete\n\t\tsatype $satype\n\t}\n\t" .
	"sa sa\n\tsrc src\n\tdst dst\n}\n";
      print PIPE $cmd;
    }
    close PIPE #or not $! 
      or XB_Log::log "err", "cannot close pipe $pipe: $!" and die close;
  };
  # print trace line
  XB_Log::log "info", "<- XB_IPsec::cairn_mkey_delete $src, $dst, $spi";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(open|close)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", 
      "XB_IPsec::cairn_mkey_delete: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_IPsec::cairn_mkey_delete";
}


# Description: 
#	Add IPsec SA between $src and $dst. See XB_IPsec::add.
# Arguments:
#	see XB_IPsec::add
# Returns:
#	1 on success
# Exceptions:
#	"XB_IPsec::cairn_add" on error 
#	(XXX: an existing SA between $src and $dst may be deleted when an 
#	exception is thrown!)
sub cairn_add ($$$$$$$) {
  my ($src, $dst, $spi, $aalg, $akey, $ealg, $ekey) =@_;
  # print trace line
  my $trace_str = sprintf "XB_IPsec::cairn_add $src, $dst, $spi, %s, %s",
    (defined $aalg ? "$aalg, <akey>" : "undef, undef"),
    (defined $ealg ? "$ealg, <ekey>" : "undef, undef");
  XB_Log::log "info", "-> $trace_str";
  eval {
    # map algorithm names to CAIRN
    if(defined $aalg) {
      if(not $cairn_aalgs{$aalg}) {
	XB_Log::log "err", "unsupported authentication algorithm $aalg"
	  and die "noaalg";
      }
      $aalg = $cairn_aalgs{$aalg};
      # strip 0x from key
      $akey =~ s/0x(.*)/$1/;
    }
    if(defined $ealg) {
      if(not $cairn_ealgs{$ealg}) {
	XB_Log::log "err", "unsupported encryption algorithm $ealg" 
	  and die "noealg";
      }
      $ealg = $cairn_ealgs{$ealg};
      # strip 0x from key
      $ekey =~ s/0x(.*)/$1/;
    }
    # put stuff into mkey
    if(defined $aalg) {
      cairn_mkey_add $src, $dst, $spi, "auth", $aalg, $akey;
      # use akey as a marker during exception handling that this was ok
      $akey = undef; 
    }
    if(defined $ealg) {
      cairn_mkey_add $src, $dst, $spi, "encrypt", $ealg, $ekey;
      # use ekey as a marker during exception handling that this was ok
      $ekey = undef; 
    }
    # put stuff into spd
    if(defined $aalg or defined $ealg) { 
      cairn_spd_add $src, $dst, $aalg, $ealg;
    }
  };
  # print trace line
  XB_Log::log "info", "<- $trace_str";
  return 1 unless $@; # success if no exception
  # exception handling
  if($@ =~ /^no(a|e)alg/ or 
     ($@ =~ /^XB_IPsec::cairn_mkey_add/ and defined $akey)) {
    # no need to recover from anything
  } elsif ($@ =~ /^XB_IPsec::cairn_mkey_add/ and not defined $akey) {
    # remove stuff from mkey
    if(defined $aalg) { eval { cairn_mkey_delete $src, $dst, $spi }; }
  } elsif ($@ =~ /^XB_IPsec::cairn_spd_add/) {
    # remove stuff from mkey
    eval { cairn_mkey_delete $src, $dst, $spi };
  } else {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", 
      "XB_IPsec::cairn_add: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_IPsec::cairn_add";
}
	       

# Description: 
#	Delete IPsec SA with $spi between $src and $dst.
# Arguments:
#	$src	source for SA (dotted decimal address)
#	$dst	destination for SA (dotted decimal address)
#	$spi	spi (in "0x" hex notation)
# Returns:
#	1 on success
# Exceptions:
#	"XB_IPsec::cairn_delete" on error, nothing to clean up by caller
sub cairn_delete ($$$) {
  my ($src, $dst, $spi) = @_;
  # print trace line
  XB_Log::log "info", "-> XB_IPsec::cairn_delete $src, $dst, $spi";
  # we will do two things even if one of them throws an exception
  eval { cairn_spd_delete $src, $dst };
  my $save_at = $@;
  eval { cairn_mkey_delete $src, $dst, $spi };
  # print trace line
  XB_Log::log "info", "<- XB_IPsec::cairn_delete $src, $dst, $spi";
  # if there was no exception on the latter cmd, test first
  unless($@) { $@ = $save_at; }
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^XB_IPsec::cairn_/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", 
      "XB_IPsec::cairn_delete: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_IPsec::cairn_delete";
}


# Description: 
#	Init CAIRN IPsec.
# Arguments:
#	none
# Returns:
#	1 on success
# Exceptions:
#	"XB_IPsec::cairn_init" on error, nothing to clean up by caller
sub cairn_init () {
  # print trace line
  XB_Log::log "info", "-> XB_IPsec::cairn_init";
  XB_Log::log "info", "<- XB_IPsec::cairn_init";
  return 1;
}


# Description: 
#	Reset CAIRN IPsec. Will remove all SAs between xbone addresses in 
#	the database.
# Arguments:
#	none
# Returns:
#	1 on success
# Exceptions:
#	"XB_IPsec::cairn_reset" on error, nothing to clean up by caller
sub cairn_reset () {
  # print trace line
  XB_Log::log "info", "-> XB_IPsec::cairn_reset";
  my $ok = 1;
  # XXX: At this time it simply removes ALL SAs!
  XB_Log::log "warning", "Removing ALL IPsec SAs, not only X-Bone ones.";
  my @cmd = ("spd", "-q", "flush");
  system(@cmd) == 0 or XB_Log::log "warning", "@cmd failed: $?, continuing"
    and $ok = 0;
  # flush mkey
  my $pipe = "| mkey";
  local $SIG{PIPE} = 
    sub { XB_Log::log "err", "pipe $pipe broke" and $ok = 0; return };
  open PIPE, $pipe 
    or XB_Log::log "warning", "cannot open pipe $pipe: $!, continuing" 
      and $ok = 0;
  my $cmd = "send {\n\tmsg {\n\t\ttype flush\n\t}\n}\n";
  print PIPE $cmd;
  close PIPE
    or XB_Log::log "warning", "cannot close pipe $pipe: $!, continuing" 
      and $ok = 0;
  XB_Log::log "info", "<- XB_IPsec::cairn_reset";
  if($ok) { return 1; } else { die "XB_IPsec::cairn_reset"; }
}


###############################################################################
# UTILITY FUNCTIONS
###############################################################################

# Description: 
#	Check if argument is a hex number in 0x notation.
# Arguments:
#	$n	test string
# Returns:
#	1 on success
#	undef on failure
# Exceptions:
#	None.
sub check_hex ($) {
  my $n = lc shift;
  # print trace line
  #XB_Log::log "info", "-> XB_IPsec::check_hex $n";
  eval {
    unless($n =~ /^0x[\da-f]+$/) { 
      XB_Log::log "err", "$n is not a valid hex string" and die "hexform"; 
    }
  };
  # print trace line
  #XB_Log::log "info", "<- XB_IPsec::check_hex $n";
  return ($@ ? undef : 1);
}


# Description: 
#	Check that spi is hex and in the correct range as specified in the RFC.
# Arguments:
#	$spi	spi in 0x notation
# Returns:
#	1 on success
#	undef on failure
# Exceptions:
#	None.
sub check_spi ($) {
  my $spi = shift;
  # print trace line
  #XB_Log::log "info", "-> XB_IPsec::check_spi $spi";
  eval {
    check_hex $spi;
    if(hex $spi < 0x100) {
      XB_Log::log "err", "SPI $spi must be between 0x100 and 0xffffffff"
	and die "spirange";
    }
  };
  # print trace line
  #XB_Log::log "info", "<- XB_IPsec::check_spi $spi";
  return ($@ ? undef : 1);
}


# Description: 
#	Check that a $key is in the correct form and length for algorithm $alg.
# Arguments:
#	$alg	valid algorithm name (see XB_IPsec::add)
#	$key	key in 0x notation
# Returns:
#	1 on success
#	undef on failure
# Exceptions:
#	None.
sub check_key ($$) {
  my ($alg, $key) = @_;
  # print trace line
  #XB_Log::log "info", "-> XB_IPsec::check_key $alg, $key";
  eval {
    check_hex $key;
    if(((length $key) - 2) * 4 != $XB_Defs::KEY_LENGTHS{$alg}) { 
      XB_Log::log "err", "expected key of length $XB_Defs::KEY_LENGTHS{$alg}" .
	" for $alg, got $key with length " . (((length $key) - 2) * 4) 
	  and die "keylength";
    }
  };
  # print trace line
  #XB_Log::log "info", "<- XB_IPsec::check_key $alg, $key";
  return ($@ ? undef : 1);
}


###############################################################################
# GENERIC EXPORTED API
###############################################################################


# Description: 
#	Delete IPsec SA with $spi between $src and $dst. This function is 
#	idempotent.
# Arguments:
#	$src	source for SA (dotted decimal address)
#	$dst	destination for SA (dotted decimal address)
#       $dir    direction of SA, either "in" or "out"
#	$spi	spi (in "0x" hex notation)
# Returns:
#	1 on success
# Exceptions:
#	"XB_IPsec::delete" on error, nothing to clean up by caller
sub delete ($$$$) {
  my ($src, $dst, $dir, $spi) = @_;
  # print trace line
  XB_Log::log "info", "-> XB_IPsec::delete $src, $dst, $dir, $spi";
  eval { 
    # IP address verification
    foreach my $addr ($src, $dst) {
      if($addr !~ /(\d{1,3}\.){3}\d{1,3}/) {
	XB_Log::log "err", "illegal IP address $addr" and die "addrform";
      }
    }
    # check spi
    check_spi $spi;
    # diddle IPsec depending on host implementation
    if($XB_Defs::NODEOS eq "cairn") { cairn_delete $src, $dst, $spi } 
    elsif($XB_Defs::NODEOS eq "kame") { kame_delete $src, $dst, $dir, $spi; } 
    elsif($XB_Defs::NODEOS eq "nist") { linux_delete $src, $dst, $spi; } 
    else {
      XB_Log::log "err", "cannot figure out how to configure IPsec" 
	and die "unknown";
    }
  };
  # print trace line
  XB_Log::log "info", "<- XB_IPsec::delete $src, $dst, $dir, $spi";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(unknown|(addr|spi)form|
		  XB_IPsec::(cairn|kame|linux)_delete)/x) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_IPsec::delete: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_IPsec::delete";
}


# Description: 
#	Add IPsec SA between $src and $dst. If $aalg and $akey are given, an AH
#	SA will be created, if $ealg and $ekey are given, an ESP SA will be
#	created, if both are given, both AH and ESP (and not ESP with ESP-auth)
#	will be created. This function is idempotent.
# Arguments:
#	$src	source for SA (dotted decimal address)
#	$dst	destination for SA (dotted decimal address)
#       $dir    direction of SA, either "in" or "out"
#	$spi	spi (in "0x" hex notation)
#	$aalg	undef or one of: md5, sha1, null
#	$akey	undef or valid key for the chosen $aalg (in "0x" hex notation)
#	$ealg	undef or one of: des, 3des, simple, blowfish, cast128, rc5
#	$ekey	undef or valid key for the chosen $ealg (in "0x" hex notation)
# Returns:
#	1 on success
# Exceptions:
#	"XB_IPsec::add" on error (XXX: an existing SA between $src and $dst 
#	may be deleted when an exception is thrown!)
sub add ($$$$$$$$) {
  my ($src, $dst, $dir, $spi, $aalg, $akey, $ealg, $ekey) = @_;
  # print trace line
  my $trace_str = sprintf "XB_IPsec::add $src, $dst, $dir, $spi, %s, %s",
    (defined $aalg ? "$aalg, <akey>" : "undef, undef"),
    (defined $ealg ? "$ealg, <ekey>" : "undef, undef");
  XB_Log::log "info", "-> $trace_str";
  eval {
    # check dir
    if($dir !~ /^(in|out)$/) { 
      XB_Log::log "err", "illegal direction $dir" and die "dir";
    }
    # check that keys are in hex
    if($aalg) { check_key $aalg, $akey; }
    if($ealg) { check_key $ealg, $ekey; }
    # IP address verification
    foreach my $addr ($src, $dst) {
      if($addr !~ /(\d{1,3}\.){3}\d{1,3}/) {
	XB_Log::log "err", "illegal IP address $addr" and die "addrform";
      }
    }
    # check spi
    check_spi $spi;
    # remove previous state, if any
    XB_IPsec::delete $src, $dst, $dir, $spi;
    # diddle IPsec depending on host implementation
    if($XB_Defs::NODEOS eq "cairn") { 
      cairn_add $src, $dst, $spi, $aalg, $akey, $ealg, $ekey;
    } elsif($XB_Defs::NODEOS eq "kame") {
      kame_add $src, $dst, $dir, $spi, $aalg, $akey, $ealg, $ekey;
    } elsif($XB_Defs::NODEOS eq "nist") {
      linux_add $src, $dst, $spi, $aalg, $akey, $ealg, $ekey;
    } else { 
      XB_Log::log "err", "cannot figure out how to configure IPsec" 
	and die "unknown";
    }
  };
  # print trace line
  XB_Log::log "info", "<- $trace_str";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(dir|unknown|((a|e)key|addr|spi)form|
		  XB_IPsec::(delete|(cairn|kame|linux)_add))/x) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_IPsec::add: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_IPsec::add";
}


# Description: 
#	Init IPsec. Must be called once before any other IPsec function.
# Arguments:
#       none
# Returns:
#	1 on success
# Exceptions:
#	"XB_IPsec::init" on error, nothing to clean up by caller
sub init () {
  # print trace line
  XB_Log::log "info", "-> XB_IPsec::init";
  eval {
    # init random number generator
    srand XB_Utils::rand_seed;
    # do platform-specific initializations
    if(not defined $XB_Defs::NODEOS or not $XB_Defs::NODEOS) { 
      XB_Log::log "warning", "\$XB_Defs::NODEOS is not set, cannot fully ".
	"init IPsec without it, only initialized random number generator";
    } elsif($XB_Defs::NODEOS eq "cairn") { cairn_init; }
    elsif($XB_Defs::NODEOS eq "kame") { kame_init; }
    elsif($XB_Defs::NODEOS eq "nist") { linux_init; }
    else {
      XB_Log::log "warning", "cannot fully initialize IPsec under ".
	$XB_Defs::NODEOS_LABEL{$XB_Defs::NODEOS} . "($XB_Defs::NODEOS), ".
	  "only initialized random number generator";
    }
  };
  # print trace line
  XB_Log::log "info", "<- XB_IPsec::init";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(unknown|XB_IPsec::(cairn|kame|linux)_init)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_IPsec::init: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_IPsec::init";
}


# Description: 
#	Reset IPsec, will delete all xbone SAs.
# Arguments:
#       none
# Returns:
#	1 on success
# Exceptions:
#	"XB_IPsec::reset" on error, nothing to clean up by caller
sub reset () {
  # print trace line
  XB_Log::log "info", "-> XB_IPsec::reset";
  eval {
    # do platform-specific initializations
    if($XB_Defs::NODEOS eq "cairn") { cairn_reset; }
    elsif($XB_Defs::NODEOS eq "kame") { kame_reset; }
    elsif($XB_Defs::NODEOS eq "nist") { linux_reset; }
    else { 
      XB_Log::log "warning", "cannot fully reset IPsec under ".
	$XB_Defs::NODEOS_LABEL{$XB_Defs::NODEOS};
    }
  };
  # print trace line
  XB_Log::log "info", "<- XB_IPsec::reset";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(unknown|XB_IPsec::(cairn|kame|linux)_reset)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_IPsec::reset: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_IPsec::reset";
}


# Description: 
#	Returns a valid (spec-conforming) key for algorithm $alg in hex form.
# Arguments:
#	$alg	algorithm for which to generate key
# Returns:
#	key as hex string ("0x..." format)
# Exceptions:
#	"XB_IPsec::key" on error, nothing to clean up by caller
sub key ($) {
  my $alg = shift;
  my $key = "0x";
  # print trace line
  XB_Log::log "info", "-> XB_IPsec::key $alg";
  eval {
    unless($XB_Defs::KEY_LENGTHS{$alg}) { 
      XB_Log::log "err", "unknown algorithm $alg" and die "badalg";
    }
    # this of course only works if the key length is a multiple of 4 bits
    my $len = $XB_Defs::KEY_LENGTHS{$alg} / 4;
    # patch the key together
    while(--$len >= 0){ $key .= sprintf "%1x", int rand 0xf; }
  };
  # print trace line
  XB_Log::log "info", "<- XB_IPsec::key $alg";
  return $key unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^badalg/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_IPsec::key: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_IPsec::key";
}


1;
