# -*- perl -*- 
#
# DO NOT MOVE THE FIRST LINE
# It identifies the rest of the file as PERL for EMACS autoformatting
# put perl options at the end of that line, e.g., -p
#
# ----------------------------------------------------------------------------
#				    X-BONE
#
#		 USC Information Sciences Institute (USC/ISI)
#		    Marina del Rey, California 90292, USA
#			   Copyright (c) 1998-2000
# ----------------------------------------------------------------------------
#
# Copyright (c) 1998-2000 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
#
# -------------------------------------------------------------------------
#
# Effort sponsored by the Defense Advanced Research Projects Agency (DARPA)
# and Air Force Research Laboratory, Air Force Materiel Command, USAF, under
# agreement number F30602-98-1-0200. The U.S. Government is authorized to 
# reproduce and distribute reprints for Governmental purposes not withstanding
# any copyright annotation therein.
# 
# The views and conclusions contained herein are those of the authors and
# should not be interpreted as necessarily representing the official policies
# or endorsements, either expressed or implied, of the Defense Advanced
# Research Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
# 
# -------------------------------------------------------------------------
# 
#
# ----------------------------------------------------------------------------
#  $RCSfile: XB_Overlay_List.pm,v $
#
# $Revision: 1.12 $
#   $Author: touch $
#     $Date: 2000/10/25 22:52:48 $
#    $State: Exp $
# ----------------------------------------------------------------------------
#
# Primary Author: Greg Finn

package XB_Overlay_List;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(XB_overlaylst);
@EXPORT_OK = qw();

use strict;
use sigtrap;

use IO::Socket;

use XB_Defs;
use XB_SSL;




######################################################################
# sub XB_overlaylst ()
#
# This routine opens a connection to the Xbone resource daemon's port on
# the running host.  A message requesting a listing of all overlays in
# which the daemon is a current participant is sent.  From the reply a
# list of the overlay names is extracted and returned to the caller as a
# list.
#
# The returned value is a tuple: ($ok, @overlays) or ($ok, $errmsg)
#
# If $ok is TRUE the number of elements following is the number of overlays
# currently detected.  If no elements are following, the @overlays was
# empty.
#
# If $ok is FALSE, the following element is a descriptive error string.
# The particular message: "Xbone daemon not running or refusing connection"
# means that the resource daemon needs to be started or restarted.
####################################################################

sub XB_overlaylst ($)
{
    my ($toaddr) = @_;
    my $debug = 0;
    my ($msg, $line, $sh);
    my ($callok, $callval, $callmsg);
    my ($ssl, $peer, $issuer);
    my @retval = ();


    ########################
    # Initialize ssl system.
    ########################

    ####################################################
    # SECURITY_CA_CERT must point to the SSL certificate
    # file for the Xbone daemon running on this host.
    ####################################################

    eval { XB_SSL::init ($XB_Defs::SECURITY_CA_CERT, undef, undef); };
    if ($@)
	{
	    return (0,
            "$callmsg\nXB_SSL::init failed.  Check CA certificate: $XB_Defs::SECURITY_CA_CERT");

	}

    #######################
    # Open TCP connection
    # to XBONE's mcast port
    #######################
    $sh = IO::Socket::INET->new (LocalAddr => $XB_Defs::LOOPBACK_IP_ADDRESS,
				 PeerAddr => $toaddr,
				 PeerPort => $XB_Defs::MCAST_PORT,
				 Proto => 'tcp',
				 );
    if (!$sh)
    { return (0, "Unable to connect to local Xbone $toaddr : $XB_Defs::MCAST_PORT"); };

    ##################################
    # Make a new ssl connection over
    # the just opened tcp connection.
    ##################################

    eval { ($ssl, $peer, $issuer) = XB_SSL::connect ($sh->fileno); };
    if ($@)
    { return (0, "Xbone daemon not running or refusing connection"); };

    #####################
    # Build query message
    #####################

    $msg = "Xbone Version=$XB_Defs::XBONE_PROTOCOL Release=$XB_Defs::XBONE_RELEASE\n";
    $msg .= "XboneControl Command=ListOverlays Transaction=Request\n";
    $msg .= "XboneSecurity UserId=none AuthType=none\n";
    $msg .= "XboneEOC\n";

    eval { XB_SSL::write ($ssl, $msg); };
    if ($@)
    {
	return (0, "$callmsg\nXB_SSL::write to local Xbone daemon failed");
    };

    ####################################
    # Look for properly terminated reply
    ####################################

    $msg = "";
    while (1)
    {
	eval { $line = XB_SSL::read ($ssl); };
	if ($@)
	{ return (0, $line); };

	if (!$line)
	{ last; }
	else
	{
	    $msg .= $line;
	    if ($line =~ /\bXboneEOC\b/i)  { last; };
	};
    };

    #################
    # Clean things up
    #################

    eval { XB_SSL::close ($ssl); };
    if ($@)
    { return (0, $callmsg); };

    eval { XB_SSL::cleanup; };
    if ($@)
    { return (0, $callmsg); };

    $sh->close;

    #################
    # Gather result
    #################
    print $msg;
    foreach $_ (split /\n/, $msg)
    {
	if ($_ =~ /XboneOverlay\s+Name=<(\S+), (\S+)>/i) { 
	    # <ovlname, ovlmgr>
	    push (@retval, [ ( $1, $2 ) ]) }
    };
    
    return (1, @retval);
}


1;                            # Insure TRUE return if module is interpreted.
