# -*- perl -*- 
#
# DO NOT MOVE THE FIRST LINE
# It identifies the rest of the file as PERL for EMACS autoformatting
# put perl options at the end of that line, e.g., -p
#
# ----------------------------------------------------------------------------
#				    X-BONE
#
#		 USC Information Sciences Institute (USC/ISI)
#		    Marina del Rey, California 90292, USA
#			   Copyright (c) 1998-2000
# ----------------------------------------------------------------------------
#
# Copyright (c) 1998-2000 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
#
# -------------------------------------------------------------------------
#
# Effort sponsored by the Defense Advanced Research Projects Agency (DARPA)
# and Air Force Research Laboratory, Air Force Materiel Command, USAF, under
# agreement number F30602-98-1-0200. The U.S. Government is authorized to 
# reproduce and distribute reprints for Governmental purposes not withstanding
# any copyright annotation therein.
# 
# The views and conclusions contained herein are those of the authors and
# should not be interpreted as necessarily representing the official policies
# or endorsements, either expressed or implied, of the Defense Advanced
# Research Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
# 
# -------------------------------------------------------------------------
# 
#
# ----------------------------------------------------------------------------
#  $RCSfile: XB_SSL.pm,v $
# 
# $Revision: 1.11 $
#   $Author: touch $
#     $Date: 2000/10/24 17:28:32 $
#    $State: Exp $
# ----------------------------------------------------------------------------
#
# Primary Author: Lars Eggert


package XB_SSL;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(init cleanup connect accept close write read reset);

use strict;
use sigtrap;

use English;

use Net::SSLeay;
use XB_Log;
use XB_Utils;


##############################################################################
### INTERNAL FUNCTIONS
##############################################################################

# global ssleay context
my $global_ctx;


# Description: 
#	Callback used during SSL authentication. Stores results of 
#	authentication (failed/succeeded) in global variable $peer_verified. 
#	XXX: This will fail badly if multithreaded!
# Arguments:
#	None.
# Returns:
#	1 on authentication success
#	0 if failed
# Exceptions:
#	None.
my $peer_verified;
sub verify_peer ($$) {
  my ($ok, $x509_store_ctx) = @_;
  XB_Log::log "info", "-> XB_SSL::verify_peer $ok, $x509_store_ctx";
  $peer_verified = $ok;
  XB_Log::log "info", "<- XB_SSL::verify_peer $ok, $x509_store_ctx";
  return $ok;
}


# Description: 
#	Internal helper doing common processing for accept and connect.
# Arguments:
#	$fd	file descriptor to work with
# Returns:
#	$ssl	handle to new ssl connection
# Exceptions:
#	"XB_SSL::open_common" on failure, nothing for caller to clean up
sub open_common ($) {
  my $fd = shift;
  XB_Log::log "info", "-> XB_SSL::open_common $fd";
  my $ssl;
  eval {
    $ssl = Net::SSLeay::new $global_ctx;
    if(Net::SSLeay::ERR_get_error or not $ssl) { 
      XB_Log::log "err", "XB_SSL::open_common: Net::SSLeay::new failed" 
	and die "new";
    }
    Net::SSLeay::set_fd $ssl, $fd;
    if(Net::SSLeay::ERR_get_error) { 
      XB_Log::log "err", "XB_SSL::open_common: Net::SSLeay::set_fd failed"
	and die "set_fd";
    }
  };
  XB_Log::log "info", "<- XB_SSL::open_common $fd";
  return $ssl unless $@; # success if no exception
  # exception handling
  if($@ =~ /^new/) {
    # known exception, nothing to handle, pass up
  } elsif($@ =~ /^set_fd/) {
    # known exception, try to free $ssl and pass up
    eval { XB_SSL::close $ssl };
  } else {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_SSL::open_common: caught unexpected " .
      "exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_SSL::open_common";
}


# Description: 
#	Internal helper doing authentication processing for accept and connect.
# Arguments:
#	$ssl	handle to ssl connection
# Returns:
#	$ssl	same as input argument
#	$peer	canonical name of peer (undef if not found or auth failed)
#	$issuer	canonical name of issuer of peer's cert (also undef same case)
# Exceptions:
#	"XB_SSL::check_verification" on failure, nothing for caller to clean up
sub check_verification ($) {
  my $ssl = shift;
  XB_Log::log "info", "-> XB_SSL::check_verification $ssl";
  my @fields = (undef, undef);
  eval {
    if($peer_verified) { 
      @fields = split /\n/, Net::SSLeay::dump_peer_certificate($ssl);
      if(Net::SSLeay::ERR_get_error) { 
	XB_Log::log "err", "XB_SSL::check_verification: Net::SSLeay::dump_" .
	  "peer_certificate failed" and die "dump_peer_certificate";
      }
      # extract the canonical names from the ASN.1 line
      foreach my $v (@fields) { 
	$v =~ s/Subject\s+Name.*\/CN=([^\/]*)\/.*/$1/; 
	$v =~ s/Issuer\s+Name.*\/OU=([^\/]*)\/.*/$1/; 
      }
    }
    $peer_verified = 0;
  };
  XB_Log::log "info", "<- XB_SSL::check_verification $ssl";
  return ($ssl, @fields) unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^dump_peer_certificate/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_SSL::check_verification: caught unexpected " .
      "exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_SSL::check_verification";
}



##############################################################################
### EXPORTED FUNCTIONS
##############################################################################

# Description: 
#	SSL module cleanup routine. Call this after all SSL connections are 
#	closed.
# Arguments:
#	None.
# Returns:
#	1 on success
# Exceptions:
#	"XB_SSL::cleanup" on failure, nothing for caller to clean up
sub cleanup () {
  # print trace line
  XB_Log::log "info", "-> XB_SSL::cleanup";
  # do cleanup
  eval {
    Net::SSLeay::CTX_free $global_ctx;
    if(Net::SSLeay::ERR_get_error) {
      XB_Log::log "err", "XB_SSL::cleanup: Net::SSLeay::CTX_free failed"
	and die "CTX_free";
    }
  };
  # print trace line
  XB_Log::log "info", "<- XB_SSL::cleanup";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^CTX_free/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_SSL::cleanup: caught unexpected exception $@";
  }
  die "XB_SSL::cleanup";
}


# Description: 
#	SSL module initialization routine. Call this once before opening SSL
#	connections. Registers CA and our client certificates with SSL. 
#	Servers (programs using XB_SSL::accept later) need to specify $cert and 
#	$key, clients (programs using XB_SSL::connect later) may leave those 
#	undef to do one-way authentication.
# Arguments:
#	$ca	file name of certification authority's certificate
#	$cert	file name of our certificate (optional)
#	$key	file name of our key (optional)
# Returns:
#	1 on success
# Exceptions:
#	"XB_SSL::init" on failure, nothing for caller to clean up
sub init ($;$$) {
  my ($ca, $cert, $key) = @_;
  # print trace line
  if(defined $cert and defined $key) {
    XB_Log::log "info", "-> XB_SSL::init $ca, $cert, $key";
  } else { XB_Log::log "info", "-> XB_SSL::init $ca"; }
  # do init
  eval {
    $peer_verified = 0;
    # the usual ssleay incantation
    Net::SSLeay::load_error_strings;
    if(Net::SSLeay::ERR_get_error) { 
      XB_Log::log "err", 
        "XB_SSL::init: Net::SSLeay::load_error_strings failed"
	  and die "no_recovery";
    }
    Net::SSLeay::ERR_load_crypto_strings;
    if(Net::SSLeay::ERR_get_error) { 
      XB_Log::log "err", 
        "XB_SSL::init: Net::SSLeay::ERR_load_crypto_strings failed"
	  and die "no_recovery";
    }
    Net::SSLeay::SSLeay_add_ssl_algorithms;
    if(Net::SSLeay::ERR_get_error) { 
      XB_Log::log "err", 
        "XB_SSL::init: Net::SSLeay::SSLeay_add_ssl_algorithms failed" 
	  and die "no_recovery";
    }
    Net::SSLeay::RAND_seed $XB_Utils::rand_seed;
    if(Net::SSLeay::ERR_get_error) { 
      XB_Log::log "err", "XB_SSL::init: Net::SSLeay::randomize failed"
	and die "no_recovery";
    }
    # create new ssleay context
    $global_ctx = Net::SSLeay::CTX_new;
    if(Net::SSLeay::ERR_get_error or not $global_ctx) { 
      XB_Log::log "err", "XB_SSL::init: Net::SSLeay::CTX_new failed"
	and die "no_recovery";
    }
    # Net::SSLeay::CTX_set_options causes warnings, suppress them
    my $w = $WARNING;
    $WARNING = 0;
    Net::SSLeay::CTX_set_options $global_ctx, &Net::SSLeay::OP_ALL;
    $WARNING = $w;
    if(Net::SSLeay::ERR_get_error) { 
      XB_Log::log "err", "XB_SSL::init: Net::SSLeay::CTX_set_options failed"
	and die "undo_ctx";
    }
    # install verify cb
    Net::SSLeay::CTX_set_verify $global_ctx, &Net::SSLeay::VERIFY_PEER, 
      \&verify_peer;
    if(Net::SSLeay::ERR_get_error) { 
      XB_Log::log "err", "XB_SSL::init: Net::SSLeay::CTX_set_verify failed"
	and die "undo_ctx";
    }
    # install our key and certificate
    if(defined $cert and defined $key) {
      unless(-s $cert) { 
	XB_Log::log "err", "XB_SSL::init: cannot find cert file $cert"
	  and die "undo_ctx";
      }
      unless(-s $key) { 
	XB_Log::log "err", "XB_SSL::init cannot find key file $key"
	  and die "undo_ctx";
      }
      # Net::SSLeay::set_server_cert_and_key causes warnings, suppress them
      $w = $WARNING;
      $WARNING = 0;
      my ($dummy, $errs) = 
	Net::SSLeay::set_server_cert_and_key($global_ctx, $cert, $key);
      $WARNING = $w;
      if($errs or Net::SSLeay::ERR_get_error) { 
	XB_Log::log "err", 
	  "XB_SSL::init: Net::SSLeay::set_server_cert_and_key failed"
	    and die "undo_ctx";
      }
    }
    # install trusted ca cert
    if(defined $ca) {
      unless(-s $ca) { 
	XB_Log::log "err", "XB_SSL::init: cannot find CA cert file $ca"
	  and die "undo_ctx";
      }
      unless(Net::SSLeay::CTX_load_verify_locations $global_ctx, $ca, "") {
	XB_Log::log "err", 
	  "XB_SSL::init: Net::SSLeay::CTX_load_verify_locations failed" 
	    and die "undo_ctx";
      }
    }
  };
  Net::SSLeay::ERR_clear_error;
  # print trace line
  if(defined $cert and defined $key) {
    XB_Log::log "info", "<- XB_SSL::init $ca, $cert, $key";
  } else { XB_Log::log "info", "<- XB_SSL::init $ca"; }
  return 1 unless $@; # success if no exception
  # exception handling
  if($@ =~ /^no_recovery/) {
    # known excpetion, nothing to handle
  } elsif($@ =~ /^undo_ctx/) {
    # need to free ctx before bailing out
    eval { cleanup };
  } else {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_SSL::init: caught unexpected exception $@";
  }
  die "XB_SSL::init";
}


# Description: 
#	Reset SSL module, is a NOP.
# Arguments:
#	none
# Returns:
#       1
# Exceptions:
#	none
sub reset () {
  # print trace line
  XB_Log::log "info", "-> XB_SSL::reset";
  XB_Log::log "info", "<- XB_SSL::reset";
  return 1;
}


# Description: 
#	Do client-side SSL handshake over SSL context $ssl.
# Arguments:
#	$ssl	handle to ssl connection
# Returns:
#	$ssl	same as input argument
#	$peer	canonical name of peer (undef if not found or auth failed)
#	$issuer	canonical name of issuer of peer's cert (also undef same case)
# Exceptions:
#	"XB_SSL::connect" on failure, nothing for caller to clean up
sub connect ($) {
  my $fd = shift;
  my ($ssl, $issuer, $peer);
  # print trace line
  XB_Log::log "info", "-> XB_SSL::connect $fd";
  eval {
    $ssl = open_common $fd;
    Net::SSLeay::connect $ssl;
    if(Net::SSLeay::ERR_get_error) { 
      XB_Log::log "err", "XB_SSL::connect: Net::SSLeay::connect failed"
	and die "connect";
    }
    ($ssl, $issuer, $peer) = check_verification $ssl; 
  };
  # print trace line
  XB_Log::log "info", "<- XB_SSL::connect $fd";
  return ($ssl, $issuer, $peer) unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(XB_SSL::(open_common|check_verification)|connect)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_SSL::connect: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_SSL::connect";
}


# Description: 
#	Do client-side SSL handshake over SSL context $ssl.
# Arguments:
#	$ssl	handle to ssl connection
# Returns:
#	$ssl	same as input argument
#	$peer	canonical name of peer (undef if not found or auth failed)
#	$issuer	canonical name of issuer of peer's cert (also undef same case)
# Exceptions:
#	"XB_SSL::accept" on failure, nothing for caller to clean up
sub accept ($) {
  my $fd = shift;
  my ($ssl, $issuer, $peer);
  # print trace line
  XB_Log::log "info", "-> XB_SSL::accept $fd";
  eval {
    $ssl = open_common $fd;
    Net::SSLeay::accept $ssl;
    if(Net::SSLeay::ERR_get_error) { 
      XB_Log::log "err", "XB_SSL::accept: Net::SSLeay::accept failed"
	and die "accept";
    }
    ($ssl, $issuer, $peer) = check_verification $ssl; 
  };
  # print trace line
  XB_Log::log "info", "<- XB_SSL::accept $fd";
  return ($ssl, $issuer, $peer) unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(XB_SSL::(open_common|check_verification)|accept)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_SSL::accept: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_SSL::accept";
}


# Description: 
#	Close ssl connection $ssl.
# Arguments:
#	$ssl	handle to ssl connection
# Returns:
#	1 on success
# Exceptions:
#	"XB_SSL::close" on failure, nothing for caller to clean up
sub close ($) {
  my $ssl = shift;
  # print trace line
  XB_Log::log "info", "-> XB_SSL::close $ssl";
  eval {
    Net::SSLeay::free $ssl;
    if(Net::SSLeay::ERR_get_error) { 
      XB_Log::log "err", "XB_SSL::close: Net::SSLeay::free failed" 
	and die "free";
    }
  };
  # print trace line
  XB_Log::log "info", "<- XB_SSL::close $ssl";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^free/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_SSL::close: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_SSL::close";
}


# Description: 
#	Write data $what over ssl connection $ssl.
# Arguments:
#	$ssl	handle to ssl connection
#	$what	data to write
# Returns:
#	number of bytes written (same semantics as write(2))
# Exceptions:
#	"XB_SSL::write" on failure, nothing for caller to clean up
sub write ($$) {
  my ($ssl, $what) = @_;
  my $result;
  # print trace line
  XB_Log::log "info", "-> XB_SSL::write $ssl, <data>";
  eval {
    $result = Net::SSLeay::ssl_write_all($ssl, $what);
    if(Net::SSLeay::ERR_get_error) {
      XB_Log::log "err", "XB_SSL::write: Net::SSLeay::ssl_write_all failed"
	and die "write";
    }
  };
  # print trace line
  XB_Log::log "info", "<- XB_SSL::write $ssl, <data>";
  return $result unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^write/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_SSL::write: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_SSL::write";
}


# Will read from peer $ssl. Returns number of bytes read. Caller may have to
# loop to obtain full message. (Same semantics as system read call.)

# Description: 
#	Read data over ssl connection $ssl.
# Arguments:
#	$ssl	handle to ssl connection
# Returns:
#	string of bytes read (same semantics as read(2))
# Exceptions:
#	"XB_SSL::read" on failure, nothing for caller to clean up
sub read ($) {
  my $ssl = shift;
  my $msg = "";
  # print trace line
  XB_Log::log "info", "-> XB_SSL::read $ssl";
  eval {
    $msg = Net::SSLeay::read($ssl);
    if(Net::SSLeay::ERR_get_error) { 
      XB_Log::log "err", "XB_SSL::read: Net::SSLeay::read failed" 
	and die "read";
    }
    # make sure we never return undef
    if(not defined $msg) { $msg = ""; }
  };
  # print trace line
  XB_Log::log "info", "<- XB_SSL::read $ssl";
  return $msg unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^read/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_SSL::read: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_SSL::read";
}


1;
