# -*- perl -*- 
#
# DO NOT MOVE THE FIRST LINE
# It identifies the rest of the file as PERL for EMACS autoformatting
# put perl options at the end of that line, e.g., -p
#
# ----------------------------------------------------------------------------
#				    X-BONE
#
#		 USC Information Sciences Institute (USC/ISI)
#		    Marina del Rey, California 90292, USA
#			   Copyright (c) 1998-2000
# ----------------------------------------------------------------------------
#
# Copyright (c) 1998-2000 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
#
# -------------------------------------------------------------------------
#
# Effort sponsored by the Defense Advanced Research Projects Agency (DARPA)
# and Air Force Research Laboratory, Air Force Materiel Command, USAF, under
# agreement number F30602-98-1-0200. The U.S. Government is authorized to 
# reproduce and distribute reprints for Governmental purposes not withstanding
# any copyright annotation therein.
# 
# The views and conclusions contained herein are those of the authors and
# should not be interpreted as necessarily representing the official policies
# or endorsements, either expressed or implied, of the Defense Advanced
# Research Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
# 
# -------------------------------------------------------------------------
# 
#
# ----------------------------------------------------------------------------
#  $RCSfile: XB_IP_Allocation.pm,v $
#
# $Revision: 1.12 $
#   $Author: touch $
#     $Date: 2000/10/24 23:07:46 $
#    $State: Exp $
# ----------------------------------------------------------------------------
#
# Primary Author: Anindo Banerjea


#############################################
#
# Common routines used by the XBONE overlay manager.
#
#############################################

package XB_IP_Allocation;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(XB_is_xbone_address XB_ip_alloc
	     XB_ip_dealloc XB_ip_dealloc_all XB_generate_internal_address);
@EXPORT_OK = qw();

use strict;
use sigtrap;

use Net::Netmask;

use XB_Defs;
use XB_RPC;
use XB_Log;

my $XB_proc_name = "XBONE ip allocator";

##############################################################################
# We no longer assume Class C addresses. The address blocks can belong to any
# class and the boundary for $subnet_bits can be arbitrary.
#############################################################################

# if this is set to zero the allocator assigns subnets to links
# otherwise it assigns contiguous addresses to maximise the number
# of hosts
my $allocate_contiguous = 0;

# Now we use this many bits of the class C to number the overlays
my $subnet_bits = $XB_Defs::XB_OVL_SUBNET_BITS;

my $eblock = new Net::Netmask($XB_Defs::XB_IP_EXTERNAL_BLOCK);
my $iblock = new Net::Netmask($XB_Defs::XB_IP_INTERNAL_BLOCK);

# Sanity checks : The number of bits in both the blocks should be same!
if ($eblock->bits() != $iblock->bits()) {
  XB_Log::log "err", "The number of network bits in External IP Block and " .
      "Internal IP Block don't match.";
    die "Fatal : External - Internal IP block mismatch";
}

# What is the combined prefix length ?
my $subnet_prefix_length = $eblock->bits() + $subnet_bits;

# How many subnets (overlays) can we get with subnet_bits bits ?
my $max_subnets = 2 ** $subnet_bits;

# How many bits are left over for hosts ?
my $host_bits = $XB_Defs::IP_ADDRESS_LENGTH - $subnet_prefix_length;

# Sanity check Make sure that we have enough host bits.
# At the minimum we need 2 bits (since all 0's and all 1's are not being
# used), though this allows us an overlay with one link only!
if ($host_bits < 2) {
  XB_Log::log "err", "Error not enough bits for hosts.";
  XB_Log::log "err", "Current configuration :";
  XB_Log::log "err", "Bits in External Block : $eblock->bits()";
  XB_Log::log "err", "Subnet bits :  $subnet_bits";
   die  "Fatal: Not enough bits for hosts";
}


# This is an internal data structure that is used to keep track
# of allocated and un-allocated subnets.
my @overlay_available = (1) x $max_subnets;


##############################################################################
# free_subnet($)
# Input: (subnet)
#	subnet: number of the block of IP addresses
#               blocks are in chunks of 2**$host_bits
#		there are $max_subnets of these blocks
# Output: NONE
# Side effect: Marks the block available for use
##############################################################################
sub free_subnet ($)
{
   XB_Log::log "info", "-> XB_IP_Allocation::free_subnet :@_\n";
   my ($subnet) = @_;
   $overlay_available[$subnet] = 1;
   XB_Log::log "info", "<- XB_IP_Allocation::free_subnet :@_\n";
}


##############################################################################
# find_free_subnet($)
# Input: NONE
# Output: the id of a free block (or -1 if none available)
# Side effect: Marks the block unavailable for use
##############################################################################
sub find_free_subnet ()
{
    XB_Log::log "info", "-> XB_IP_Allocation::find_free_subnet :\n";
    my $ix;
    my $i = -1;
    for ($ix = 0; $ix < $max_subnets; $ix++) {
	if ($overlay_available[$ix]) {
	    $overlay_available[$ix] = 0;
	    $i = $ix;
	    last ;
	}
    }

      #if ($ix == 0) { 
#	$ix = -1;
#    }
    
    XB_Log::log "info", "<- XB_IP_Allocation::find_free_subnet :$i\n";
    return $i;
}


##############################################
# XB_ip_alloc($$$$)
# Input: (ovlname, header_string, requests, dns_flag)
#	ovlname: the name of an overlay to allocate
#	header_string: a string to send to the DNS server as header
#			to satisfy the protocol requriements
#			this contains the security information etc.
#	requests: a pointer to an array of host_pairs, (pairs of names)
#		  that consitute the tunnels of the overlay. The allocator
#		  can allocate IP addresses so that each such pair is
#		  on a 2 bit subnet, or can allocate addresses contiguously
#		  depending on the variable $allocate_contiguous
#       dns_flag: a flag to enable/disable DNS for the overlay.
#
# Output: (\%allocation) 
#		Pointer to the allocation data structure defined as follows:
#		$allocation{"netaddr"} = base net address of the overlay
#		$allocation{"netmask"} = network mask of the overlay
#		$allocation{"address"} = hash from name to array of addresses
#		$allocation{"name"} =    hash from address to name
#    		$allocation{"structure"} = list of ip address pairs, one
#				           pair for each tunnel
# Side effect:
#	Enters the allocated address into Primary and secondary DNS databses
##############################################################################
sub XB_ip_alloc ($$$$)
{
    my $procname = "XB_ip_alloc";
    XB_Log::log "info", "-> $procname: @_\n";
    my ($ovlname, $header_string, $requests, $dns_flag) = @_;
    my %allocation = ();
    my $dcmdmsg = $header_string;
    my $subnet;

    $subnet = find_free_subnet();
    if ($subnet < 0) {
       XB_Log::log "info", "<- $procname: 0\n";
       die "Error: Could not find free subnet";
    }
    
    if ($subnet > $max_subnets) {
      XB_Log::log "info", "Something's wrong!! find_free_subnet returned" . 
	  " a larger than permissible value.";
	die "Invalid value returned by find_free_subnet";
    }

    # We first find the network IP Address for this overlay.
    # The network IP Address is the [(2 ** $host_bits) * $subnet]th
    # address in the eblock
    my $naddr = $eblock->nth((2 ** $host_bits) * $subnet);
    my $subnetblock = new Net::Netmask($naddr . "/$subnet_prefix_length");

    $allocation{"netaddr"} = $subnetblock->base();
    $allocation{"netmask"} = $subnetblock->mask();
    $allocation{"address"} = {};
    $allocation{"name"} = {};
    $allocation{"structure"} = [];

    if ($allocate_contiguous) {

	# We want to allocate all IP addresses in the space to
	# maximize the number of nodes. So just allocate addresses
	# to nodes sequentially
        $allocation{"linkmask"} = "0xffffffff";
        my $address_count = 1;                 # We don't want to use all zeros
        my $max_host = (2 ** $host_bits) - 2 ; # We also don't want to all ones

        foreach my $hostpair (@$requests) {
	    my $address_pair = [];
	    foreach my $hostname (@$hostpair) {
	        if ($address_count > $max_host) {
		   free_subnet($subnet);
		   XB_Log::log "info", "<- $procname: 0\n";
	           die "XB_ip_alloc: address count exceeded in ip ".
		       "allocation\n";
	        }
		
		# To determine addresses for each host on the overlay
		# we use the Net::Netmask module. Makes it easy, no
		# manual calculations are needed.
		my $new_address = $subnetblock->nth($address_count);

	        $allocation{'address'}->{$hostname} = [] unless 
		    defined($allocation{'address'}->{$hostname});
	        push @{ $allocation{'address'}->{$hostname} }, $new_address;
		$allocation{'name'}->{$new_address} = $hostname;
	        push @$address_pair, $new_address;
                $dcmdmsg .= "XboneNodeCommand dns add " . $hostname . "." . 
		    $ovlname . " " . $new_address . "\n";
	        $address_count++;
	    }
	    push @{ $allocation{"structure"} }, $address_pair;
        }
    } else {

	# In this case we want to allocate a 30 bit link mask to each
	# point to point link (i.e. tunnel).
        $allocation{"linkmask"} = "0xfffffffc";

	my $link_count = 0;

	# We want to allocate a 2 bit host ID to each link, and not use
	# numbers 0 and 3. This means that each link will need 4 IP addresses
	# so the number of links we can have is:
	my $max_link = (2 ** $host_bits) / 4 ;
	
        foreach my $hostpair (@$requests) {
	    my $x = scalar(@$requests);
	    my $address_pair = [];
	    if ($link_count >= $max_link) {
		   free_subnet($subnet);
		   XB_Log::log "info", "<- $procname: 0\n";
	           die "XB_ip_alloc: link count exceeded in ip allocation\n";
	    }
	    my $address = $link_count * 4 + 1;
	    foreach my $hostname (@$hostpair) {


		my $new_address = $subnetblock->nth($address);

	        $allocation{'address'}->{$hostname} = [] unless 
		    defined($allocation{'address'}->{$hostname});
	        push @{ $allocation{'address'}->{$hostname} }, $new_address;
		$allocation{'name'}->{$new_address} = $hostname;
	        push @$address_pair, $new_address;
                $dcmdmsg .= "XboneNodeCommand dns add " . $hostname . "." . 
		    $ovlname . " " . $new_address . "\n";
	        $address++;
	    }
	    push @{ $allocation{"structure"} }, $address_pair;
	    $link_count++;
	}
    }
    $dcmdmsg .= "XboneNodeCommand dns commit\nXboneEOC\n";


    ### Should we send DNS messages or not
    if ($dns_flag == 1)
    {


      XB_Log::log "debug1",  "sending to the dns server $XB_Defs::DNS_SERVER:\n";
      XB_Log::log "debug2", "$dcmdmsg###########\n";


	#####################################################################
	# now we send the constructed message to the primary DNS server
	# We don't catch the error because this is a problem as
	# we want to let the higher level know about it
	#####################################################################
	$_ = XB_dns_rpc( $XB_Defs::DNS_SERVER, $XB_Defs::MCAST_PORT, $dcmdmsg);
        XB_Log::log "debug1", "Response from $XB_Defs::DNS_SERVER:\n";
        XB_Log::log "debug2", "$_############\n";
	
	# Handle the response from the DNS server
	if (/Code=(\S+)\s*/) {
	    if ($1 != 0) {
	      XB_Log::log "info", "<- $procname: 0\n";
	      die "Error: XB_ip_alloc failed in XB_dns_rpc: $XB_Defs::DNS_SERVER returned code not equal to zero\n$_";
	    }
	} else {
            die "Fatal: Must have Code field in return message from DNS server\n$_";
	}
	
	# Construct message to secondary DNS server
	if ( $XB_Defs::DNS_UPDATE_MODE eq "update_both") {
	    ###################################################################
	    # Do nothing, the message to send to secondary is same as primary #
	    ###################################################################
	    ;
	} elsif ( $XB_Defs::DNS_UPDATE_MODE eq "update_1_hup_2" ) {
	    ###################################################################
	    # To the secondary we just need to send a hup. It will then get   #
	    # the zone records from the primary. We do this by sending a hup  #
	    # without any corresponging commands			      #
	    ###################################################################
	    $dcmdmsg = $header_string;
	    $dcmdmsg .= "XboneNodeCommand dns commit\nXboneEOC\n";
	} else {
	    die "Fatal: Unknown DNS_UPDATE_MODE: $XB_Defs::DNS_UPDATE_MODE\n";
	}
	
      XB_Log::log "debug1", "sending to the secondary dns server $XB_Defs::DNS_SERVER_SECONDARY:\n";
      XB_Log::log "debug2", "$dcmdmsg###########\n";
	
	######################################################################
	# Send the message to the secondary DNS server
	# We catch and handle this error because we want to continue
	# even if the secondary is not able to receive the message
	######################################################################
	eval {
	    my $return_string = XB_dns_rpc( $XB_Defs::DNS_SERVER_SECONDARY, $XB_Defs::MCAST_PORT, $dcmdmsg);
	  XB_Log::log "debug1", "Response from $XB_Defs::DNS_SERVER_SECONDARY:\n";
	  XB_Log::log "debug2", "$return_string###########\n";
	};
	if ($@) {
	    my $errorstring = $@;
	  XB_Log::log "err", "Error while talking to secondary DNS. $XB_Defs::DNS_SERVER_SECONDARY\n";
	  XB_Log::log "err", "$errorstring\n";
	}
	
    } # terminating brace for DNS
    else {
      XB_Log::log "debug1", "DNS messages skipped.\n";
    }


    XB_Log::log "info", "<- $procname: 0\n";
    return (\%allocation);
}

##############################################################################
# XB_ip_dealloc($$$$)
# Input: (ovlname, header, allocation, dns_flag)
#	ovlname: name of overlay to remove
#	header: header to send to DNS servers (contains security etc)
#	allocation: the allocation structure constructed by XB_ip_alloc
#       dns_flag : whether dns is enabled/disabled for this overlay.
# Output: NONE
# Side effects:
#	Removes the DNS entries from primary and secondary DNS server
#	Marks the block of addresses as free for reuse
##############################################################################
sub XB_ip_dealloc ($$$$)
{
  my $procname = "XB_ip_dealloc";
  XB_Log::log "info", "-> $procname: @_\n";
  my ($ovlname, $header, $allocation, $dns_flag) = @_;

  # for the time being just do deallocation based on the netaddr
  my $naddr = $$allocation{"netaddr"};

  # We find the subnet number from the network address, using the 
  # formula used earlier in the reverse manner.
  # subnet = position of the network address in the whole block / 2 ^ host_bits
  my $n = $eblock->match($naddr);
  if (!$n) {
    XB_Log::log "err", "Network address : $$allocation{netaddr} from the ".
	"data structure does not match the External Block addresses.";
      die "Fatal: Netaddr format mismatch in XB_ip_dealloc: ".
	  "$$allocation{netaddr}\n";
  }
  my $subnet = $n / (2 ** $host_bits);

  if (($subnet > $max_subnets) or ($subnet < 0)) {
    XB_Log::log "err", "Some error occured! Calculated subnet value $subnet ".
	"invalid!";
      die "Wrong subnet value $subnet calculated";
  }
  free_subnet($subnet);

  my $dcmdmsg = $header;

  # do we need to send to DNS servers?
  if ($dns_flag ==1)
  {
      ##########################################################
      # Send a message to the DNS server to remove the overlay #
      ##########################################################
      $dcmdmsg .=  "XboneNodeCommand dns delete $ovlname\n";
      $dcmdmsg .=  "XboneNodeCommand dns commit\nXboneEOC\n";
      
      XB_Log::log "debug1", "sending to the dns server $XB_Defs::DNS_SERVER:\n";
      XB_Log::log "debug2", "$dcmdmsg###########\n";
      
      eval {
	  my $return_string = XB_dns_rpc( $XB_Defs::DNS_SERVER, $XB_Defs::MCAST_PORT, $dcmdmsg);
	XB_Log::log "debug1", "Response from $XB_Defs::DNS_SERVER:\n";
	XB_Log::log "debug2", "$return_string###########\n";
      };
      if ($@) { XB_Log::log "err", "$@\n"; }
      
      if ( $XB_Defs::DNS_UPDATE_MODE eq "update_both") {
	  ###################################################################
	  # Do nothing, the message to send to secondary is same as primary #
	  # This case only works if both the servers are set up as primary  #
	  # It depends on the file format on the secondary being parseable  #
	  # by the xb_daemon code which gets violated if the secondary's    #
	  # zone files are overwritten by DNS while as a slave.             #
	  ###################################################################
	  ;
      } elsif ( $XB_Defs::DNS_UPDATE_MODE eq "update_1_hup_2" ) {
	  ###################################################################
	  # To the secondary we just need to send a hup. It will then get   #
	  # the zone records from the primary. We do this by sending a hup  #
	  # without any corresponding commands			        #
	  ###################################################################
	  $dcmdmsg = $header;
	  $dcmdmsg .= "XboneNodeCommand dns commit\nXboneEOC\n";
      } else {
	  die "Fatal: Unknown DNS_UPDATE_MODE: $XB_Defs::DNS_UPDATE_MODE\n";
      }
      
    XB_Log::log "debug1", "sending to the secondary dns server $XB_Defs::DNS_SERVER_SECONDARY:\n";
    XB_Log::log "debug2", "$dcmdmsg###########\n";
      
      eval {
	  my $return_string = XB_dns_rpc( $XB_Defs::DNS_SERVER_SECONDARY, $XB_Defs::MCAST_PORT, $dcmdmsg);
        XB_Log::log "debug1", "Response from $XB_Defs::DNS_SERVER_SECONDARY:\n";
        XB_Log::log "debug2", "$return_string###########\n";
      };
      if ($@) { XB_Log::log "err", "$@\n"; }
  } else {
    XB_Log::log "debug1", "***DNS messages skipped.***\n";
  }
 XB_Log::log "info", "<- $procname: 0\n";
  
}

##############################################################################
# XB_ip_dealloc_all($)
# Input: (header)
#	header: header to send to DNS servers (contains security etc)
# Output: NONE
# Side effects:
#	Removes all Xbone DNS entries from primary and secondary DNS server
#	Marks all blocks of addresses as free for reuse
##############################################################################
sub XB_ip_dealloc_all ($)
{
  my $procname = "XB_ip_dealloc_all";
  XB_Log::log "info", "-> $procname: @_\n";
  my ($header) = @_;

  # This is called in response to overlay delete all
  # Clean up my internal structures and tell DNS daemon to
  # release all xbone related addresses
  for (my $ix = 0; $ix < $max_subnets; $ix++) {
    free_subnet($ix);
  }
  my $dcmdmsg = $header;
  $dcmdmsg .=  "XboneNodeCommand dns delete $XB_Defs::XBONE_NET\n";
  $dcmdmsg .=  "XboneNodeCommand dns commit\nXboneEOC\n";

  XB_Log::log "debug1", "sending to the dns server $XB_Defs::DNS_SERVER:\n";
  XB_Log::log "debug2", "$dcmdmsg############\n";

  eval {
	my $return_string= XB_dns_rpc( $XB_Defs::DNS_SERVER, $XB_Defs::MCAST_PORT, $dcmdmsg);
        XB_Log::log "debug1", "Response from $XB_Defs::DNS_SERVER:\n";
        XB_Log::log "debug2", "$return_string############\n";
  };
  if ($@) { XB_Log::log "warning", "$@\n"; }

  if ( $XB_Defs::DNS_UPDATE_MODE eq "update_both") {
      ###################################################################
      # Do nothing, the message to send to secondary is same as primary #
      ###################################################################
      ;
  } elsif ( $XB_Defs::DNS_UPDATE_MODE eq "update_1_hup_2" ) {
      ###################################################################
      # To the secondary we just need to send a hup. It will then get   #
      # the zone records from the primary. We do this by sending a hup  #
      # without any corresponging commands			      #
      ###################################################################
      $dcmdmsg = $header;
      $dcmdmsg .=  "XboneNodeCommand dns commit\nXboneEOC\n";
  } else {
        die "Fatal: Unknown DNS_UPDATE_MODE: $XB_Defs::DNS_UPDATE_MODE\n";
  }

  XB_Log::log "debug1", "sending to the secondary dns server $XB_Defs::DNS_SERVER_SECONDARY:\n";
  XB_Log::log "debug2", "$dcmdmsg############\n";

  eval {
      my $return_string = XB_dns_rpc( $XB_Defs::DNS_SERVER_SECONDARY, $XB_Defs::MCAST_PORT, $dcmdmsg);
      XB_Log::log "debug1", "Response from $XB_Defs::DNS_SERVER_SECONDARY:\n";
      XB_Log::log "debug2", "$return_string############\n";
  };
  if ($@) { XB_Log::log "warning", "$@\n"; }

  XB_Log::log "info", "<- $procname: 0\n";

}

##############################################################################
# XB_generate_internal_address($)
# Input: (address)
#	address: An address from the class C allocated for xbone external use
# Output: ($internal_address)
#       internal_address: Corresponding address from internal class C	
##############################################################################
sub XB_generate_internal_address($)
{
    XB_Log::log "info", "-> XB_generate_internal_address @_";
    # Given an address from the External block,  generate the corresponding
    # address from the Internal block by finding it's position in the 
    # external block and then finding an address at the same position
    # in the internal block.
    my ($external_address) = @_;
    
    my $n = $eblock->match($external_address);
    if (!$n) {
      XB_Log::log "err", "Given address $external_address does not belong ".
	  "to the External block of addresses.";
	die "Address $external_address does not belong to external block";
    }
    my $address = $iblock->nth($n);
    if (!$address) {
      XB_Log::log "err", "Some error occured, couldn't generate the $n th ".
	  "address in the internal block.";
	die "Couldn't generate the internal address";
    }
      
    XB_Log::log "info", "<- XB_generate_internal_address $address\n";
    return ($address);
}

1;







