# -*- perl -*- 
#
# DO NOT MOVE THE FIRST LINE
# It identifies the rest of the file as PERL for EMACS autoformatting
# put perl options at the end of that line, e.g., -p
#
# ----------------------------------------------------------------------------
#				    X-BONE
#
#		 USC Information Sciences Institute (USC/ISI)
#		    Marina del Rey, California 90292, USA
#			   Copyright (c) 1998-2000
# ----------------------------------------------------------------------------
#
# Copyright (c) 1998-2000 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
#
# -------------------------------------------------------------------------
#
# Effort sponsored by the Defense Advanced Research Projects Agency (DARPA)
# and Air Force Research Laboratory, Air Force Materiel Command, USAF, under
# agreement number F30602-98-1-0200. The U.S. Government is authorized to 
# reproduce and distribute reprints for Governmental purposes not withstanding
# any copyright annotation therein.
# 
# The views and conclusions contained herein are those of the authors and
# should not be interpreted as necessarily representing the official policies
# or endorsements, either expressed or implied, of the Defense Advanced
# Research Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
# 
# -------------------------------------------------------------------------
# 
#
# ----------------------------------------------------------------------------
#  $RCSfile: XB_RPC.pm,v $
#
# $Revision: 1.17 $
#   $Author: touch $
#     $Date: 2000/10/25 22:52:49 $
#    $State: Exp $
# ----------------------------------------------------------------------------
#
# Primary Author: Anindo Banerjea

package XB_RPC;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(XB_init_network XB_multicast_message XB_request_overlay 
	     XB_tcp_open XB_resource_daemon_rpc XB_dns_rpc);
@EXPORT_OK = qw();

use strict;
use sigtrap;

use Socket;
use FileHandle;
use POSIX qw(:errno_h);

use XB_SSL;
use XB_Defs;
use XB_Utils;
use XB_Log;


##########################################################################
# These variables are useful for debugging.
# If you set security to 0 you turn of the SSL code and use straight TCP
my $security = 1;

# If you set skip_dns to 1 you fake the message to the DNS server. 
# This is useful for testing the ovlmgr when the DNS servers are not up.
my $skip_dns = 0;

# If you set dummy_rpc to 1 you fake the unicast communication to the RDs
# This is useful for testing the ovlmgr when the resource daemons are not up.
my $dummy_rpc = 0;

# If you set dummy_rpc to 1 you fake the multicast message to the RDs
# This is useful for testing the ovlmgr when the resource daemons are not up.
my $dummy_mcast = 0;
##########################################################################



#######################################################################
# XB_init_network ()
# 
# Input: none
# Output: none
#
# Side effect: initializes the SSL software through larse's wrapper
#######################################################################
sub XB_init_network ()
{
  my $procname = "XB_init_network";
  XB_Log::log "info", "-> $procname: @_\n";

  if ($security) {
    XB_SSL::init $XB_Defs::SECURITY_CA_CERT,
	      $XB_Defs::SECURITY_SERVER_CERT, 
    	      $XB_Defs::SECURITY_SERVER_KEY;
    XB_Log::log "info", "<- $procname:\n";
    return;
  }
  XB_Log::log "info", "<- $procname:\n";
  return;
}

#######################################################################
# XB_multicast_message ($$)
#
# Input: (msg_pointer, radius)
#   msg_pointer: a pointer to a message to be sent
#   radius:      a distance over which this message should be multicast
#
# Output: ($udph) the handle of the socket used for mcasting
#######################################################################
sub XB_multicast_message ($$)
{
    my $procname = "XB_multicast_message";
    XB_Log::log "info", "-> $procname: @_\n";
    my ($msgptr, $radius) = @_;
    if ($dummy_mcast) {
       XB_Log::log "info", "<- $procname:\n";
       return (1);
    }
    my $udph;
    unless ($udph = IO::Socket::INET->new(Proto => "udp", LocalAddr => $XB_Defs::XBONE_OVERLAY_MANAGER)) {
	     XB_Log::log "err", "$procname: IO::Socket::INET->new failed\n";
             die "$procname: IO::Socket::INET->new failed\n";
    }
    unless ($udph->setsockopt ($XB_Defs::SOL_IP,
	    $XB_Defs::IP_MULTICAST_TTL, pack ("C", $radius))) {
	XB_Log::log "err", "$procname: cannot set multicast socket TTL: $!\n";
	die "$procname: cannot set multicast socket TTL: $!.";
    }
    my $inaddr = inet_aton ($XB_Defs::MCAST_ADDR);
    my $paddr = sockaddr_in ($XB_Defs::MCAST_PORT, $inaddr);
    unless ($udph->send ($$msgptr, 0, $paddr)) {
	XB_Log::log "err", "$procname: Cannot multicast message: $!.\n";
	die "$procname: Cannot multicast message: $!.";
    }
    XB_Log::log "info", "<- $procname:\n";
    return ($udph);
}

##############################################
# XB_request_overlay ( message, radius )
#
# This routine opens a non-specific UDP socket and multicasts across it
# the message passed by reference in \msg.  The message contains an
# overlay: identification record followed by a sequence of resource
# request description records.
# This routine gathers response messages from the UDP socket
# passed in \udph.  It returns them as a list of scalar messages.
# The length of time that messages are gathered is determined
# by $XB_Defs::ACTUATOR_WAIT.
#
# Input: (msg_pointer, radius)
#   msg_pointer: a pointer to a message to be sent
#   radius:      a distance over which this message should be multicast
#
# Output: (@msgs) an array of messages collected from the nodes that
#		  responded
#
##############################################

# holds copies of all udp messages received in the last 2*MSL, to
# prevent against duplicate messages (XXX this should probably go
# somewhere persistent, in case the OM crashes/restarts)
my %udp_msg_cache;


sub XB_request_overlay ($$) {
  my ($msgptr, $radius) = @_;
  my @msgs;
  my $trace_msg = "XB_request_overlay @_\n";
  XB_Log::log "info", "-> $trace_msg";
  
  # XXX is this still needed? i won't touch it - larse
  if($dummy_mcast) {
    # If dummy_mcast is required we just fake the response from from
    # the resource daemons. This allows the overlay manager to be
    # tested in isolation
    $msgs[0]="Xbone Version=$XB_Defs::XBONE_PROTOCOL\nXboneControl Command=ResourceDiscovery Transaction=Reply\nXboneOverlay Name=t-ovl\nXboneNode Class=host ID=mul.isi.edu IPaddr=128.9.160.211 FreeBSD IPv6\nXboneNetwork Topology=star\nXboneEOC\n";
    $msgs[1]="Xbone Version=$XB_Defs::XBONE_PROTOCOL\nXboneControl Command=ResourceDiscovery Transaction=Reply\nXboneOverlay Name=t-ovl\nXboneNode Class=host ID=div.isi.edu IPaddr=128.9.160.213 FreeBSD IPv6\nXboneNetwork Topology=star\nXboneEOC\n";
    $msgs[2]="Xbone Version=$XB_Defs::XBONE_PROTOCOL\nXboneControl Command=ResourceDiscovery Transaction=Reply\nXboneOverlay Name=t-ovl\nXboneNode Class=host ID=sub.isi.edu IPaddr=128.9.160.215 FreeBSD IPv6\nXboneNetwork Topology=star\nXboneEOC\n";
    $msgs[3]="Xbone Version=$XB_Defs::XBONE_PROTOCOL\nXboneControl Command=ResourceDiscovery Transaction=Reply\nXboneOverlay Name=t-ovl\nXboneNode Class=router ID=add.isi.edu IPaddr=128.9.160.196 OSPF RIO\nXboneNetwork Topology=star\nXboneEOC\n";
    $msgs[4]="Xbone Version=$XB_Defs::XBONE_PROTOCOL\nXboneControl Command=ResourceDiscovery Transaction=Reply\nXboneOverlay Name=t-ovl\nXboneNode Class=router ID=tan.isi.edu IPaddr=128.9.160.198 OSPF RIO\nXboneNetwork Topology=star\nXboneEOC\n";
    $msgs[5]="Xbone Version=$XB_Defs::XBONE_PROTOCOL\nXboneControl Command=ResourceDiscovery Transaction=Reply\nXboneOverlay Name=t-ovl\nXboneNode Class=router ID=sec.isi.edu IPaddr=128.9.160.199 OSPF RIO\nXboneNetwork Topology=star\nXboneEOC\n";

  } else {
    
    # otherwise, this is the real thing. We will actually multicast the
    # message to the resource daemons
    my $udph = XB_multicast_message $msgptr, $radius;

    # purge messages older than our duplicate detection interval from the cache
    my $current_time = time;
    my %updated_udp_msg_cache;
    while(my ($msg, $time) = each %udp_msg_cache) {
      # we MUST NOT delete from the hash during iterating, see "perldoc
      # -f each"
      if($time + $XB_Defs::DUPLICATE_DETECTION_INTERVAL >= $current_time) {
	# this message is younger than the detection interval, keep it
	$updated_udp_msg_cache{$msg} = $time;
	XB_Log::log "debug", "kept message received %d seconds ago in ".
	  "duplicate detection cache", $current_time - $time;
      } else {
	# this message has expired, do not keep it in the cache
	XB_Log::log "notice", "message timed out from the duplicate ".
	  "detection cache, was received %d seconds ago", 
	  $current_time - $time;
      }
    }
    # move updated cache into place
    %udp_msg_cache = %updated_udp_msg_cache;

    # issue a limited select on input from RDs
    my ($inputflags, $errorflags, $iflg, $eflg) = ("") x 4;
    vec($inputflags, $udph->fileno, 1) = 1;
    $errorflags = $inputflags;
    while(1) {
	# select doesn't return updated time left
	select($iflg = $inputflags, undef, 
	       $eflg = $errorflags, $XB_Defs::ACTUATOR_WAIT);
	if(vec($iflg, $udph->fileno, 1)) {
	    my $msg = "";
	    unless (recv($udph, $msg,
			 $XB_Defs::ACTUATOR_MSG_LIMIT, 0))
	    {
	      XB_Log::log ("err", "Reception error on UDP socket: $!.");
		die "$!";
	    }
	    
	    push @msgs, $msg;
	} elsif(vec($eflg, $udph->fileno, 1)) {
	  XB_Log::log "err", "cannot read from UDP socket: $!" 
	      and die "XB_request_overlay";
	} else { 
	    # must be due to time limit
	    last; 
	}
    }
    
    # update current time in case recv took a while
    $current_time = time;
    
    # @msgs now holds all RD responses we received, make sure none of
    # them are duplicates of earlier messages
    my @updated_msgs;
    foreach my $msg (@msgs) {
      # check if this message we just received is a duplicate
      if(exists $udp_msg_cache{$msg}) {
	# yes, we have received this message in the recent past
	XB_Log::log "warning", "received a duplicate of a message already ".
	  "received %d seconds ago, ignoring it", 
	  $current_time - $udp_msg_cache{$msg};
      } else {
	# this is a fresh message, cache and return it
	XB_Log::log "debug", "fresh message in duplicate detection cache";
	$udp_msg_cache{$msg} = $current_time;
	push @updated_msgs, $msg
      }
    }
    # move updated msgs into place
    @msgs = @updated_msgs;    
  }
  
  XB_Log::log "info", "<- $trace_msg";
  return @msgs;
}


##############################################
# XB_rpc ($server, $port, $message)
#
# This procedure sends and Xbone message to the server and port number
# specified in the arguments.
#
# Input: (server, port, message)
#	server: IP name of machine to send message to 
#	port: port number where to send the message
#	message: message to send
#
# Output: ($response) the Xbone message returned by the node
#
##############################################
sub XB_rpc ($$$)
{
    my $procname = "XB_rpc";
    XB_Log::log "info", "-> $procname: @_\n";

    my ($server, $port, $message) = @_;

    XB_Log::log "debug1", "XB_rpc: $server,  $port \n";

    my $return_message = "";
    my $sock;
    if ($security) {
	unless ($sock = IO::Socket::INET->new(LocalAddr => $XB_Defs::XBONE_OVERLAY_MANAGER,
					      PeerAddr => $server,
					      PeerPort => $port, 
					      Proto => "tcp",
					      Timeout => $XB_Defs::CONNECTION_TIMEOUT)) {
	   XB_Log::log "err", "$procname: Could not open tcp connection to $server, $port: $!.\n";
	   die "$procname: Could not open tcp connection to $server, $port: $!.\n";
	}
	$sock->autoflush(1);

	XB_Log::log "debug3", "Authenticating TCP peer. sock->fileno: " . $sock->fileno . "\n";

        my ($ssl, $peer, $issuer) = XB_SSL::connect $sock->fileno;

# No need to compare issuer to anything (like the old version did), since
# authentication will only succeed if the peer has a certifiacte that is
# signed by one of the CAs whose certs are in the file we passed to XB_SSL::init
# earlier.

	unless (defined($peer) and ($peer eq $server)) {
	    $sock->close or XB_Log::log "warning", "$!: Close failed\n";
	    XB_Log::log "err", "$procname: Certificate id ($peer) does not match server id ($server). Spoofing attempt?\n";
	    XB_SSL::close $ssl;
	    $sock->close or XB_Log::log "warning", "$!: Close failed\n";
	    die "$procname: Certificate id ($peer) does not match server id ($server). Spoofing attempt?";
	} 

	XB_Log::log "debug3", "SSL using cipher " . Net::SSLeay::get_cipher($ssl) . "\n";
	XB_Log::log "debug3", "Authenticated peer " . $peer . " (certificate signed by " .  $issuer . ").\n";

	XB_SSL::write ($ssl, $message);

	XB_Log::log "debug3", "Done writing to TCP peer.\n";

	my $start_time = time();
	########################################################
	# Select to wait until the socket becomes ready to read
	########################################################
	my ($inputflags, $errorflags, $iflg, $eflg) = ("") x 4;
        my ($events, $timeleft, $timelimit) = (0, $XB_Defs::CONNECTION_TIMEOUT,
					       $XB_Defs::CONNECTION_TIMEOUT);
        vec ($inputflags, $sock->fileno, 1) = 1;
        $errorflags = $inputflags;
        ($events, $timeleft) =
	    select ( $iflg = $inputflags, undef, $eflg = $errorflags,
		     $timelimit );

	###########################################################
	# When something (either the socket or the timer goes off
	###########################################################
        if (vec ($iflg, $sock->fileno, 1))
        {
	    ########################################################
	    # Make the read call non-blocking so we can avoid getting
	    # blocked if the other end crashes
	    ########################################################
	    fcntl($sock, F_SETFL, O_NDELAY);
            while (1) {
		if (my $buf = XB_SSL::read($ssl)) {
                    $return_message .= $buf;
	            XB_Log::log "debug4", "PARTIAL READ: $buf\n";
	            if ($buf =~ /\bXboneEOC\b/) { last;};
	        } else {
	            my $current_time = time();
	            if ($current_time >
			    $start_time + $XB_Defs::CONNECTION_TIMEOUT) {
			    XB_Log::log "err", "$procname: Connection to $server timed out\n";
                            die "$procname: Connection to $server timed out";
		        last;
	            } else {
		        if ((not defined ($!)) or ($! == 0)) {
			    # No error, but null return. This is EOF
			    last;
			} elsif ($! == EWOULDBLOCK) {
			    undef $!;
		            sleep(1);
		            next;
		        } else {
			    # we have other kind of error we don't know
			    # how to deal with
			    XB_Log::log "err", "$procname: XB_SSL::read returned null buffer and system error=$!\n";
			    die "$procname: XB_SSL::read returned null buffer and system error=$!";
		        }
		    }
	        }
            }
	} elsif (vec ($eflg, $sock->fileno, 1)) {
	    die "Fatal: $procname: I/O error on TCP socket. $!.\n";
	} else {
	      # Must be due to time limit.
	      XB_Log::log "err", "$procname: Connection to $server timed out.\n";
              die "$procname: Connection to $server timed out. I guess the machine rebooted/crashed.";
	}

	XB_Log::log "debug3", "Done reading from TCP peer.\n";

	XB_SSL::close $ssl;

	XB_Log::log "debug3", "XB_SSL::close peer.\n";

	$sock->close or XB_Log::log "warning", "$!: Close failed\n";

	XB_Log::log "debug3", "Closed socket.\n";

    } else {
	###########################
	# This is the non-SSL case
	###########################
	unless ($sock = IO::Socket::INET->new(
 	        LocalAddr => $XB_Defs::XBONE_OVERLAY_MANAGER,
		Proto => "tcp",
		PeerAddr => $server,
		PeerPort => $port)) {
	   XB_Log::log "err", "$procname: Could not open tcp connection to $server, $port. $!\n";
	   die "$procname: Could not open tcp connection to $server, $port. $!";
	}
	$sock->autoflush(1);
        print $sock $message;
        while (my $buf = <$sock>) {
            $return_message .= $buf;
        }
        $sock->close or XB_Log::log "warning", "$!: Close failed\n";
    }
    if ($return_message eq "") {
       XB_Log::log "err", "$procname: No return messsage from $server.\n";
       die "$procname: No return messsage from $server. I guess xb_daemon crashed";
    }
    XB_Log::log "info", "<- $procname:\n";
    return ($return_message);
}


##############################################
# XB_resource_daemon_rpc ($server, $port, $message)
#
# This procedure sends and Xbone message to the server and port number
# specified in the arguments. It is a wrapper around xb_rpc which 
# does the actual communication. This procedure just adds the 
# additional functionality of being able to fake a message
# in case the debugging flag ($dummy_rpc) is turned on.
#
# Input: (server, port, message)
#	server: IP name of machine to send message to 
#	port: port number where to send the message
#	message: message to send
#
# Output: ($response) the Xbone message returned by the node
#
##############################################
sub XB_resource_daemon_rpc ($$$)
{
    my $procname = "XB_resource_daemon_rpc";
    XB_Log::log "info", "-> $procname: @_\n";
    my ($server, $port, $message) = @_;
    if ($dummy_rpc) {
	XB_Log::log "debug1", "Transmitting to $server:\n";
        my $response = "";
        # Just to test without steve's daemons lets fake some
        # return values with the correct formatting
	unless ($message =~ /Command=(\S*)\s/) {
	  die "Fatal: Could not find message type in XB_resource_daemon_rpc";
	}
        $response = "Xbone Version=$XB_Defs::XBONE_PROTOCOL\nXboneControl Command=$1 Transaction=Reply\n";
	if ($1 eq "OverlayStatus") {
	    $response .= "XboneTunnel LocalAddr=198.32.6.2 RemoteAddr=198.32.6.3 Status=up\n";
            $response .= "XboneReturn Code=0\n";
            #$response .= "XboneReturn Code=1 Msg=\"Test\"\n";
	} elsif ($1 =~ /Destroy/i) {
            $response .= "XboneReturn Code=0\n";
	    #$response .= "XboneReturn Code=1 Msg=\"Test\"\n";
	} else {
            $response .= "XboneReturn Code=0\n";
	    #$response .= "XboneReturn Code=1 Msg=\"Test\"\n";
	}
        $response .= "XboneEOC\n";
	XB_Log::log "debug3", "return message from $server:\n";
	XB_Log::log "info", "<- $procname:\n";
        return ($response);
    } else {
        my $return_message = XB_rpc($server, $port, $message);
	XB_Log::log "info", "<- $procname:\n";
	return $return_message;
    }
}

##############################################
# XB_dns_rpc ($server, $port, $message)
#
# This procedure sends an Xbone DNS message to the server and port number
# specified in the arguments. It is a wrapper around xb_rpc which 
# does the actual communication. This procedure just adds the 
# additional functionality of being able to fake a message
# in case the debugging flag ($dns_skip) is turned on.
#
# Input: (server, port, message)
#	server: IP name of machine to send message to 
#	port: port number where to send the message
#	message: message to send
#
# Output: ($response) the Xbone message returned by the node
#
#############################
sub XB_dns_rpc ($$$)
{
    my $procname = "XB_dns_rpc";
    XB_Log::log "info", "-> $procname: @_\n";
    my ($server, $port, $message) = @_;
    unless ($skip_dns) { 
        my $return_message = XB_rpc($server, $port, $message);
	XB_Log::log "info", "<- $procname:\n";
	return $return_message;
    } else {
	my $response = "Xbone Version=$XB_Defs::XBONE_PROTOCOL\nXboneReturn Code=0\n";
	# $response = "Xbone Version=$XB_Defs::XBONE_PROTOCOL\nXboneReturn Code=1 Msg=\"Testing the non-zero Code return\"\n";
	$response .= "XboneEOC\n";
	XB_Log::log "info", "<- $procname:\n";
        return ($response);
    }
}

1;                            # Insure TRUE return if module is interpreted.
