# -*- perl -*- 
#
# DO NOT MOVE THE FIRST LINE
# It identifies the rest of the file as PERL for EMACS autoformatting
# put perl options at the end of that line, e.g., -p
#
# ----------------------------------------------------------------------------
#				    X-BONE
#
#		 USC Information Sciences Institute (USC/ISI)
#		    Marina del Rey, California 90292, USA
#			   Copyright (c) 1998-2000
# ----------------------------------------------------------------------------
#
# Copyright (c) 1998-2000 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
#
# -------------------------------------------------------------------------
#
# Effort sponsored by the Defense Advanced Research Projects Agency (DARPA)
# and Air Force Research Laboratory, Air Force Materiel Command, USAF, under
# agreement number F30602-98-1-0200. The U.S. Government is authorized to 
# reproduce and distribute reprints for Governmental purposes not withstanding
# any copyright annotation therein.
# 
# The views and conclusions contained herein are those of the authors and
# should not be interpreted as necessarily representing the official policies
# or endorsements, either expressed or implied, of the Defense Advanced
# Research Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
# 
# -------------------------------------------------------------------------
# 
#
# ----------------------------------------------------------------------------
#  $RCSfile: XB_Topology.pm,v $
#
# $Revision: 1.7 $
#   $Author: touch $
#     $Date: 2000/10/24 17:28:34 $
#    $State: Exp $
# ----------------------------------------------------------------------------
#
# Primary Author: Anindo Banerjea

package XB_Topology;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw();

use strict;
use sigtrap;

use Data::Dumper;

my $XB_proc_name = "XBONE Topology creator";

##############################################
# sanity_check ($Topology, $host_count, $router_count) 
#
# This procedure performs some simple checks on the input parameters
#
# Input: ($Topology, $host_count, $router_count) 
#       Topology: a string. one of "ring" "star" "linear"
#       host_count: the number of hosts involved
#       router_count: the number of routers involved
#
# Output: (error_string)
#	error_string: If error_string = "" then there is no problem
#		      otherwise the problem is described in the error_string
#
##############################################
sub sanity_check($$$)
{
    my ($Topology, $host_count, $router_count) = @_;
    if ($Topology eq "star") {
	if ($router_count != 1) {
	   return ("Error: Router count should be exactly one for a star: $router_count\n");
	};
    } elsif ($Topology eq "ring") {
	if ($router_count <= 2) {
	    return ("Error: Cannot have a ring with less than three routers: $router_count\n");
	};
    } elsif ($Topology eq "linear") {
	if ($host_count != 2) {
	    return ("Error: Must have exactly two hosts in linear topology: $host_count\n");
	};
    }  else {
        return ("Error: Unsupported topology type: $Topology\n");
    }
    return ("");

}

##############################################
# tunnel_count ($Topology, $class, $host_count, $router_count) 
#
# Computes the number of tunnels to ask from each node depending
# on the topology being set up and the type of node, and the number
# of hosts and routers to be connected
# Assumes sanity_check has been performed
#
# Input: ($Topology, $class, $host_count, $router_count) 
#       Topology: a string. one of "ring" "star" "linear"
#       class: the class of the specific node to compute tunnels for
#       host_count: the number of hosts involved
#       router_count: the number of routers involved
#
# Output: ($tunnel_count)
#         The number of tunnels appropriate for the class        
#
##############################################
sub tunnel_count($$$$)
{
    my ($Topology, $class, $host_count, $router_count) = @_;
    my $tunnel_count = 0;
    if ($Topology eq "star") {
	if ($class eq "host") {
	   $tunnel_count = 1;
	} else {
	   $tunnel_count = $host_count;
	}
     } elsif ($Topology eq "linear") {
	if ($class eq "host") {
	   $tunnel_count = 1;
	} else {
	   $tunnel_count = 2;
	}
     } elsif ($Topology eq "ring") {
	if ($class eq "host") {
	   $tunnel_count = 1;
	} else {
           ########################################################
	   #                 _            _
	   #                |  host_count  |
	   # tunnel_count = | ------------ | + 2
	   #                | router_count |
	   #
	   ########################################################
	   my $ratio = $host_count/$router_count;
	   if ($ratio == int $ratio) {
	       $tunnel_count = 2 + int $ratio;
	   } else {
	       $tunnel_count = 3 + int $ratio;
	   }
	}
    }  else {
        die "Error: Unsupported topology type: $Topology";
    }
    return ($tunnel_count);

}

##############################################
# response_check ($Topology, $host_count, $router_count, $no_host_lines, $no_router_lines) 
#
# Simple sanity check on the number of responses and the number of
# requested nodes to see if we have succeeded in finding enough
#
# Input: ($Topology, $host_count, $router_count, $no_host_lines, $no_router_lines) 
#       Topology: a string. one of "ring" "star" "linear"
#       host_count: the number of hosts required
#       router_count: the number of routers required
#       no_host_lines: the number of hosts found
#       no_router_lines: the number of routers found
#
# Output: ($errorstring)
#       $errorstring: everything went OK the string is "" otherwise
#		      the string contains a description of the error
#
##############################################
sub response_check($$$$$)
{
    my ($Topology, $host_count, $router_count, $no_host_lines, $no_router_lines) = @_;
    if ($no_router_lines < $router_count) {
	return ("Error: Could only find $no_router_lines routers\n");
    };

    if ($no_host_lines < $host_count) {
        return ("Error: Could only obtain $no_host_lines hosts\n");
    };
    return ("");

}

##############################################
# create_ip_alloc_request ($Topology, $host_name, $router_name) 
#
# Construct a ip allocation request based on topology
#
# Input: ($Topology, $host_name, $router_name)
#       Topology: a string. one of "ring" "star" "linear"
#       host_name: an array of hosts to be given numbers
#       router_name: an array of routers to be given numbers
#
# Output: (\@ipAllocRequest)
#       pointer to the allocation request structure
#	as required by xb_ip_alloc
#
##############################################
sub create_ip_alloc_request($$$)
{
    my ($Topology, $host_name, $router_name) = @_;
    my $host_count = scalar (@$host_name);
    my $router_count = scalar (@$router_name);
    my @ipAllocRequest = ();
    my $ix;
    if ($Topology eq "star") {
        for ($ix = 0;  $ix < $host_count;  $ix++)
        {
	    $ipAllocRequest[$ix] = [ $$router_name[0], $$host_name[$ix] ];
        }
    } elsif ($Topology eq "linear") {
	if ($router_count == 0) {
	    $ipAllocRequest[0] = [ $$host_name[0], $$host_name[1] ];
	} else {
	    $ipAllocRequest[0] = [ $$host_name[0], $$router_name[0] ];
            for ($ix = 0;  $ix < ($router_count-1);  $ix++)
            {
	        $ipAllocRequest[$ix+1] = [ $$router_name[$ix], $$router_name[$ix+1] ];
            }
	    $ipAllocRequest[$ix+1] = [ $$router_name[$router_count-1], $$host_name[1] ];
	}
    } elsif ($Topology eq "ring") {
	# Connect the routers in a ring
	for ($ix = 0;  $ix < $router_count;  $ix++)
        {
	    my $next = ($ix + 1)% $router_count;
	    $ipAllocRequest[$ix] = [ $$router_name[$ix], $$router_name[$next] ];
        }
	# Assign the hosts to the routers one by one
	my $which_router = 0;
        for ($ix = 0;  $ix < $host_count;  $ix++)
        {
	    $ipAllocRequest[$ix+$router_count] = [ $$router_name[$which_router], $$host_name[$ix] ];
	    $which_router = ($which_router + 1)% $router_count;
        }
    }  else {
        die "Fatal: Unsupported topology type: $Topology. This should have been caught during sanity checking";
    }
    return \@ipAllocRequest;
}

##############################################
# create_ovlTopology($Topology, $host_name, $router_name, $allocation) 
#
# Construct an ovlTopology structure as requried by Oscar's app code
# based on the allocation of IP addresses returned by xb_ip_alloc
#
# Input: ($Topology, $host_name, $router_name, $allocation)
#       Topology: a string. one of "ring" "star" "linear"
#       host_name: an array of hosts to be given numbers
#       router_name: an array of routers to be given numbers
#	allocation: a pointer to an allocation structure created by
#		    xb_ip_alloc::XB_ip_alloc
#
# Output: $pointer_to_ovlTopology
#	structure of ovlTopology is as needed by the application code
#	written by Oscar and should be documented there.
######################################################################
sub create_ovlTopology ($$$$)
{
    my ($Topology, $host_name, $router_name, $allocation) = @_;

    my %ovlTopology = ();
    my $host_count = scalar (@$host_name);
    my $router_count = scalar (@$router_name);
    
    $ovlTopology{"Type"} = $Topology;
    $ovlTopology{"Structure"} = ();
    $ovlTopology{"Structure"}{"Routers"} = [];
    $ovlTopology{"Structure"}{"Hosts"} = [];
    $ovlTopology{"Vertices"} = ();
    $ovlTopology{"Edges"} = [];

    for (my $ix = 0;  $ix < $router_count;  $ix++)
    {
	$ovlTopology{"Structure"}{"Routers"}[$ix] = $$router_name[$ix];
	$ovlTopology{"Vertices"}{$$router_name[$ix]} = $allocation->{'address'}->{$$router_name[$ix]}[0];
    }
    for (my $ix = 0;  $ix < $host_count;  $ix++)
    {
	$ovlTopology{"Structure"}{"Hosts"}[$ix] = $$host_name[$ix];
	$ovlTopology{"Vertices"}{$$host_name[$ix]} = $allocation->{'address'}->{$$host_name[$ix]}[0];
    }

    foreach my $address_pair (@{ $allocation->{'structure'} })
    {
        push @{ $ovlTopology{"Edges"}},
				[ $allocation->{'name'}->{$address_pair->[0]},
		    		  $allocation->{'name'}->{$address_pair->[1]} ];
    }

    return \%ovlTopology;

}

##############################################
#
# INTERNAL FUNCTION. NOT EXPORTED
#
# dfs_tree($address0, $address1, $allocation)
#
# Computes a list of nodes that need to have a route at address0
# pointing at address 1. I.e. at address0 in order to get to these
# nodes, you need to go through address1.
#
# Input: ($address0, $address1, $allocation)
#	address0: an IP address from the allocation
#	address1: a neighbouring IP address from the allocation
#		  i.e. address0 is a immediate neighbour across a tunnel
#		  otherwise it will lead to a fatal error
#	$allocation: pointer to allocation structure
#
# Output: ($return_list)
#	return_list: list of addresses that need to go through address1
#	             if coming from address0
##############################################
sub dfs_tree ($$$)
{
    my ($address0, $address1, $allocation) = @_;
    my $return_list = [];
    my $node1 = $allocation->{'name'}->{$address1};
    my $addr1_list = $allocation->{'address'}->{$node1};
    foreach my $address (@{$addr1_list}) {
	if ($address ne $address1) {
	    push @{$return_list}, $address;
	    my $otheraddr = "";
            foreach my $address_pair (@{$allocation->{'structure'}}) {
		if ($address eq $address_pair->[0]) {
		    $otheraddr = $address_pair->[1];
		} elsif ($address eq $address_pair->[1]) {
		    $otheraddr = $address_pair->[0];
	        }
	    }
	    die "Fatal: Could not find $address in " . Dumper($allocation->{'structure'}) if ($otheraddr eq "");
	    push @{$return_list}, $otheraddr;
	    push @{$return_list}, @{dfs_tree($address1, $otheraddr, $allocation)};
	}
    }
    return $return_list;
}

##############################################
#
# generate_route (Topology, address0, address1, Class, allocation, structure)
#
# Spits out the appropriate string of formatted route commands corresponding
# to a given tunnel
#
# Input: ($Topology, $address0, $address1, $Class, $allocation, $structure)
#	Topology: string: one of "ring" "star" or "linear"
#	address0, address1: IP addresses of a tunnel
#	$class: class of node having address 0 (host or router)
#	$allocation: pointer to allocation structure
#	$structure: the "structure" component of the ovlTopology
#		generated by create_ovlTopology 
#
# Output: ($return_list)
#	return_list: string containing the route commands.
##############################################

sub generate_route($$$$$$)
{
    my ($Topology, $address0, $address1, $Class, $allocation, $structure) = @_;

    my $node0 = $allocation->{'name'}->{$address0};
    my $node1 = $allocation->{'name'}->{$address1};
    my $class0 = $$Class{$node0};
    my $class1 = $$Class{$node1};
    my $netaddr = $allocation->{'netaddr'};
    my $netmask = $allocation->{'netmask'};
    my $return_string = "";
    if ($Topology eq "star") {
	if ($class0 eq "host") {
            $return_string .= "XboneNodeCommand route add -net $netaddr -netmask $netmask $address1\n" ;
	}
    } elsif ($Topology eq "ring") {
	if ($class0 eq "host") {
            $return_string .= "XboneNodeCommand route add -net $netaddr -netmask $netmask $address1\n" ;
	} elsif (($class0 eq "router") and ($class1 eq "router")) {
	    ############################################################
	    # if node0 and node1 are connected and left to right order
	    ############################################################
	    my $nrouters = scalar (@{$structure->{'Routers'}});
	    for (my $ix = 0; $ix < $nrouters; $ix ++) {
		my $next = ($ix + 1) % $nrouters;
		if (($node0 eq $structure->{"Routers"}[$ix]) and
		    ($node1 eq $structure->{"Routers"}[$next])) {
            	    $return_string .= "XboneNodeCommand route add -net $netaddr -netmask $netmask $address1\n" ;
		}
	    }
	}
    } elsif (($Topology eq "linear") or
	     ($Topology eq "spanningTree") or
	     ($Topology eq "staggered")) {
	if ($class0 eq "host") {
            $return_string .= "XboneNodeCommand route add -net $netaddr -netmask $netmask $address1\n" ;
	} elsif (($class0 eq "router") and ($class1 eq "router")) {
	    #Find the set of addresses reachable through address1
	    my $address_list = dfs_tree($address0, $address1, $allocation);
	    if (scalar(@{$address_list}) > 0) {
	      foreach my $address (@{$address_list}) {
		$return_string .= "XboneNodeCommand route add -host $address $address1\n" ;
	      }
	    }
	}
    }
    return $return_string;
}
 
1;
