# -*- perl -*- 
#
# DO NOT MOVE THE FIRST LINE
# It identifies the rest of the file as PERL for EMACS autoformatting
# put perl options at the end of that line, e.g., -p
#
# ----------------------------------------------------------------------------
#				    X-BONE
#
#		 USC Information Sciences Institute (USC/ISI)
#		    Marina del Rey, California 90292, USA
#			   Copyright (c) 1998-2000
# ----------------------------------------------------------------------------
#
# Copyright (c) 1998-2000 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
#
# -------------------------------------------------------------------------
#
# Effort sponsored by the Defense Advanced Research Projects Agency (DARPA)
# and Air Force Research Laboratory, Air Force Materiel Command, USAF, under
# agreement number F30602-98-1-0200. The U.S. Government is authorized to 
# reproduce and distribute reprints for Governmental purposes not withstanding
# any copyright annotation therein.
# 
# The views and conclusions contained herein are those of the authors and
# should not be interpreted as necessarily representing the official policies
# or endorsements, either expressed or implied, of the Defense Advanced
# Research Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
# 
# -------------------------------------------------------------------------
# 
#
# ----------------------------------------------------------------------------
#  $RCSfile: XB_DNS.pm,v $
#
# $Revision: 1.10 $
#   $Author: touch $
#     $Date: 2000/10/24 17:28:36 $
#    $State: Exp $
# ----------------------------------------------------------------------------
#
# Primary Author: Greg Finn

#############################################
#
# DNS database action routines used by XBONE Perl modules.
#
#############################################

package XB_DNS;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(XB_dns_test XB_file_lock XB_file_unlock XB_write_dns_database
	     XB_check_dns_additions XB_check_dns_deletions
	     XB_increment_dns_database XB_commit_dns_database);
@EXPORT_OK = qw();

use strict;
use sigtrap;

use FileHandle;
use XB_Defs;
use XB_Log;




############################################################
#
# XB_file_lock ( fd )
#
# Uses PERL's flock() procedure for exclusively locking a
# file whose file descriptor is passed in fd.
############################################################

sub XB_file_lock ($)
{
    my ($fd) = @_;
    my $LOCK_EX = 2;

    XB_Log::log ("info", "->XB_file_lock (@_)");

    flock ($fd, $LOCK_EX);

    XB_Log::log ("info", "<-XB_file_lock");
}





############################################################
#
# XB_file_unlock ( fd )
#
# Uses PERL's flock() procedure for unlocking a file whose
# file descriptor is passed in fd.
############################################################

sub XB_file_unlock ($)
{
    my ($fd) = @_;
    my $LOCK_UN = 8;

    XB_Log::log ("info", "->XB_file_unlock (@_)");

    flock ($fd, $LOCK_UN);

    XB_Log::log ("info", "<-XB_file_unlock");
}






############################################################
#
# XB_write_dns_database ( file, \linelst )
#
# Writes the DNS database whose string filename is passed in file.
# Each line is retrieved from \linelst, a list that is passed by
# reference.
############################################################

sub XB_write_dns_database ($$)
{
    my ($filename, $linelstptr) = @_;
    my $dnsfh = new FileHandle;  my $line;

    my $subrname = "XB_write_dns_database";

    XB_Log::log ("info", "->$subrname (@_)");
    
    if (!$dnsfh->open (">$filename"))
    { die "Cannot open DNS database file: $filename.\n"; };

    while ($line = shift @$linelstptr)
    {
	if (!print $dnsfh $line)
	{ die "Cannot write DNS database file: $filename.\n"; };
    };

    if (!$dnsfh->close)
    {  XB_Log::log ("warning", "Close of $filename failed [$!].");  }

    XB_Log::log ("info", "<-$subrname");
}





############################################################
#
# XB_get_dns_record ( \dnslinelst, \index )
#
# Extracts and returns a text DNS database record from the list of text
# lines that is passed by reference in \dnslinelst.
#
# The search begins from the line indexed by \index, which is also passed
# by reference.  Upon return, \index will have been updated to point just
# beyond the record that has been returned.
#
# All comment lines and empty lines will be discarded from the record
# that is returned.
############################################################

sub XB_get_dns_record ($$)
{
    my ($linelstptr, $index ) = @_;

    my $line;  my $subrname = "XB_get_dns_record";
    my $record = "";

    XB_Log::log ("info", "->$subrname (@_)");

    while ($line = $linelstptr->[$$index])
    {
	if ($line =~ /^;/)         	# comment lines start with semicolon
	{
	    $$index++;		        # discard it 
	    next;
	}
	elsif ($line !~ /\S/)         	# empty line
	{
	    $$index++;
	    next;
	}
	elsif ($line =~ /^\s/)   	# must be non-empty line
	{			        # with leading whitespace 
	    if ($record)
	    { $record = $record . $line; }  # continuation line
	    else
	    {
		$record = $line;        # should not happen
		goto EXIT;
	    }

	}
	elsif ($record)		# non-empty line without leading whitespace
	{ goto EXIT; }	        # indicates start of a new record
	else
	{ $record = $line; };	# start of first record

	$$index++;
    }

EXIT:
    XB_Log::log ("info", "<-$subrname ($record)");
    return ($record);
}





############################################################
#
# XB_get_dns_records ( \dnslinelst, \dnsrecordlst )
#
# Returns a list containing all the individual records of the DNS
# database that are extracted from a list of DNS database text lines.
# The list that contains the text lines of the DNS database is
# passed by reference in \dnslinelst, while the list that receives
# the individual records is passed by reference in \dnsrecordlst.
############################################################

sub XB_get_dns_records ($$)
{
    my ($linelstptr, $reclstptr) = @_;
    
    my $index = 0;  my $subrname = "XB_get_dns_records";
    my $record;

    XB_Log::log ("info", "->$subrname (@_)");

    while (1)
    {
	$record = XB_get_dns_record ($linelstptr, \$index);
	if ($record)
	{
	    push (@$reclstptr, $record);
	    next;
	}
	elsif ($index >= scalar (@$linelstptr))
	{  goto EXIT;  }
    }

EXIT:
    XB_Log::log ("info", "<-$subrname");
    return;
}




############################################################
#
# XB_increment_dns_database ( linelstptr )
#
############################################################

sub XB_increment_dns_database ($)
{
    my ($linelstptr) = @_;
    my ($ix, $iy, $line, $field, @fields);
    my $subrname = "XB_increment_dns_database";

    XB_Log::log ("info", "->$subrname (@_)");

    ######################
    # Find SOA record
    ######################

    for ($ix = 0;  $ix < scalar (@$linelstptr);  $ix++)
    {
	$line = $linelstptr->[$ix];
	if ($line =~ /\s+SOA\s+/)  { last; };
    }

    $line = $linelstptr->[++$ix];
    if (!$line)
    { die "XB_increment_dns_serial: No SOA record in DNS database.\n"; };

    #########################################
    # Increment first non-empty field in following line
    #########################################

    @fields = split /\s+/, $line;
    for ($iy = 0;  $iy < scalar (@fields);  $iy++)
    {
	if ($fields[$iy])
	{ $fields[$iy]++;  last; };
    }

    $line = "";
    foreach $field (@fields)
    { $line = join ("\t", $line, $field); };

    $linelstptr->[$ix] = $line . "\n";

    XB_Log::log ("info", "<-$subrname");
}



############################################################
#
# XB_check_dns_additions ( \@fwdlines, \@revlines )
#
# This routine checks to make sure that all the DNS forward
# record additions stored in @XB_Defs::DNS_ADD_FWD_PARMS are
# not already in the \@fwdlines database.  A similar check is
# undertaken in \@revlines for the implied DNS reverse record
# additions.
#
# Upon success the empty string "" is returned.  In this case
# the new forward and reverse mappings will have been added
# to \@fwdlines and \@revlines databases.
#
# Otherwise, a descriptive error message is returned and the
# databases are left unaltered.
############################################################

sub XB_check_dns_additions ($$)
{
    my ($fwdlinesptr, $revlinesptr) = @_;
    my ($newline, @newline, $line, $revaddr, @fields, $field);
    my $subrname = "XB_check_dns_additions";
    my $retval = "";
    my @newfwdlines = ();
    my @newrevlines = ();

    XB_Log::log ("info", "->$subrname (@_)");

    while ($newline = shift (@XB_Defs::DNS_ADD_FWD_PARMS))
    {
	my $tmpline = $newline;      # Because Perl is screwed up!

	$tmpline =~ s/\./\\./g;      # Quote any "." characters.
	@newline = split (/\s+/, $tmpline);

	###################################
	# The new "A" record must not
	# already be in current database.
	###################################

	foreach $line (@$fwdlinesptr)
	{
	    if ($line =~ /^\s*$newline[0].+A\s+$newline[1]/)
	    {
		$retval = "$XB_Defs::DNS_FILE_FORWARD already contains: $line.";
		goto EXIT;
	    };
	};

	###################################
	# As per DNS specification, reverse
	# the order of the address.
	###################################

	$revaddr = "";
	@fields = split (/\./, $newline[1]);
	foreach $field (@fields)  { $revaddr = $field . "." . $revaddr; };
	$revaddr =~ s/\.$//;

	###################################
	# The new "PTR" record must not
	# already be in the current database.
	###################################

	foreach $line (@$revlinesptr)
	{
	    if ($line =~ /^\s*$revaddr\.in-addr\.arpa\..+PTR\s+$newline[0]/)
	    {
		$retval = "$XB_Defs::DNS_FILE_REVERSE already contains: $line.";
		goto EXIT;
	    };
	};

	#########################
	# Add the new "A" record
	#########################

	$newline[0] =~ s/\\//g;	       # Remove search quoting characters.
	$newline[1] =~ s/\\//g;
	$revaddr =~ s/\\//g;

	if ($newline[0] !~ /\.$/)      # Make sure name ends in "."
	{  $newline[0] .= ".";	};

	push (@newfwdlines,
	      "$newline[0]\t$XB_Defs::DNS_TIMEOUT\tIN A\t$newline[1]\n");

	###########################
	# Add the new "PTR" record 
	###########################

	push (@newrevlines,
          "$revaddr.in-addr.arpa.\t$XB_Defs::DNS_TIMEOUT\tIN PTR\t$newline[0]\n");
    }

    #########################################
    # Apply new records to current databases
    #########################################

    foreach $line (@newfwdlines)
    { push (@$fwdlinesptr, $line); };

    foreach $line (@newrevlines)
    { push (@$revlinesptr, $line); };

EXIT:
    XB_Log::log ("info", "<-$subrname ($retval)");
    return ($retval);
}





############################################################
#
# XB_check_dns_deletions ( \@fwdlines, \@revlines )
#
# This routine deletes from \@fwdlines all the DNS forward record
# deletions stored in @XB_Defs::DNS_DEL_FWD_PARMS.  A similar
# is given \@revlines for the implied DNS reverse record
# deletions.
#
# Upon success the empty string "" is returned.  In this case
# the forward and reverse mappings will have been removed
# from the \@fwdlines and \@revlines databases.
#
# Otherwise, a descriptive error message is returned.
# The databases may have been altered upon error return.
############################################################

sub XB_check_dns_deletions ($$)
{
    my ($fwdlinesptr, $revlinesptr) = @_;
    my ($newline, @newline, $revaddr, @fields, $field );
    my ($fwdline, $revline, $found, @newdbase);
    my $retval = "";
    my $subrname = "XB_check_dns_deletions";

    XB_Log::log ("info", "->$subrname (@_)");

    @newdbase = ();

    ###################################
    # All lines above demarcation line
    # may not be deleted.
    ###################################

    while ($fwdline = shift (@$fwdlinesptr))
    {
	push (@newdbase, $fwdline);

	if ($fwdline =~ /DO NOT REMOVE THIS LINE/i)
	{ last; }
    };

    #######################################
    # Create a new forward database without
    # deletions.  Replace old database with
    # new one.  Do same for reverse database.
    #######################################

    foreach $fwdline (@$fwdlinesptr)
    {
	$found = 0;

	foreach $newline (@XB_Defs::DNS_DEL_FWD_PARMS)
	{
	    my $tmpline = $newline;      # Because Perl is screwed up!

	    $tmpline =~ s/\./\\./g;      # Quote any "." characters.
	    @newline = split (/\s+/, $tmpline);

	    if (!defined ($newline[1])) { $newline[1] = ""; };

	    XB_Log::log ("debug0",
			 "newline[0]=$newline[0], newline[1]=$newline[1].");

	    if ($fwdline =~ /^.*?$newline[0].+A.+$newline[1]/)
	    {
	        XB_Log::log ("debug0", "Match found: $fwdline.");

		$found = 1;
	    };
	};

	if (!$found)
	{ push (@newdbase, $fwdline); };   # No matches to DNS_DEL_FWD_PARMS
    };

    @$fwdlinesptr = @newdbase;


    @newdbase = ();

    foreach $revline (@$revlinesptr)
    {
	$found = 0;

	foreach $newline (@XB_Defs::DNS_DEL_FWD_PARMS)
	{
	    my $tmpline = $newline;      # Because Perl is screwed up!

	    $tmpline =~ s/\./\\./g;      # Quote any "." characters.
	    @newline = split (/\s+/, $tmpline);

	    if (!defined ($newline[1])) { $newline[1] = ""; };

	    ###################################
	    # As per DNS specification, reverse
	    # the order of the address.
	    ###################################

	    $revaddr = "";
	    @fields = split (/\./, $newline[1]);
	    foreach $field (@fields)  { $revaddr = $field . "." . $revaddr; };
	    $revaddr =~ s/\.$//;

	    if ($revline =~ /^.*?$revaddr.+PTR.+$newline[0]/)
	    { $found = 1; };
	};

	if (!$found)  { push (@newdbase, $revline); };
    };

    @$revlinesptr = @newdbase;


EXIT:
    XB_Log::log ("info", "<-$subrname ($retval)");
    return ($retval);
}




############################################################
#
# XB_commit_dns_database ( )
#
# Applies all pending DNS add/delete command lines that reside
# in the global variables $XB_Defs::DNS_ADD_FWD_PARMS and
# $XB_Defs::DND_DEL_FWD_PARMS.
#
# The DNS database is exclusively locked and then read.
# The commands are checked one by one against the current DNS
# database for consistency.  Any failure results in an error
# return without application of any of the command to the DNS
# database.
#
# If the commands are all consistent, they are applied to the
# DNS database.
#
# The pending commands are cleared from the global variables
# and the DNS database is unlocked.  Subsequently, if no errors
# occurred, the DNS process is sent a HUP signal that causes it
# to read the new DNS database files.
#
# If a non-empty string is returned, it contains an error
# message.  An empty string implies successful execution.
############################################################

sub XB_commit_dns_database ()
{
    my @fwddnslines = ();
    my @revdnslines = ();
    my $fwdfd = new FileHandle;
    my $revfd = new FileHandle;
    my $fd = new FileHandle;
    my $retval = "";
    my $modified = 0;
    my ($line, @newline, $newline, $pid);
    my $subrname = "XB_commit_dns_database";


    XB_Log::log ("info", "->$subrname (@_)");

    ####################################
    # Perform any pending DNS database
    # file additions or deletions.
    ####################################

    if ( (scalar (@XB_Defs::DNS_ADD_FWD_PARMS)) ||
	 (scalar (@XB_Defs::DNS_DEL_FWD_PARMS)) )
    {

	####################################
	# Open and lock the forward and
	# reverse DNS database files.
	####################################

	if (!$fwdfd->open ("<$XB_Defs::DNS_FILE_FORWARD"))
	{
	    $retval = "cannot open forward Xbone DNS databse.";
	    goto EXIT;
	};

	if (!$revfd->open ("<$XB_Defs::DNS_FILE_REVERSE"))
	{
	    $fwdfd->close;
	    $retval = "cannot open reverse Xbone DNS databse.";
	    goto EXIT;
	};

	XB_file_lock ($fwdfd);
	XB_file_lock ($revfd);

	####################################
	# Read forward and reverse DNS files
	####################################

	while (<$fwdfd>)
	{ push (@fwddnslines, $_); }

	while (<$revfd>)
	{ push (@revdnslines, $_); }
 
	##################################
	# Process pending DNS ADD commands
	##################################

	if (scalar (@XB_Defs::DNS_ADD_FWD_PARMS))
	{
	    $modified = 1;

	    $retval = XB_check_dns_additions (\@fwddnslines, \@revdnslines);
	    if ($retval) { goto CLEANUP; };
	};

	#####################################
	# Process pending DNS DELETE commands
	#####################################

	if (scalar (@XB_Defs::DNS_DEL_FWD_PARMS))
	{
	    $modified = 1;

	    $retval = XB_check_dns_deletions (\@fwddnslines, \@revdnslines);
	    if ($retval) { goto CLEANUP; };
	};

	#######################################
	# If we have modified the databases
	# increment DNS database serial number,
	# write the new temporary database file
	# and rename it to be the new database.
	#######################################

	if ($modified)
	{
	    XB_increment_dns_database (\@fwddnslines);
	    XB_increment_dns_database (\@revdnslines);

	    XB_write_dns_database ("$XB_Defs::DNS_FILE_FORWARD.tmp",
				   \@fwddnslines);
	    XB_write_dns_database ("$XB_Defs::DNS_FILE_REVERSE.tmp",
				   \@revdnslines);

	    if (!rename ("$XB_Defs::DNS_FILE_FORWARD.tmp",
			 $XB_Defs::DNS_FILE_FORWARD) )
	    {
		die ("Cannot rename forward DNS database.\n$!\n");
	    };

	    if (!rename ("$XB_Defs::DNS_FILE_REVERSE.tmp",
			 $XB_Defs::DNS_FILE_REVERSE) )
	    {
		die ("Cannot rename reverse DNS database.\n$!\n");
	    };
	};


CLEANUP:
	XB_file_unlock ($fwdfd);
	XB_file_unlock ($revfd);

	$fwdfd->close;
	$revfd->close;
    };


    ###############################
    # Find the root process that is
    # running the named program.
    ###############################

    if (!$fd->open ("<$XB_Defs::NAMED_PID") )
    {
	$retval = "DNS $XB_Defs::NAMED_PID not found";
	goto EXIT;
    }
    else
    {
	$pid = <$fd>;

	if (!defined ($pid))
	{
	    $fd->close or warn "$!: Close failed";
	    $retval = "DNS $XB_Defs::NAMED_PID empty";
	    goto EXIT;
	};

	chomp ($pid);

	$pid =~ /(\d+)/;	# un-Taint PID, /etc/named.pid is absolute
	$pid = $1;

	$fd->close or warn "$!: Close failed";

	if (!defined ($pid))
	{
	    $retval = "$XB_Defs::NAMED_PID has non-numeric value.";
	    goto EXIT;
	};
    };


    #####################################
    # Give the named program a hup signal
    #####################################

    XB_Log::log ("debug0", "kill HUP => $pid.");

    unless (kill ('HUP', $pid))
    {
        XB_Log::log ("warning", "HUP of DNS failed.");

	$retval = "HUP signal to Xbone DNS failed.";
	goto EXIT;
    };

EXIT:
    XB_Log::log ("info", "<-$subrname ($retval)");
    return ($retval);
}



1;                            # Insure TRUE return if module is interpreted.
