# -*- perl -*- 
#
# DO NOT MOVE THE FIRST LINE
# It identifies the rest of the file as PERL for EMACS autoformatting
# put perl options at the end of that line, e.g., -p
#
# ----------------------------------------------------------------------------
#				    X-BONE
#
#		 USC Information Sciences Institute (USC/ISI)
#		    Marina del Rey, California 90292, USA
#			   Copyright (c) 1998-2000
# ----------------------------------------------------------------------------
#
# Copyright (c) 1998-2000 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
#
# -------------------------------------------------------------------------
#
# Effort sponsored by the Defense Advanced Research Projects Agency (DARPA)
# and Air Force Research Laboratory, Air Force Materiel Command, USAF, under
# agreement number F30602-98-1-0200. The U.S. Government is authorized to 
# reproduce and distribute reprints for Governmental purposes not withstanding
# any copyright annotation therein.
# 
# The views and conclusions contained herein are those of the authors and
# should not be interpreted as necessarily representing the official policies
# or endorsements, either expressed or implied, of the Defense Advanced
# Research Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
# 
# -------------------------------------------------------------------------
# 
#
# ----------------------------------------------------------------------------
#  $RCSfile: XB_Node_DB.pm,v $
#
# $Revision: 1.34 $
#   $Author: touch $
#     $Date: 2000/10/25 22:52:49 $
#    $State: Exp $
# ----------------------------------------------------------------------------
#
#############################################
#
# Xbone host node database routines
#
# Principal author: Gregory Finn
#
#############################################

package XB_Node_DB;


require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(XB_write_state XB_check_node_resources
	     XB_check_access XB_scan_msg_syntax XB_hash_msg
	     XB_read_config_file XB_read_state XB_init_nodeovl
	     XB_remove_ovlstruct XB_overlay_names
	     XB_find_overlay XB_delete_overlays XB_overlay_delete
	     XB_push_overlay_state XB_sweep_overlays
	     XB_increment_overlaycount XB_increment_tunnelcount
             XB_check_release XB_scan_xbone_line);

@EXPORT_OK = qw();


use strict;
use sigtrap;

use Data::Dumper;
use FileHandle;

use XB_Defs;
use XB_Log;

my $XB_debug_flag = 1;		# Set TRUE to enable debug printing.




###################################################################
#
# XB_node_check ( nodehref )
#
# Checks to make sure that an Xbone node that claims to be a router
# is in fact configured to forward IP packets as a router.  It
# returns the empty string "" in the event that the node is configured
# properly.  Otherwise, a non-empty error string is returned.
#
# NOTE: This routine is sensitive to operating system types.  If an
#       unrecognized OS is seen, this routine will return an error.
###################################################################

sub XB_node_check ($)
{
    my ($nodehref) = @_;
    my ($os, $class) = ( $nodehref->{'OS'}, $nodehref->{'Class'} );
    my $result = "";
    my (@tmp, $tmp);

    XB_Log::log ("info", "->XB_node_check (@_)");

    if ($os =~ /freebsd|cairn|kame/i)
    {
	$tmp = `sysctl net.inet.ip.forwarding`;
    }
    elsif ($os =~ /linux|nist/i)
    {
	$tmp = `cat /proc/sys/net/ipv4/ip_forward`;
	$tmp = "placeholder $tmp";

	print "##### tmp = $tmp\n";
    }
    elsif ($os =~ /solaris/i)
    {
	if (($class =~ /host/i) || ($class =~ /router/i))
        {
	   $result = "Solaris OS not supported for either host or router nodes.";
	   goto EXIT;
        }
    }
    else
    {
	$result = "Unrecognized operating system in $XB_Defs::DAEMON_CONF_FILE";
	goto EXIT;
    };

    ##############################
    # At this point we have a node
    # that could legally be either
    # a router or a host.
    ##############################

    if ($class !~ /router/i)  { goto EXIT; };       # No need to check further

    ##############################
    # tmp should hold two-tuple
    # "string value" and value
    # should not be zero.
    ##############################

    @tmp = split /\s+/, $tmp;

    if (!$tmp[1])
    {
	$result = "To be an Xbone router, this node must first be configured\n";
	$result .= "to forward IP packets.";
	goto EXIT;
    };


EXIT:
    XB_Log::log ("info", "<-XB_node_check ($result)");
    return $result;
}





###################################################################
#
# XB_write_state ( nodehref )
#
# Writes the Xbone state as a data structure out to the Xbone state
# file.  This is intended to make the xb_daemon process crash
# recoverable.  The state must be rewritten each time a change is
# made to an overlay.
###################################################################

sub XB_write_state ($)
{
    my ($nodehref) = @_;  my $state;
    my $fh = new FileHandle;  my $timenow = time();

    XB_Log::log ("info", "->XB_write_state (@_)");

    if (!$fh->open ($XB_Defs::DAEMON_STATE_FILE, ">"))
    {
        XB_Log::log ("err", "Unable to open state file: $XB_Defs::DAEMON_STATE_FILE");
	die "$!";
    };

    if (!chmod (0600, $XB_Defs::DAEMON_STATE_FILE))
    {
        XB_Log::log ("err",
		"Unable to protect state file: $XB_Defs::DAEMON_STATE_FILE");
	die "$!";
    };

    $XB_Defs::STATETIME = time();

    $Data::Dumper::Purity = 1;

    $state = Data::Dumper->Dump ([$nodehref, $XB_Defs::STATETIME],
		   [qw(*XB_Defs::DAEMON_STATE XB_Defs::DAEMON_STATE_TIME)]);

    $fh->print ("$state");
    $fh->close;

    XB_Log::log ("info", "<-XB_write_state ()");
}





####################################################################
#
# XB_erase_host_state ()
#
# Clears all route, tunnel and IPsec state in the host.  This is
# meant to be called only by developers and is triggered by the
# XB_Defs::ERASE_AT_STARTUP flag.
####################################################################

sub XB_erase_host_state ()
{
    my $subrname = "XB_erase_host_state";

    XB_Log::log ("info", "-> $subrname ()\n");

    ##########################################
    #  FOR DEVELOPMENT ONLY!!!!
    #  Don't for DNS version of xb_daemon
    ##########################################

    unless (-e $XB_Defs::DNS_FILE_FORWARD)
    {
      XB_Log::log ("warning", "\n****************************\n");
      XB_Log::log ("warning", "ERASE_AT_STARTUP was requested.\n");
      XB_Log::log ("warning", "****************************\n");

      XB_Log::log ("debug0", "Flushing ALL Xbone routes.");

	if (!XB_IPsec::reset ())
	{
	  XB_Log::log ("err", "XB_IPsec::reset() failed");
	    die "";
	}

      XB_Log::log ("debug0", "Flushing Xbone routes.");

	if (!XB_Route::flush_xbone_route ())
	{
	  XB_Log::log ("err", "XB_Route::flush_xbone_route() failed");
	    die "";
	}

      XB_Log::log ("warning", "****************************\n");
      XB_Log::log ("warning", "ERASE_AT_STARTUP finished.\n");
      XB_Log::log ("warning", "****************************\n\n");
    };

    XB_Log::log ("info", "<- $subrname\n");
}





####################################################################
#
# XB_read_state ( nodehref, cfgfile )
#
# The argument, nodehref, is a reference to the daemon state hash, while
# the daemon state configuration file is read from cfgfile.  The
# current host configuration file is always reread to insure that
# any changes made there are incorporated into the state.
####################################################################

sub XB_read_state ($$)
{
    my ($nodehref, $cfgfile) = @_;
    my $fh = new FileHandle;
    my ($dsdata, $end_found, $dstime, %newstate);

    XB_Log::log ("info", "->XB_read_state (@_)");

    XB_Log::log ("debug0", "Initializing node database ...");

    $end_found = 0;

    if ($fh->open ($XB_Defs::DAEMON_STATE_FILE, "<"))
    {
	##############################################
	# Read in state data and look for STATE_TIME
	# written, used to indicate that file is good.
	##############################################

	XB_Log::log ("debug0",
		  "Reading daemon state file: $XB_Defs::DAEMON_STATE_FILE");

	$dsdata = $dstime = "";
READLINE:
	while (<$fh>)
	{
	    if ($_ !~ m'\$XB_Defs::DAEMON_STATE_TIME\s*=')
	    { $dsdata .= $_; }
	    else
	    {
		$end_found = 1;
		$dstime = $_;
		last READLINE;
	    };
	};

	$fh->close;

	if ($end_found)
	{
	    $dsdata =~ /(.*)/s;       # Taint fixup.  State file is in a root
	    $dsdata = $1;	      # privileged directory, so we trust it.

	    eval $dsdata;	      # This EVAL sets $XB_Defs::DAEMON_STATE.

            $nodehref = \%XB_Defs::DAEMON_STATE;

	    $XB_Defs::NODEOS = $XB_Defs::DAEMON_STATE{'OS'};
	};
    };

    ###########################################
    # If we either cannot access the state file
    # or did not see it properly terminated the
    # state is initialized from the config file.
    ###########################################

    if (!$end_found)
    {
        XB_Log::log ("warning", "State data file damaged or missing.");
        XB_Log::log ("warning", "Reinitializing using config file alone.");

        %XB_Defs::DAEMON_STATE = ();

        XB_read_config_file ($cfgfile, \%XB_Defs::DAEMON_STATE);

        $nodehref = \%XB_Defs::DAEMON_STATE;

	$XB_Defs::NODEOS = $XB_Defs::DAEMON_STATE{'OS'};

#{
#    my $state = Dumper ($nodehref);
#    XB_Log::log ("info", "###################################\n");
#    XB_Log::log ("info", "########## INIT STATE ##########\n");
#    XB_Log::log ("info", "\n$state\n");
#    XB_Log::log ("info", "############### END ###############\n");
#    XB_Log::log ("info", "###################################\n");
#}

    };


    eval { XB_IPsec::init() };
    if ($@)
    {
      XB_Log::log ("err", "XB_IPsec::init failed with $@");
	die "";
    };

    eval { XB_Tunnel::init() };
    if ($@)
    {
      XB_Log::log ("err", "XB_Tunnel::init failed with $@");
	die "";
    };

    ###########################################
    # If developing, ensure a clean route,
    # tunnel and IPsec state in the host.
    #
    # NOTE: ERASE_AT_STARTUP will eliminate
    #       ALL host tunnels and IPsec rules,
    #       not just Xbone tunnels and rules.
    ###########################################

    if ($XB_Defs::ERASE_AT_STARTUP)
    {
	XB_erase_host_state ();
    }

    ###########################################
    # If we read what we believe is good state
    # the overlays are reinitialized to reflect
    # that state.
    ###########################################

    XB_Log::log ("debug0", "Restoring using daemon state file ...");

    XB_restore_state ($nodehref);

    XB_Log::log ("debug0", "State has been restored.");

    ####################################
    # Always reread current config file.
    # This can alter some of the just
    # reloaded state data, particularly
    # changed access control lists.
    ####################################

    XB_Log::log ("debug0", "Reading current host config file.");

    %newstate = ();
    XB_read_config_file ( $cfgfile, \%newstate );

    XB_update_daemon_state ($nodehref, \%newstate);
    %$nodehref = %newstate;

    #################################
    # Now that config state is loaded
    # perform gross legality checks.
    #################################

    {
	my $tmp = XB_node_check ($nodehref);

	if ($tmp)
	{
	  XB_Log::log ("err", "\n****************************\n");
	  XB_Log::log ("err", "$tmp\n");
	  XB_Log::log ("err", "****************************\n\n");

	  die "";
	}
    }

EXIT:
    XB_Log::log ("debug0", "Node database initialized.");
    XB_Log::log ("info", "<-XB_read_state ()");
}




#######################################################################
# XB_init_nodeovl ( ovlhref )
#
# The argument, ovlhref, is a reference to an overlay hash.  The
# default name/value pairs for ovlhref are defined.  The time values
# are set to the current time.
#######################################################################

sub XB_init_nodeovl ($)
{
    my ($ovlhref) = @_;

    XB_Log::log ("info", "->XB_init_nodeovl (@_)");

    %$ovlhref = (
		OverlayName => '',
		OverlayManager => '',
		CreatorName => '',
		Email => '',
		Tunnel_Tags => [],
		Routes => [],
		IPsec => [],
		AppProcs => [],
                Create_Time => time(),
		Alive_Time => time()
               );

    XB_Log::log ("info", "<-XB_init_nodeovl ()");
}




#######################################################################
# XB_update_daemon_state ( shref, nshref )
#
# The argument, shref, is a reference to a daemon state hash, while
# nshref is a reference to a newer daemon config hash.
# 
# The dynamic state from shref is copied into nshref.  Currently,
# that state consists of TunnelCount and OverlayCount properties,
# along with dynamic state in the AclLists property.
#######################################################################

sub XB_update_daemon_state ($$)
{
    my ($shref, $nshref) = @_;
    my @keyvalue;

    XB_Log::log ("info", "->XB_update_daemon_state (@_)");

    $nshref->{'TunnelCount'} = $shref->{'TunnelCount'};
    $nshref->{'OverlayCount'} = $shref->{'OverlayCount'};

    XB_update_acllist_state ($shref->{'AclLists'}, $nshref->{'AclLists'});

    XB_Log::log ("info", "<-XB_update_daemon_state ()");
}





#######################################################################
# XB_update_acllist_state ( oacllref, nacllref )
#
# The argument, oacllref, is a reference to the older internal acl list
# state.  The argument nacllref is a reference to a newly read acl list
# from a newer daemon config file.
#
# Each acl from nacllref is updated without affecting its static state.
# It is updated from the state of the matching acl in oacllref.
#
# Currently, the dynamic state updated consists of TunnelCount and
# OverlayCount properties.
#######################################################################

sub XB_update_acllist_state ($$)
{
    my ($oacllref, $nacllref) = @_;
    my ($oaclhref, $naclhref, $match);

    XB_Log::log ("info", "->XB_update_acllist_state (@_)");

    foreach $oaclhref (@$oacllref)
    {
	$match = 0;

	foreach $naclhref (@$nacllref)
	{
	    if ($naclhref->{'UserAuth'} eq $oaclhref->{'UserAuth'})
	    {
		$match = 1;
		XB_update_acl_state ($oaclhref, $naclhref);
		last;
	    };
	};

	########################################
	# If the acl in $oaclhref did not match
	# an acl in the new acl list, we have
	# an obsolete acl.
	########################################

	if (!$match)
	{
	    ########################################
	    # If dynamic state is associated with
	    # this ACL we must temporarily preserve
	    # it, copying it into the new acl list.
	    ########################################

	    if ($oaclhref->{'OverlayCount'} || $oaclhref->{'TunnelCount'})
	    {
		####################################
		# Make sure no security violation by
		# clearing maximum overlay count.
		####################################

		$oaclhref->{'MaxOverlayCount'} = 0;

		push (@$nacllref, $oaclhref);
	    };
	};
    };

    XB_Log::log ("info", "<-XB_update_acllist_state ()");
}





#######################################################################
# XB_update_acl_state ( oaclhref, naclhref )
#
# The argument, oaclhref, is a reference to an older internal acl hash
# state, while naclhref is a reference to a newly read acl from a daemon
# config file.
#
# The dynamic state in naclhref is updated from oclhref.  Currently,
# that state consists of TunnelCount, OverlayCount and Overlays
# properties.
#######################################################################

sub XB_update_acl_state ($$)
{
    my ($oaclhref, $naclhref) = @_;

    XB_Log::log ("info", "->XB_update_acl_state (@_)");

    $naclhref->{'TunnelCount'} = $oaclhref->{'TunnelCount'};
    $naclhref->{'OverlayCount'} = $oaclhref->{'OverlayCount'};
    $naclhref->{'Overlays'} = $oaclhref->{'Overlays'};

    XB_Log::log ("info", "->XB_update_acl_state");
}





#######################################################################
# XB_remove_ovlstruct ( aclhref, ovlname )
#
# The argument, aclhref, is a reference to an access control list hash
# and ovlname is an overlay name string.
#
# The {Overlays} list of overlay hash structures is searched to find
# one with a matching name.  If found, that hash structure is removed
# from the list and true is returned.  Otherwise, false is returned.
#######################################################################

sub XB_remove_ovlstruct ($$)
{
    my ($aclhref, $ovlname) = @_;
    my $href;

    XB_Log::log ("info", "->XB_remove_ovlstruct (@_)");

    foreach $href (@{$aclhref->{'Overlays'}})
    {
	if ($href->{'OverlayName'} eq $ovlname)
	{
	    XB_hlshrink ( $aclhref->{'Overlays'}, $href);

	    XB_Log::log ("info", "<-XB_remove_ovlstruct (1)");
	    return (1);
	};
    };

    XB_Log::log ("info", "<-XB_remove_ovlstruct (0)");
    return (0);
}





#######################################################################
# XB_hlshrink ( lohlref, obshref )
#
# The argument, lohlref, is a reference to a list of hashes, while
# obshref is a reference to an obsolete hash element that is to be deleted
# from the lohlref list.  Upon return, obshref will have been removed from
# the lohlref list.
#######################################################################

sub XB_hlshrink ($$)
{
    my ($lohlref, $obshref) = @_;
    my $href;    my @newacll = ();

    XB_Log::log ("info", "->XB_hlshrink (@_)");

    foreach $href (@$lohlref)
    { if ($href != $obshref) { push (@newacll, $href); }; };

    @$lohlref = @newacll;

    XB_Log::log ("info", "<-XB_hlshrink ()");
}




#######################################################################
# XB_find_overlay ( nodehref, ovlname, ovlmgrname )
#
# The argument, nodehref, is a reference to the daemon state data hash.
# This routine searches all current members of the AclLists for the
# matching overlay.
#
# This routine returns (<aclhref>, <ovlhref>) upon success, where
# <aclhref> is a reference to the AclLists element that contains the
# matching overlay hash and <ovlhref> is a reference to the matched
# overlay hash.  Upon failure to find a match, <aclref> will be FALSE.
#######################################################################

sub XB_find_overlay ($$$)
{
    my ($nodehref, $ovlname, $ovlmgrname) = @_;
    my ($aclhref, $ovlhref, $args);
    my ($maclhref, $movlhref) = (0, 0);

    $args = "$nodehref, $ovlname";
    if (defined ($ovlmgrname)) { $args .= ", $ovlmgrname"; }

    XB_Log::log ("info", "->XB_find_overlay ($args)");

    if (!defined(@{$nodehref->{'AclLists'}}))
    {  goto EXIT;  };

ACLSEARCH:
    foreach $aclhref (@{$nodehref->{'AclLists'}})
    {
	if (!defined(@{$aclhref->{'Overlays'}}))
	{ next; };

	foreach $ovlhref (@{$aclhref->{'Overlays'}})
	{
	    if ($ovlhref->{'OverlayName'} eq $ovlname)
	    {
		$maclhref = $aclhref;
		$movlhref = $ovlhref;
		last ACLSEARCH;
	    }
	}
    };

EXIT:
    XB_Log::log ("info", "<-XB_find_overlay ($maclhref, $movlhref)");
    return ($maclhref, $movlhref);
}





#######################################################################
# XB_push_overlay_state ( nodehref, ovlhref, statename, state )
#
# Adds new state to the overlay hash pointed to by ovlhref.  The
# addition takes the form pushing state onto the list associated with
# the statename hash key.
#
# A TRUE return indicates success, a FALSE return indicates that there
# was no matching overlay named ovlname.
#######################################################################

sub XB_push_overlay_state ($$$$)
{
    my ($nodehref, $ovlhref, $statename, $state) = @_;

    XB_Log::log ("info", "->XB_push_overlay_state (@_)");

    if (!defined ($ovlhref->{$statename})) { $ovlhref->{$statename} = []; };
    push (@{$ovlhref->{$statename}}, $state);

    XB_Log::log ("info", "<-XB_push_overlay_state (1)");
    return (1);
}





#######################################################################
# XB_overlay_names ( nodehref )
#
# Returns a list of all the current overlays residing in the
# current daemon state.  This data takes the form of a list of
# two-tuples: ( [<ovlname>, <ovlmgrname>], ... )
#######################################################################

sub XB_overlay_names ($)
{
    my ($nodehref) = @_;
    my ($aclhref, $ovlhref);
    my @retval = ();

    XB_Log::log ("info", "->XB_overlay_names (@_)");

    if (!defined(@{$nodehref->{'AclLists'}}))
    { return @retval; };

    foreach $aclhref (@{$nodehref->{'AclLists'}})
    {
	if (!defined(@{$aclhref->{'Overlays'}}))
	{ next; };

	foreach $ovlhref (@{$aclhref->{'Overlays'}})
	{ push ( @retval,
		[ $ovlhref->{'OverlayName'}, $ovlhref->{'OverlayManager'} ] ); }
    };

    XB_Log::log ("info", "<-XB_overlay_names (@retval)");
    return @retval;
}





#######################################################################
# XB_read_config_unit ( fh )
#
# The argument, fh, is an opened file handle.  The format of the Xbone
# daemon node configuration file contains long lines of:
#      <name>       <name>=<value> ... <name>=<value>
# that are line-continued with "\" characters.  Comment lines that
# begin with the "#" character are also allowed.  The end of a line is
# signalled by the lack of a continuation character.
#
# This routine reads the file, fh, and strips the comments and
# continuation characters.  It returns a lisy of lines or the empty
# list if the file is exhausted.
#######################################################################

sub XB_read_config_unit (*)
{
    my ($fh) = @_;
    my @unit = ();

    XB_Log::log ("info", "->XB_read_config_unit (@_)");

READLINE:
    while (<$fh>)
    {
	if ($_ =~ /^\s*\x23/)	      # Skip comment lines
	    { next READLINE; }
	elsif ($_ =~ /\\\s*$/)	      # Continuation of line exists
	{
	    chomp;
	    $_ =~ s/\\\s*$//;	      # Remove continuation suffix
	    push (@unit, $_);	      # and accumulate the line
	}
	else
	{			      # End of line continuation
	    chomp;
	    push (@unit, $_);	      # Keep last newline
	    last READLINE;
	};
    };

    XB_Log::log ("info", "<-XB_read_config_unit (@unit)");
    return (@unit);
}





#######################################################################
# XB_read_nvpairs ( href, lineref )
#
# The argument, href, is a reference to a hash that is being built to
# match the configuration <name>=<value> pairs passed in the text
# line referenced by lineref.
#######################################################################

sub XB_read_nvpairs ($$)
{
    my ($href, $lineref) = @_;
    my (@doublet, $pair, $key, $value);

    XB_Log::log ("info", "->XB_read_nvpairs (@_)");

    foreach $pair (@$lineref)
    {
	@doublet = split /\s*=\s*/, $pair;
	$key = shift @doublet;
	$key =~ s/\s+//g;
	$value = shift @doublet;
	$value =~ s/\s+//g;

	######################################
        # Transform "{xxx,yyy}" into "xxx yyy"
	######################################

	if ($value =~ /^{/)
	{
	    $value =~ s/^{//;
	    $value =~ s/}//;
	    $value =~ s/,/ /g;
	};

	$href->{$key} = $value;
    }

    XB_Log::log ("info", "<-XB_read_nvpairs ()");
}





#######################################################################
# XB_read_config_file ( filename, nodehref )
#
# The argument, filename, names the Xbone Daemon node configuration
# file.  It is read in and the corresponding internal node state is
# stored into the hash referenced by nodehref.
#######################################################################

sub XB_read_config_file ($$)
{
    my ($filename, $nodehref) = @_;

    my ($line, @line);
    my $fh = new FileHandle;

    XB_Log::log ("info", "->XB_read_config_file (@_)");

    ########################
    # Open the node database
    ########################

    if (!$fh->open ("<$filename"))
    {
        XB_Log::log ("err", "Cannot open Xbone node database file: $filename");
	die ("$!");
    };

    ##############################
    # Read in the configuration
    # file and fill the database.
    ##############################

READLINE:
    while (@line = XB_read_config_unit ($fh))
    {
	################################
	# Expect either "XboneNode" or
	# "XboneAclList" as initial word
	################################

	$_ = shift @line;

	if (/XboneNode/i)
	{
	    XB_read_nvpairs ($nodehref, \@line);
	}
	elsif (/XboneAclList/i)
	{
	    my %nh = ();
	    my ($aclrefs, $aclref, $size);

	    push (@{$nodehref->{'AclLists'}}, \%nh);    # Add AclList hash ref to
	    $size = scalar @{$nodehref->{'AclLists'}};  # end of AclLists list

	    $aclref = @{$nodehref->{'AclLists'}}[$size-1];

	    XB_read_nvpairs ($aclref, \@line);
	}
    };

    ########################
    # Close the config file
    ########################


    if (!$fh->close)
    {
        XB_Log::log ("err", "Cannot close config file: $filename");
	die ("$!");
    };

    XB_Log::log ("info", "<-XB_read_config_file ()");
}





sub XB_hash_pairs ($)
{
    my ($plref) = @_;
    my ($key, $value, $href, $pl);

    XB_Log::log ("info", "->XB_hash_pairs (@_)");

    $href = {};
    foreach $pl (@$plref)
    {
	($key, $value) = split /=/, $pl;
	$href->{$key} = $value;
    };

    XB_Log::log ("info", "<-XB_hash_pairs ($href)");
    return ($href);
}




##################################################################
# XB_hash_msg ( sref )
#
# Argument, sref, is  a reference to a string that is composed of
# lines of the format:
#        <category name> <name>=<value> ... <name>=<value>
# That is eventualy terminated by a line containing the string
# "XboneEOC".
#
# Returns a hash of hashes, with the outermost hash having the
# <category name> as its key, with the value being an inner hash
# that contains the associated <name>,<value> pairs.
#
# If the <category name> is "XboneNodeCommand", the rest of the
# line is pushed onto the list associated with the  <category name>
# "XboneNodeCommands" (note the plural).  The rest of the line
# associated with this <category name> is unstructured text and
# there may be more than one such line in the string passed.
#
# If the <category name> is "XboneNode", the rest of the line is
# pushed onto the list of hashes associated with the  <category name>
# "XboneNodes" (note the plural).
#
# Illegally formatted data is ignored.
##################################################################

sub XB_hash_msg ($)
{
    my ($sref) = @_;
    my %msghash = ();
    my ($line, $key, $value, $category, $pair, @pairs);

    XB_Log::log ("info", "->XB_hash_msg (@_)");

    foreach $line (split /\n/, $$sref)
    {
	if ($line =~ /\bXboneEOC\b/i)
	{ goto EXIT; };

	$_ = $line;
	@pairs = split;
	$category = shift @pairs;
	if (!$category)  { next; };

	if ($category =~ /XboneNodeCommand/i)
	{
	    if (!defined ($msghash{'XboneNodeCommands'}))
	    { $msghash{'XboneNodeCommands'} = [ "@pairs" ]; }
	    else
	    { push (@{$msghash{'XboneNodeCommands'}}, "@pairs"); }
	}
	elsif ($category =~ /XboneNode/i)
	{
	    if (!defined ($msghash{'XboneNodes'}))
	    { $msghash{'XboneNodes'} = [ XB_hash_pairs (\@pairs) ]; }
	    else
	    { push (@{$msghash{'XboneNodes'}}, XB_hash_pairs (\@pairs) ); }
	}
	else
	{
	    foreach $pair (@pairs)
	    {
		($key, $value) = split /=/, $pair;
		if ($key && $value) { $msghash{$category}{$key} = $value; };
	    }
	};
    };

EXIT: 
    XB_Log::log ("info", "<-XB_hash_msg (%msghash)");
    return %msghash;
}





#####################################################################
# XB_scan_msg_syntax ( mhref )
#
# Performs a basic syntax check of an XboneCommand message sent to
# the Resource Daemon by the Overlay Manager.
#
# The three lines that must be in all such messages are:
#
#    Xbone  Version=<version_number>  Release=<release number>
#    XboneControl  Command=<cmd> Transaction="Request"
#    XboneSecurity  UserId=<user string>  AuthType=<auth>
#
# The <version_number> must agree with  $XB_Defs::XBONE_PROTOCOL.
# The Release=<release_number> must appear and agree with
# $XB_Defs::XBONE_RELEASE.
# 
# These are the only semantic checks performed.
#
# A result of (1, <command>, <msg>) is returned if the XboneControl
# message fails a basic syntax check.  A result of (0, "", "") is
# returned if the check succeeds.
#####################################################################

sub XB_scan_msg_syntax ($)
{
    my ($mhref) = @_;
    my @retval = (0, "", "");
    my ($cmd, $reply_msg) = ("", "");

    XB_Log::log ("info", "->XB_scan_msg_syntax (@_)");


    $cmd = "XboneControl:";

    @_ = XB_scan_xbonecontrol_line ($mhref);
    if (!$_[0])
    {
	@retval = (1, $cmd, $_[1]);
	goto EXIT;
    }

    $cmd = "Xbone:";

    ##################################
    # UDP messages can be duplicated.
    # A Seq=<num> field appears in
    # the Xbone line for some of them.
    ##################################

    if ($mhref->{'XboneControl'}{'Command'} eq 'ResourceDiscovery')
    { @_ = XB_scan_xbone_line ($mhref, 1); }
    else
    { @_ = XB_scan_xbone_line ($mhref); }
	
    if (!$_[0])
    {
	@retval = (1, $cmd, $_[1]);
	goto EXIT;
    }


    $cmd = "XboneSecurity:";

    @_ = XB_scan_xbonesecurity_line ($mhref);
    if (!$_[0])
    {
	@retval = (1, $cmd, $_[1]);
	goto EXIT;
    }


EXIT:
    XB_Log::log ("info", "<-XB_scan_msg_syntax (@retval)");
    return @retval;
}





#####################################################################
# XB_scan_xbone_line ( mhref, seqflg )
#
# Performs a basic syntax check of an Xbone message line.  If seqnumflg
# is true, then the Seq=<seqnum> fields must be included.
#
# If seqflg is provided a Seq=<seqnum> name/value pair must be present.
#
# A result of (1, "") is returned if the Xbone message line passes
# the basic syntax check.  Otherwise, (0, "errmsg") is returned.
#####################################################################

sub XB_scan_xbone_line ($;$)
{
    my ($mhref, $seqflg) = @_;
    my (@retval) = (1, "");

    XB_Log::log ("info", "->XB_scan_xbone_line (@_)");

    if (!defined ($mhref->{'Xbone'}))
    {
	@retval = (0, "Xbone line is missing from message.");
	goto EXIT;
    };

    if (!defined ($mhref->{'Xbone'}{'Version'}))
    {
	@retval = (0, "Xbone Version field is missing from message.");
	goto EXIT;
    }

    if ($mhref->{'Xbone'}{'Version'} ne $XB_Defs::XBONE_PROTOCOL)
    {
	@retval = (0, "Xbone Version protocol number mismatch.");
	goto EXIT;
    }

    if (!defined ($mhref->{'Xbone'}{'Release'}))
    {
	###################################
	# The Release field is not required
	# for DNS related commands.
	###################################

	if ( (defined ($mhref->{'XboneControl'}{'Command'})) &&
	     ($mhref->{'XboneControl'}{'Command'} ne 'DNSUpdate')  )
	{
	    @retval = (0, "Xbone Release field is missing from message.");
	    goto EXIT;
	}
    }
    else
    {
	my ($match, $str);

	$_ = $mhref->{'Xbone'}{'Release'};
	$_ =~ s/\"//m;
	@_ = split;

	$match = 0;
	foreach $str (@_)
	{
	    if ($str eq $XB_Defs::XBONE_RELEASE)
	    { $match = 1; }
	}

	if (!$match)
	{
	    @retval = (0, "Xbone Release number mismatch.");
	    goto EXIT;
	}
    }


    if (defined ($seqflg) && !(defined ($mhref->{'Xbone'}{'Seq'})) )
    {
	@retval = (0, "Xbone Seq field is missing from message.");
	goto EXIT;
    }
	      

EXIT:
    XB_Log::log ("info", "<-XB_scan_xbone_line (@retval)");
    return @retval;
}





#####################################################################
# XB_scan_xbonecontrol_line ( mhref )
#
# Performs a basic syntax check of an XboneControl message line.
#
# A result of (1, "") is returned if the XboneControl line passes
# the basic syntax check.  Otherwise, (0, "errmsg") is returned.
#####################################################################

sub XB_scan_xbonecontrol_line ($)
{
    my ($mhref) = @_;
    my @retval = (1, "");

    XB_Log::log ("info", "->XB_scan_xbonecontrol_line (@_)");


    if (!defined ($mhref->{'XboneControl'}))
    {
	@retval = (0, "XboneControl line is missing from message.");
	goto EXIT;
    }
    else
    {
	if (!defined ($mhref->{'XboneControl'}{'Command'}))
	{
	    @retval = (0, "XboneControl Command field is missing from message.");
	    goto EXIT;
	};

	if (!defined ($mhref->{'XboneControl'}{'Transaction'}))
	{
	    @retval = (0,
		       "XboneControl Transaction field is missing from message.");
	    goto EXIT;
	}

	if ($mhref->{'XboneControl'}{'Transaction'} !~ /Request/i)
	{
	    @retval = (0,
		       "XboneControl Transaction=Request missing from message.");
	    goto EXIT;
	}
    };
	      

EXIT:
    XB_Log::log ("info", "<-XB_scan_xbonecontrol_line (@retval)");
    return @retval;
}





#####################################################################
# XB_scan_xbonesecurity_line ( mhref )
#
# Performs a basic syntax check of an XboneSecurity message line.
#
# A result of (1, "") is returned if the XboneSecurity line passes
# the basic syntax check.  Otherwise, (0, "errmsg") is returned.
#####################################################################

sub XB_scan_xbonesecurity_line ($)
{
    my ($mhref) = @_;
    my @retval = (1, "");

    XB_Log::log ("info", "->XB_scan_xbonesecurity_line (@_)");


    if (!defined ($mhref->{'XboneSecurity'}))
    {
	@retval = (0, "XboneSecurity line is missing from message.");
	goto EXIT;
    }
    else
    {
	if (!defined ($mhref->{'XboneSecurity'}{'UserId'}))
	{
	    @retval = (0, "XboneSecurity UserId field is missing from message.");
	    goto EXIT;
	}

	if (!defined ($mhref->{'XboneSecurity'}{'AuthType'}))
	{
	    @retval = (0, "XboneSecurity AuthType field is missing from message.");
	    goto EXIT;
	}
    };
	      

EXIT:
    XB_Log::log ("info", "<-XB_scan_xbonesecurity_line (@retval)");
    return @retval;
}





#####################################################################
# XB_check_access ( nodehref, mhref )
#
# The argument, mhref, is a reference to the hash of hashes that was
# derived from a message received from the overlay manager.
# The argument, nodehref, is a referendce to this host's Xbone
# data structure.
#
# This routine performs access control checks between the requested
# operation inside the mhref hash and the ACL hashes derived from the
# host Xbone configuration file.
#
# A result of (1, aclhref) is returned if the access check passes,
# where aclhref is a reference to the first ACL hash encountered
# that passed.  A result of (0, <error msg>) is returned if the access
# check failed for all ACL hashes.
#####################################################################

sub XB_check_access ($$)
{
    my ($nodehref, $mhref) = @_;
    my (@result, @retval);

    XB_Log::log ("info", "->XB_check_access (@_)");

    if (!defined ($mhref->{'XboneControl'}{'Command'}))
    {
	@retval = (0, "XboneControl Command field missing");
	goto EXIT;
    };

    if (!defined ($mhref->{'Xbone'}{'Version'}))
    {
	@retval = (0, "Xbone Version field missing");
	goto EXIT;
    };

    if ($XB_Defs::XBONE_PROTOCOL ne $mhref->{'Xbone'}{'Version'})
    {
	@retval = (0, "Xbone Version mismatch");
	goto EXIT;
    };

    #################################
    # Check node-wide resource limits
    #################################

    if ($nodehref->{'OverlayCount'} == $nodehref->{'MaxOverlayCount'} )
    {
	@retval = (0, 'Node maximum overlay count would be exceeded.');
	goto EXIT;
    };

    my ($bool, $string) = XB_check_node_resources ($nodehref, $mhref);
    if (!$bool)
    {
	@retval = (0, $string);
	goto EXIT;
    };

    ############################
    # Check access control lists
    ############################

    @result = XB_candidate_acls ($nodehref, \@{$nodehref->{'AclLists'}}, $mhref);
    if (!$result[0])
    {
	@retval = (0, $result[1]);
	goto EXIT;
    }
    else
    { @retval = (1, $result[1]); };    # Pass back aclhref

EXIT:
    XB_Log::log ("info", "<-XB_check_access (@retval)");
    return (@retval);
}





#####################################################################
# XB_candidate_acls ( nodehref, acllref, mhref )
#
# acllref is a reference to a list of ACL hashes for this host, while
# mhref is a reference to the hash of hashes obtained from an overlay
# manager message.
# The ACLs are matched one at a time with the XboneSecurity UserId
# extracted from mhref.  Each acllref whose UserAuth value pattern
# matches the XboneSecurity UserId value is gathered
# A (1, aclref ... aclref) result is returned if there was a match.
# Otherwise, a (0, <error string>) result will be returned.
#####################################################################

sub XB_candidate_acls ($$$)
{
    my ($nodehref, $acllref, $mhref) = @_;
    my ($aclhref, $user);
    my (@retval) = (1);

    XB_Log::log ("info", "->XB_candidate_acls (@_)");

    ###########################
    # Currently no AuthType
    # verification implemented
    ###########################

    if (!defined ($mhref->{'XboneSecurity'}{'UserId'}))
    {
	@retval = (0, "XboneSecurity UserId missing in overlay manager message");
	goto EXIT;
    }
    else { $user = $mhref->{'XboneSecurity'}{'UserId'}; };

    foreach $aclhref (@$acllref)
    {
	my ($bool, $string);

	#################################
	# Check this ACL element's limits
	#################################

	if ($aclhref->{'OverlayCount'} == $aclhref->{'MaxOverlayCount'} )
	{
	    XB_Log::log ("debug0", "Would exceed this ACL's MaxOverlayCount");
	    next;
	};

#       if ($aclhref->{'AccessLevel'} ne 'deploy' )
#       { next; };

	if ($user =~ /$aclhref->{'UserAuth'}/i)
	{ push (@retval, $aclhref); }
	else
	{
	    XB_Log::log ("debug0",
		      "User=$user does not match ACL's $aclhref->{'UserAuth'}");
	};
    };

    if (!defined ($retval[1]))	# Condition of no aclhref's pushed.
    {
	@retval = (0, 'No ACLs were matched.');
	goto EXIT;
    };

EXIT:
    XB_Log::log ("info", "<-XB_candidate_acls (@retval)");
    return @retval;
}




#####################################################################
# XB_check_node_resources ( nodehref, mhref )
#
# The argument, nodehref, is a reference to the daemon state hash of
# hashes structure.  The argument, mhref, is a reference to the message
# hash of hash structure derived from the overlay manager message.
#
# If the access control checks applied between the ACL hash and the
# message hash pass successfully, a (1, "") result is returned.
# Otherwise, a (0, <error string>) result is returned.
#
# The access control checks consist of a pattern match of all the
# XboneNodes hashes.  Within each hash the property name/value pairs
# are checked to see if they match a property name/value pairs in nodehref.
# If a match is not found, an access control failure has occurred.
#
## The "InterfaceCount" property is special.   For it a numeric
## comparison is performed.
#####################################################################

sub XB_check_node_resources ($$$)
{
    my ($nodehref, $mhref) = @_;
    my ($xbnlref, $xbnhref, $key, $val, @retval);

    XB_Log::log ("info", "->XB_check_node_resources (@_)");

    ###########################################
    # Check all XboneNode hashes for a match.
    ## The "InterfaceCount" property is special.
    ## For it numeric comparisons are performed.
    ###########################################

    $xbnlref = $mhref->{'XboneNodes'};

    if (!defined ($xbnlref))	# If no XboneNode commands, it passes by default.
    {
	@retval = (1, "");
	goto EXIT;
    };

SCANHASH:
    foreach $xbnhref (@$xbnlref)    # Loop over all XboneNode hashes
    {	
	while ( ($key, $val) = each (%$xbnhref) ) # Loop over all <key> <value>
	{			                  # pairs in an XboneNode hash
	    if ($key eq "InterfaceCount")
	    {}

#	    {
#		if ( ($val > $nodehref->{'MaxTunnelCount'}
#		      - $nodehref->{'TunnelCount'} ) ||
#		     ($val > $aclhref->{'MaxTunnelCount'}
#		      - $aclhref->{'TunnelCount'} ) )
#		{ next SCANHASH; }
#	    }

	    elsif (!defined ($nodehref->{$key}))  # Failure, no <key> match
	    {
	        XB_Log::log ("debug0", "Config file $key property absent");
		next SCANHASH;
	    }
	    elsif ($nodehref->{$key} !~ /$val/i)  # Failure, no <value> match
	    {
	        XB_Log::log ("debug0",
			"Config file $key=$nodehref->{$key} does not match $val");
		next SCANHASH;
	    }
	};

	@retval = (1, "");    # Completely satisfied an XboneNode hash
	goto EXIT;
    }

    @retval = (0, "Xbone node resource requests not satisifed.");

EXIT:
    XB_Log::log ("info", "<-XB_check_node_resources (@retval)");
    return @retval;
}





#####################################################################
# XB_delete_overlays ( nodehref, ovlmgr_name, preserve )
#
# Tear down the state associated with all the overlays associated
# with the overlay manager whose canonical name is ovlmgr_name. 
# nodehref is a reference to the daemon state hash.
#
# If ovlmgr_name is not defined or FALSE, all overlays will be deleted.
#
# If preserve is TRUE, the state data associated with the overlays
# WILL NOT BE REMOVED.  If not defined or FALSE, the overlay state data
# will be removed from the aclhref and ovlhref data structures.
#
# The normal return is TRUE.  A FALSE result indicates that an error
# occured while deleting some of the state from an overlay.
#####################################################################

sub XB_delete_overlays ($;$$)
{
    my ($nodehref, $ovlmgr_name, $preserve) = @_;
    my $subrname = "XB_delete_overlays";

    my (@overlays, $ovllref, $aclhref, $ovlhref);
    my $retval = 1;

    if (!defined ($ovlmgr_name))
    { $ovlmgr_name = 0; };

    if (!defined ($preserve))
    { $preserve = 0; };

    XB_Log::log ("info", "->$subrname ($nodehref, $ovlmgr_name, $preserve)");


    #################################
    # Delete tunnel, IPsec and routes
    # associated with all overlays
    #################################

    @overlays = XB_overlay_names ($nodehref);

    if (!scalar (@overlays))
    {
	XB_Log::log ("debug0", "Overlays: NONE FOUND");
	  
        XB_Log::log ("info", "<-$subrname ()");
	return $retval;
    }
    else
    {
	XB_Log::log ("debug0", "Overlays: @overlays");
    };


    foreach $ovllref (@overlays)
    {
	if ($ovlmgr_name)     # Filter on ovlmgr_name if present
	{
	    if ($ovllref->[1] ne $ovlmgr_name)
	    { next; }
	}

	($aclhref, $ovlhref) = XB_find_overlay ($nodehref,
						$ovllref->[0], $ovllref->[1]);
	if (!$aclhref)
	{  $retval = 0; }
	else
	{
	    if (!XB_overlay_delete ($nodehref, $aclhref, $ovlhref, $preserve))
	    { $retval = 0; };
	}
    };


    XB_Log::log ("info", "<-$subrname ($retval)");
    return ($retval);
}




#####################################################################
# XB_overlay_delete ( nodehref, aclhref, ovlhref, preserve )
#
# Tear down the state associated with the overlay referred to be
# aclhref and ovlhref.  This includes tunnels, possible IPsec
# associations, routes and the data structure associated with the
# overlay itself.
#
# The order in which an overlay is torn down is:  First any overlay
# IPsec associations are destroyed.  Second, any associated tunnels
# are removed.  Finally, any associated routes are removed.
#
# If preserve is defined, the state data associated with the overlay
# WILL NOT BE REMOVED.  Otherwise, the overlay state data will be
# removed from the aclhref and ovlhref data structures.
#
# The normal return is TRUE.  A FALSE result indicates that an error
# occurred while deleting some of the state.  This problem must be
# handled by the caller.  However, as much state as possible is
# deleted in all cases.
#
# There should be no die() calls that are untrapped by this routine.
#####################################################################

sub XB_overlay_delete ($$$;$)
{
    my ($nodehref, $aclhref, $ovlhref, $preserve) = @_;
    my (@tmp, $tmp, $rule);
    my $retval = 1;

    my $subrname = "XB_overlay_delete";


    if (!defined ($preserve))
    { $preserve = 0; };

    XB_Log::log ("info", "->$subrname (@_)");

    ########################################
    # IPsec commands create state associated
    # with an overlay that must be cleared.
    ########################################

    if (defined (@{$ovlhref->{'IPsec'}}))
    {
	XB_Log::log ("debug0",
"Deleting <$ovlhref->{'OverlayName'}, $ovlhref->{'OverlayManager'}> IPsec state.");

	foreach $rule (@{$ovlhref->{'IPsec'}})
	{
	    if (!defined ($rule)) { next; };

	    @tmp = split /\s/, $rule;

	    XB_Log::log ("debug0",
			 "XB_IPsec::delete ($tmp[0], $tmp[1], $tmp[2], $tmp[3])");

	    eval { $tmp = XB_IPsec::delete ( $tmp[0], $tmp[1], $tmp[2], $tmp[3]) };
	    if ($@ || (!$tmp))
	    {
		XB_Log::log ("err", "**** FAILURE: XB_IPsec::delete ()");
		$retval = 0;
	    }
	}
    };

    ######################################
    # Each tunnel requires a process.
    # Kill the processes associated with
    # the tunnels for this overlay.
    ######################################

    if (defined (@{$ovlhref->{'Tunnel_Tags'}}))
    {
	my @tdata;

	XB_Log::log ("debug0", "Killing tunnel processes.");

	foreach $rule (@{$ovlhref->{'Tunnel_Tags'}})
	{
	    @tdata = split /\s+/, $rule;

	    XB_Log::log ("debug0", "Calling XB_Tunnel::down ( $tdata[0] )");

	    eval { XB_Tunnel::down ($tdata[0]) };
	    if ($@)
	    {
	        XB_Log::log ("err", "**** FAILURE: XB_Tunnel::down()");
		$retval = 0;
	    };

	    ############################
	    # Remove tunnel count info
	    # unless asked not to.
	    ############################

	    if (!$preserve)
	    {  XB_decrement_tunnelcount ($nodehref, $aclhref);  }
	};

    };

    ###########################################
    # "route add" commands initially received
    # for this node are turned into a string of
    # corresponding "route delete" commands.
    ###########################################

    if (defined (@{$ovlhref->{'Routes'}}))
    {
	XB_Log::log ("debug0",
"Deleting <$ovlhref->{'OverlayName'}, $ovlhref->{'OverlayManager'}> route state.");

	foreach $rule (@{$ovlhref->{'Routes'}})
	{
	    if (!defined ($rule)) { next; };

	    my @args = split /\s+/, $rule;

	    XB_Log::log ("debug0", "DELETING: delete_route (@args)");

	    if (!XB_Route::get_route (@args))
	    {
		XB_Log::log ("debug0", "Route @args already deleted.");
	    }
	    else
	    {
		if (!XB_Route::delete_route (@args))
		{
		    XB_Log::log ("err", "**** FAILURE: XB_Route::delete_route()");
		    $retval = 0;
		}
	    }
	}
    };

#    ########################################
#    # Application execute commands spawn an
#    # inferior tree of processes that must
#    # be halted and removed.  **** NOT IMPLEMENTED NOW
#    ########################################
#
#    if (defined (@{$ovlhref->{'App_PIDs'}}))
#    {
#	XB_Log::log ("debug0",
#"Killing <$ovlhref->{'OverlayName'}, $ovlhref->{'OverlayManager'}> application processes.");
#
#	foreach $rule (@{$ovlhref->{'App_PIDs'}})
#	{
#	    @tmp = split " ", $rule;
#
#	    XB_Log::log ("debug0", "Killing application processes: @tmp");
#
#	    foreach $tmp (@tmp)
#	    {
#		$tmp =~ /(\d+)/;	   # Taint mode fixup.
#		if (!$1) { next; };
#		$tmp = $1;
#
#		XB_Log::log ("debug0", "Issuing KILL ('KILL', $tmp)");
#
#		if (!kill ('KILL', $tmp))
#		{
#		    XB_Log::log ("err", "**** FAILURE: Issuing KILL ('KILL', $tmp)");
#
#		    $retval = 0;
#		}
#	    }
#	}
#    };


    #######################################
    # The overlay has now ceased to exist
    # Remove it from its acllist and update
    # overlay counts unless asked not to.
    #######################################

    if (!$preserve)
    {
	XB_hlshrink ($aclhref->{'Overlays'}, $ovlhref);

	XB_decrement_overlaycount ($nodehref, $aclhref);
    }

    XB_Log::log ("info", "<-$subrname ($retval)"); 
    return ($retval);
}




#####################################################################
# XB_sweep_overlays ( $nodehref )
#
# This routine is called to examine the keep-alive times for each
# overlay.  If the time since an overlay's last keep-alive message
# exceeds $XB_Defs::OVERLAY_ALIVE_TIME, that overlay will be deleted.
#
# The number of overlays deleted by the sweep is returned.
#####################################################################

sub XB_sweep_overlays ($)
{
    my ($nodehref) = @_;
    my (@overlays, $ovllref, $aclhref, $ovlhref);  my $retval = 0;
    my $subrname = "XB_sweep_overlays";


    XB_Log::log ("info", "->$subrname (@_)"); 

    @overlays = XB_overlay_names ($nodehref);

    if (!scalar (@overlays))
    {
        XB_Log::log ("debug0", "Overlays: NONE FOUND");
	goto EXIT;
    }
    else
    { XB_Log::log ("debug0", "Overlays: @overlays"); };

    foreach $ovllref (@overlays)
    {
	($aclhref, $ovlhref) = XB_find_overlay ($nodehref,
						$ovllref->[0], $ovllref->[1]);
	if (!$aclhref)
	{
	    XB_Log::log ("warning", "Overlay @$ovllref data structure is absent");
	    next;
	};

	if ( (time() - $ovlhref->{'Alive_Time'}) > $XB_Defs::OVERLAY_ALIVE_TIME)
	{
            XB_Log::log ("alert",
			 "Overlay @$ovllref deleted (keep-alive expired)");

	      ($aclhref, $ovlhref) = XB_find_overlay ($nodehref,
						      $ovllref->[0],
						      $ovllref->[1]);

	    if (!XB_overlay_delete ($nodehref, $aclhref, $ovlhref))
	    {
                XB_Log::log ("warning", "Overlay $ovllref could not be deleted");
		next;
	    };

	    $retval++;
	};
    };


EXIT:
    XB_Log::log ("info", "<-$subrname ($retval)");
    return $retval;
}





#####################################################################
# XB_restore_state ( $nodehref )
#
# This routine is called at start-up time to recover any overlay
# state that may still be presumed to be active.  Any overlay
# whose last recorded age is less than XB_Defs::OVERLAY_ALIVE_TIME.
#
# Overlays whose state is older than the alive-time limit will be
# deleted and NOT rebuilt.  An attempt will be made to delete and
# rebuild those overlays that remain.  If an eror occurs when trying
# to rebuild an overlay, it will be deleted.
#
# This strategy recognizes that at start-up time the resource daemon
# will not know whether or not it or the kernel has crashed.  Thus it
# cannot assume that kernel state for overlays is consistent.  By
# deleting all overlay state and rebuilding from scratch, a
# consistent internal state can be reached.
#####################################################################

sub XB_restore_state ($)
{
    my ($nodehref) = @_;
    my (%oldnode, $oldnodehref, @overlays, $ovllref, $aclhref, $ovlhref, $tmp);
    my $subrname = "XB_restore_state";


    XB_Log::log ("info", "->$subrname (@_)"); 

    ##################################
    # Delete all overlays that have
    # exceeded their alive time limit.
    ##################################

    XB_Log::log ("debug0", "Deleting expired overlays ..."); 

    $tmp = XB_sweep_overlays ($nodehref);

    XB_Log::log ("debug0", "Expired overlays deleted: $tmp.\n"); 


    #####################################
    # Delete remaining overlay kernel
    # state, but preserve the state data.
    #####################################

    XB_Log::log ("debug0", "Deleting remaining overlay state ..."); 

    XB_delete_overlays ( $nodehref, 0, 1 );

    XB_Log::log ("debug0", "Remaining overlay state deleted.\n"); 

    ########################
    # Recreate overlay state
    ########################

    @overlays = XB_overlay_names ($nodehref);

    if (!scalar (@overlays))
    {
        XB_Log::log ("debug0", "Overlays to be restored: NONE FOUND");
	goto EXIT;
    };

    foreach $ovllref (@overlays)
    {
        XB_Log::log ("debug0",
		     "Restoring overlay <$ovllref->[0], $ovllref->[1]>.");
	
	if (!XB_restore_overlay ($nodehref, $ovllref->[0], $ovllref->[1]))
	{
	  XB_Log::log ("warning",
	     "Overlay <$ovllref->[0], $ovllref->[1]> could not be be restored.");
	};
    };


EXIT:
    #######################
    # Save resulting state.
    #######################

    XB_Log::log ("debug0", "Saving resulting state."); 

    XB_write_state ($nodehref);

    XB_Log::log ("info", "<-$subrname ()"); 
    return;    
}





#####################################################################
# XB_restore_overlay ( nodehref, ovlname, ovlmgr_name )
#
# This routine restores the overlay indicated, whose state may be
# found in the daemon state hash pointed to by nodehref.  If there
# is a problem detected, the overlay is completely deleted.
#
# Upon success a TRUE value is returned, with tunnels, routes, etc.
# restored.
#
# If FLASE is returned, an error occurred in attempting to restore
# the overlay.  In this case all state data regarding the overlay
# is removed from the nodehref daemon state hash.
#####################################################################

sub XB_restore_overlay ($$$)
{
    my ($nodehref, $ovlname, $ovlmgr_name) = @_;
    my ($aclhref, $ovlhref, @args, $retval, $tmp);
    my $subrname = "XB_restore_overlay";


    XB_Log::log ("info", "->$subrname (@_)"); 

    XB_Log::log ("debug0", "Restoring overlay: <$ovlname, $ovlmgr_name>.");

    $retval = 1;

    ########################
    # Find the overlay state
    ########################

    ($aclhref, $ovlhref) = XB_find_overlay ( $nodehref, $ovlname, $ovlmgr_name );
    if (!$aclhref)
    {
        XB_Log::log ("warning",
		     "Structure for <$ovlname, $ovlmgr_name> is absent");
	$retval = 0;
	goto EXIT;
    };


    #################################
    # Recreate the tunnels but gather
    # and store new resulting state.
    #################################

    if (defined (@{$ovlhref->{'Tunnel_Tags'}}))
    {
	my $tag;  my @newstate;

	XB_Log::log ("debug0",
		     "Restoring tunnels for <$ovlname, $ovlmgr_name>.");

	@newstate = ();
	foreach $tmp (@{$ovlhref->{'Tunnel_Tags'}})
	{
	    @args = split /\s+/, $tmp;

	    shift (@args);	# Remove saved tag

	    XB_Log::log ("debug0", "Calling XB_Tunnel::up (@args)");

	    eval { $tag = XB_Tunnel::up (@args) };
	    if ($@)
	    {
	      XB_Log::log ("err", "**** FAILURE: XB_Tunnel::up()");

	      XB_overlay_delete ($nodehref, $aclhref, $ovlhref);

	      $retval = 0;
	      goto EXIT;
	    }

	    push (@newstate, "$tag @args");
	}

	$ovlhref->{'Tunnel_Tags'} = [ @newstate ];
    };


    ##########################
    # Recreate the IPsec state
    ##########################

    if (defined (@{$ovlhref->{'IPsec'}}))
    {
	my ($ix, $args);

        XB_Log::log ("debug0", "Restoring IPsec for <$ovlname, $ovlmgr_name>.");

	foreach $args (@{$ovlhref->{'IPsec'}})
	{
	    @args = split /\s+/, $args;

	    XB_Log::log ("debug0", "XB_IPsec::add (@args)");

	    for ($ix = 3;  $ix <= 7;  $ix++)   # "undef" substitution, see
	    {                                  # the IPsec::add() definition.
		if ($args[$ix] =~ /undef/i)
		{ undef ($args[$ix]); };
	    };

	    eval { $tmp = XB_IPsec::add ($args[0], $args[1], $args[2], $args[3],
					 $args[4], $args[5], $args[6], $args[7]) };

	    if ($@ || (!$tmp))
	    {
	        XB_Log::log ("err", "**** FAILURE: XB_IPsec::add ()");

		XB_overlay_delete ($nodehref, $aclhref, $ovlhref);

		$retval = 0;
		goto EXIT;
	     }
	}
    };


    #####################
    # Recreate the routes
    #####################

    if (defined (@{$ovlhref->{'Routes'}}))
    {
        XB_Log::log ("debug0", "Restoring routes for <$ovlname, $ovlmgr_name>.");

	foreach $tmp (@{$ovlhref->{'Routes'}})
	{
	    @args = split /\s+/, $tmp;

	    XB_Log::log ("debug0", "Calling add_route (@args)");

	    if (!XB_Route::add_route (@args))
	    {
	      XB_Log::log ("debug0", "**** FAILURE: add_route (@args).");

	      XB_overlay_delete ($nodehref, $aclhref, $ovlhref);

	      $retval = 0;
	      goto EXIT;
	    }
	}
    };


EXIT:
    XB_Log::log ("info", "<-$subrname ($retval)"); 
    return $retval;    
}





#####################################################################
# XB_increment_tunnelcount ( nodehref, aclhref )
#
# Adjusts counts for both the entire node and the particular ACL.
# Ensures that upper bound of MaxTunnalCount cannot be exceeded.
# This can occur as result of change made to ACLs while overlays are
# in existance that are associated with that ACL.
#####################################################################

sub XB_increment_tunnelcount ($$)
{
    my ($nodehref, $aclhref) = @_;
    my $subrname = "XB_increment_tunnelcount";

    XB_Log::log ("info", "->$subrname (@_)"); 

    if ($nodehref->{'TunnelCount'} < $nodehref->{'MaxTunnelCount'})
    {  $nodehref->{'TunnelCount'}++;  }

    if ($aclhref->{'TunnelCount'} < $aclhref->{'MaxTunnelCount'})
    {  $aclhref->{'TunnelCount'}++;  }

    XB_Log::log ("info", "<-$subrname ()"); 
    return;
}





#####################################################################
# XB_decrement_tunnelcount ( nodehref, aclhref )
#
# Adjusts counts for both the entire node and the particular ACL.
# Ensures that proper lower bound of zero cannot be exceeded.  This
# can occur as result of change made to ACLs while overlays are
# in existance that are associated with that ACL.
#####################################################################

sub XB_decrement_tunnelcount ($$)
{
    my ($nodehref, $aclhref) = @_;
    my $subrname = "XB_decrement_tunnelcount";

    XB_Log::log ("info", "->$subrname (@_)"); 

    if ($nodehref->{'TunnelCount'} > 0)
    {  $nodehref->{'TunnelCount'}--;  }

    if ($aclhref->{'TunnelCount'} > 0)
    {  $aclhref->{'TunnelCount'}--;  }

    XB_Log::log ("info", "<-$subrname ()"); 
    return;
}





#####################################################################
# XB_increment_overlaycount ( nodehref, aclhref )
#
# Adjusts counts for both the entire node and the particular ACL.
# Ensures that upper bound of MaxOverlayCount cannot be exceeded.
# This can occur as result of change made to ACLs while overlays are
# in existance that are associated with that ACL.
#####################################################################

sub XB_increment_overlaycount ($$)
{
    my ($nodehref, $aclhref) = @_;
    my $subrname = "XB_increment_overlaycount";

    XB_Log::log ("info", "->$subrname (@_)"); 

    if ($nodehref->{'OverlayCount'} < $nodehref->{'MaxOverlayCount'})
    {  $nodehref->{'OverlayCount'}++;  }

    if ($aclhref->{'OverlayCount'} < $aclhref->{'MaxOverlayCount'})
    {  $aclhref->{'OverlayCount'}++;  }

    XB_Log::log ("info", "<-$subrname ()"); 
    return;
}





#####################################################################
# XB_decrement_overlaycount ( nodehref, aclhref )
#
# Adjusts counts for both the entire node and the particular ACL.
# Ensures that proper lower bound of zero cannot be exceeded.  This
# can occur as result of change made to ACLs while overlays are
# in existance that are associated with that ACL.
#####################################################################

sub XB_decrement_overlaycount ($$)
{
    my ($nodehref, $aclhref) = @_;
    my $subrname = "XB_decrement_overlaycount";

    XB_Log::log ("info", "->$subrname (@_)"); 

    if ($nodehref->{'OverlayCount'} > 0)
    {  $nodehref->{'OverlayCount'}--;  }

    if ($aclhref->{'OverlayCount'} > 0)
    {  $aclhref->{'OverlayCount'}--;  }

    XB_Log::log ("info", "<-$subrname ()"); 
    return;
}





#####################################################################
# XB_check_release ( mhref )
#
# Checks compatible releases in message received, passed via mhref,
# with current release of the Xbone system.  Returns the current
# release if it is compatible.  Otherwise, returns empty string.
#####################################################################

sub XB_check_release ($)
{
    my ($mhref) = @_;
    my ($om_version, @om_version, $result);

    my $subrname = "XB_check_release";

    XB_Log::log ("info", "->$subrname (@_)"); 

    $_ = $mhref->{'Xbone'}{'Release'};
    s/\"//m;
    @om_version = split /\s+/, $_;

    #########################
    # Check for compatibility
    #########################

    $result = "";

    foreach $om_version (@om_version)
    {
	if ($om_version eq $XB_Defs::XBONE_RELEASE)
	{
	    $result = $XB_Defs::XBONE_RELEASE;
	    goto EXIT;
	}
    }

    XB_Log::log ("debug0",
	       "Release versions in message not compatible");

EXIT:
    XB_Log::log ("info", "<-$subrname ($result)"); 
    return $result;
}




1;                            # Insure TRUE return if module is interpreted.
