# -*- perl -*- 
# ----------------------------------------------------------------------------
#				    X-BONE
#
#		 USC Information Sciences Institute (USC/ISI)
#		    Marina del Rey, California 90292, USA
#			   Copyright (c) 1998-2000
# ----------------------------------------------------------------------------
#
# Copyright (c) 1998-2000 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
#
# -------------------------------------------------------------------------
#
# Effort sponsored by the Defense Advanced Research Projects Agency (DARPA)
# and Air Force Research Laboratory, Air Force Materiel Command, USAF, under
# agreement number F30602-98-1-0200. The U.S. Government is authorized to 
# reproduce and distribute reprints for Governmental purposes not withstanding
# any copyright annotation therein.
# 
# The views and conclusions contained herein are those of the authors and
# should not be interpreted as necessarily representing the official policies
# or endorsements, either expressed or implied, of the Defense Advanced
# Research Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
# 
# -------------------------------------------------------------------------
# 
#
# ----------------------------------------------------------------------------
#  $RCSfile: XB_Route.pm,v $
#
# $Revision: 1.28 $
#   $Author: yushunwa $
#     $Date: 2000/10/27 23:21:44 $
#    $State: Exp $
# ----------------------------------------------------------------------------
#
#
# Primary Author: Amy Hughes
#
# This script is used in place of the unix 'route' command.  If gated is
# running, the script will modify gated.conf and restart the gated daemon.
# If not, it will issue the unix 'route' command.
# It takes the same input as 'route'. 
#
# usage:
# add_route( [-net|-host], destinationIP, [-netmask mask], 
#                                   (gateway | -interface name))
# delete_route ([-net|-host], destinationIP, [-netmask mask])
# get_route ([-net|-host], destinationIP, [-netmask mask])
# flush_route (prefix, netmask, [prefix, netmask, ...])
# flush_xbone_route()
#
# Supported OS:
# FreeBSD (Default OS)
# Linux (modifications are marked)

package XB_Route;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(add_route delete_route get_route flush_route restart_gated 
		check_gated config_route flush_xbone_route check_mrtd 
		mrtd_init);

use strict;
use sigtrap;
use Fcntl;
use POSIX qw(:errno_h);

use File::Copy;
use Socket;
use IPC::Open3;
use Net::Netmask;
use XB_Log;
use XB_Defs;
use IO::Select;

my $log = 0;

# Variable used to store mrtd password
my $mrtd_pass = "";


###########################################################
#  Declarations
###########################################################
sub restart_gated();


###########################################################
#  Utilities
###########################################################

# Description: 
#	Return interface type of passed interface name.
# Arguments:
#       None.
# Returns:
#	0 if "rip no"
#	1 otherwise
# Exceptions:
#	"XB_Route::check_rip" on error, nothing to clean up by caller
sub check_rip() {
    XB_Log::log "info" => "-> XB_Route::check_rip";
    my $rip = 1;
    
    eval {
	my $gated_conf = $XB_Defs::GATED_PATH . "gated.conf";
	open(GATED, "$gated_conf") or 
	    XB_Log::log "err" => "Can't open $gated_conf: $!" 
		and die "open $gated_conf";
	while (<GATED>) {
	    if (m/rip\s*no/) {
		$rip = 0;
		XB_Log::log "debug" => "XB_Route::check_rip: RIP is off";
		last;
	    }
	}
	close GATED or XB_Log::log "err" => "cannot close $gated_conf: $!"
	    and die "close";
	
	XB_Log::log "debug" => "XB_Route::check_rip: RIP is on";
    };
    XB_Log::log "info" => "<- XB_Route::check_rip $rip";
    return $rip unless $@; #success if no exception
    # exception handling
    unless ($@ =~ /^(open|close)/) {
	XB_Log::log "warning" => "XB_Route::check_rip: caught unexpected exception $@";
    }
    # pass defined exceptions up to caller
    die "XB_Route::check_rip";
} #sub check_rip()


# Description: 
#	Checks if mrtd is running or not.
# Arguments:
#       Flag to indicate if this routine should read from the socket or not.
# Returns
#	A socket to the mrtd process running on this machine
#       0 if mrtd is not running.
# Exceptions:
#	"XB_Route::check_mrtd" on error, nothing to clean up by caller
sub check_mrtd($)
{
    my ($flag) = @_;
    XB_Log::log "info", "-> XB_Route::check_mrtd @_";
    my ($sock, $port, $line);
    
    eval { 
	$sock = IO::Socket::INET->new(PeerAddr =>$XB_Defs::LOOPBACK_IP_ADDRESS,
				      PeerPort =>$XB_Defs::MRTD_PORT,
				      Proto => 'tcp');
	
	# return after cleaning up any junk that we've received.
	if (defined($sock)) {
	  XB_Log::log "info","Connected to mrtd at port ".
		"$XB_Defs::MRTD_PORT $sock";
	    $sock->autoflush(1); 	# for backward compatibility
	    if ($flag) {
		my $buf;
		fcntl($sock, F_SETFL, O_NDELAY) or 
		  XB_Log::log "err", "Error trying to set the socket options"
		      and die "XB_Route::check_mrtd fcntl";
		my ($f, $reply) = mrtd_recv($sock);
		if ($f) {
		  XB_Log::log "err", "mrtd_recv_error" and 
		      die "XB_Route::check_mrtd mrtd_recv";
		}
		if ($mrtd_pass ne "") {
		    if (!mrtd_send($sock, "$mrtd_pass\r\n"))
		    {
		      XB_Log::log "err","Error sending password to mrtd";
			die "XB_Route::check_mrtd password";
		    }
		    ($f, $flag) = mrtd_recv($sock, "MRTd\#");
		    if ($f) {
		      XB_Log::log "err", "mrtd_recv_error" and 
			  die "XB_Route::check_mrtd mrtd_recv";
		    }
		} # if ($mrtd_pass ne "")
	    } # if ($flag)
	} # if (defined($sock) && ....)
    }; # eval block
    
    if (defined($sock)) {
      XB_Log::log "info", "<- XB_Route::check_mrtd $sock";
	return $sock;
    } 
    
    XB_Log::log "info" => "<- XB_Route::check_mrtd 0" ;
    return 0 unless ($@);
    if ($@ =~ /(open|close|Connection)/) {
	undef $@;
      XB_Log::log "warning" => "Couldn't connect to mrtd, assuming that " .
	  "mrtd not running.";
	return 0;
    } elsif ($@ =~ /(fcntl|password|mrtd_send|mrtd_recv)/) {
      XB_Log::log "warning" => "XB_Route::check_mrtd caught exception $@";
    } else {
      XB_Log::log "warning" => "XB_Route::check_mrtd caught unexpected".
	  " exception $@";
    }
    die "XB_Route::check_mrtd";
    
} #sub check_mrtd()
    

# Description: 
#       Used only one time, checks if mrtd is running and a password
#       is required. if yes, then prompts the user for the password.
# Arguments:
#       None
# Returns
#	Nothing.
# Exceptions:
#	"XB_Route::mrtd_init" on error, nothing to clean up by caller
sub mrtd_init() {
  XB_Log::log "info", "->XB_Route::mrtd_init";
    eval {
	my $msock = check_mrtd(0);
        XB_Log::log "info", "Value is $msock!!!";
	if ($msock) {
	    fcntl($msock, F_SETFL, O_NDELAY) or 
	      XB_Log::log "err", "Error trying to set the socket options"
		  and die "XB_Route::mrtd_init fcntl";
	    my ($flag, $message) = mrtd_recv($msock);
	    if ($flag) {
	      XB_Log::log "err", "Error reading from mrtd";
		die "XB_Route::mrtd_init mrtd_recv";
	    }
	    if ($message =~ /password/i) {
		# we need a password
		while (1) {
		    print "Enter password for mrtd:";
		    $mrtd_pass = <STDIN>;
		    chomp($mrtd_pass);
		    if (!mrtd_send($msock,"$mrtd_pass\r\n")) {
		      XB_Log::log "err", "Error sending password";
			die "XB_Route::mrtd_init mrtd_send";
		      }
		    ($flag, $message) = mrtd_recv($msock);
		    if ($flag) {
		      XB_Log::log "err", "Error reading from mrtd";
			die "XB_Route::mrtd_init mrtd_recv";
		    }			
		    if ($message =~ /password/i) {
			print "Incorrect password\n";
		    } else {
			system("clear");
			last ;
		    }
		}
	    } #if ($message =~ /password/i)
	    if (!mrtd_send($msock, "exit\r\n")) {
	      XB_Log::log "err", "Error sending command to exit mrtd";
		die "XB_Route::mrtd_init mrtd_send";
	      }
	} # if ($sock)
    };

    if (($@) && ($@ !~ /(check_mrtd|fcntl|password|mrtd_send|mrtd_recv)/)){
      XB_Log::log "err", "Caught unexpected exception during mrtd_init: $@";
	die "XB_Route::mrtd_init";
    }
    XB_Log::log "info", "<-XB_Route::mrtd_init";
} # sub mrtd_init()


# Description: 
#	Prints the RIP clause to the given file handle, if needed
# Arguments:
#       $fh -- The file handle for gated.conf
#       $rip -- A status variable that is 0 before the RIP clause has been 
#               printed
#       $interface -- the interface ID to be configured
# Returns:
#	1 always
#	also modified $rip
# Exceptions:
#       None
sub print_rip(*$$){
    my($fh, $rip, $interface) = @_;
    XB_Log::log "info" => "-> XB_Route::print_rip $rip $interface";
    
    unless ($$rip) {
	XB_Log::log "debug" => "XB_Route::print_rip: Adding RIP clause to gated.conf";
	
	print $fh "rip yes { #XBONE ADDED\n";
	print $fh "interface $interface noripin noripout; #XBONE ADDED\n";
	print $fh "}; #XBONE ADDED\n\n";
	$$rip++;
    }
    
    XB_Log::log "info" => "<- XB_Route::print_rip $rip $interface";
    return 1;
} # sub print_rip(*$$)

# Description: 
#	Prints the static clause to the given file handle, if needed
# Arguments:
#       $fh -- The file handle for gated.conf
#       $static -- status variable that is 0 before clause is printed
#       $destarg -- "-host" or "-net"
#       $destination -- IP address of destination
#       $netmask -- netmask for destination
#       $gateway -- IP address of gateway
# Returns:
#	1 always
#       also modifies $static
# Exceptions:
#       None
sub print_static(*$$$$$) {
    my($fh, $static, $destarg, $destination, $netmask, $gateway) = @_;
    
    XB_Log::log "info" => "->XB_Route::print_static $static $destarg $destination $netmask $gateway";
    
    unless ($$static) {
	XB_Log::log "debug" => "Adding Static clause to GATED";
	print $fh "static { #XBONE ADDED\n";
	if ($destarg eq "-host") {
	    print $fh "host $destination gateway $gateway; #XBONE ADDED\n";
	} elsif ($destarg eq "-net") {
	    print $fh "$destination mask $netmask gateway $gateway; #XBONE ADDED\n";
	}
	print $fh "}; #XBONE ADDED\n\n";
	$$static++; # set static flag
    }
    
    XB_Log::log "info" => "<- XB_Route::print_static $static $destarg $destination $netmask $gateway";
    return 1;
} # sub print_static(*$$$$$)

# Description: 
#       Makes an immediate backup copy of gated.conf to roll-back in case
#       of local failure
# Arguments:
#       None.
# Returns:
#	1 unless error
# Exceptions:
#	"XB_Route::temp_backup" on error, nothing to clean up by caller
sub temp_backup(){
    XB_Log::log "info" => "->XB_Route::temp_backup";
    
    my $gated_bak = $XB_Defs::GATED_PATH . "gated.conf.bak";
    my $gated_conf = $XB_Defs::GATED_PATH . "gated.conf";
    eval {
	unless (-e $gated_bak) {
	    # backup does not exist, make one
	    XB_Log::log "debug" => "Making a backup for $gated_conf";
	    copy("$gated_conf", "$gated_bak") or
		XB_Log::log "err" => "Copy $gated_conf -> $gated_bak failed: $!"
		    and die "copy";
	}
    }; 
    XB_Log::log "info" => "<- XB_Route::temp_backup";
    return 1 unless $@;
    # exception handling
    unless ($@ =~ /^copy/) {
	XB_Log::log "warning" => "XB_Route::temp_backup: caught unexpected exception $@";
    }
    # pass defined exceptions up to caller
    die "XB_Route::temp_backup()";
} # sub temp_backup()

# Description: 
#       Makes a (pre XBone) backup copy of gated.conf if one doesn't exist
# Arguments:
#       None.
# Returns:
#	1 unless error
# Exceptions:
#	"XB_Route::make_backup" on error, nothing to clean up by caller
sub make_backup(){
    XB_Log::log "info" => "->XB_Route::make_backup";
    
    my $gated_bak = $XB_Defs::GATED_PATH . "gated.conf.xbone.bak";
    my $gated_conf = $XB_Defs::GATED_PATH . "gated.conf";
    eval {
	unless (-e $gated_bak) {
	    # backup does not exist, make one
	    XB_Log::log "debug" => "Making a backup for $gated_conf";
	    copy("$gated_conf", "$gated_bak") or
		XB_Log::log "err" => "Copy $gated_conf -> $gated_bak failed: $!"
		    and die "copy";
	}
	temp_backup or XB_Log::log "err" => "Could not make temp backup" 
	    and die "temp_backup";
    }; 
    XB_Log::log "info" => "<- XB_Route::make_backup";
    return 1 unless $@;
    # exception handling
    unless ($@ =~ /^(copy|temp_backup)/) {
	XB_Log::log "warning" => "XB_Route::make_backup: caught unexpected exception $@";
    }
    # pass defined exceptions up to caller
    die "XB_Route::make_backup()";
} # sub make_backup()

# Description: 
#	expands partial destination IP address as determined by $destarg
# Arguments:
#       $destarg -- "-net" or "-host"
#       $destination -- IP address
# Returns:
#	$destination
# Exceptions:
#	"XB_Route::expand_destination" on error, nothing to clean up by caller
sub expand_destination($$) {
    my ($destarg, $destination) = @_;
    XB_Log::log "info" => "-> XB_Route::expand_destination $destarg $destination";

    # modify destination as specified by $destarg if necessary for IPv4
    my (@dest, @newdest);
    unless ($destination =~ m/^(\d{1,3})(\.(\d{1,3})){3}$/) {
	@dest = split('\.', $destination);
	# the IP address is not complete, must be expanded
	
	eval {
	    if ($destarg eq "-net") {
		# we have a net address
		my @newdest = (0,0,0,0);	
	    } elsif ($destarg eq "-host") {
		# we have a host address
		my @newdest = (0,0,0,pop(@dest));	
	    } else {
		# error
		# when we get to here, shouldn't $destarg be error free?
		XB_Log::log "err" => "Address modifier $destarg is incorrect" and
		    die "destarg";
	    }
	   	    
	    # fill out rest of new destination
	    for (0 .. $#dest) {
		$newdest[$_] = shift @dest;
	    }
	    $destination = join('.', @newdest);
	} 
    };
    XB_Log::log "info" => "<- XB_Route::expand_destination $destarg $destination";
    return $destination unless $@;
    unless ($@ =~ /^destarg/) {
	XB_Log::log "warning" => "XB_Route::expand_destination caught unexpected exception $@";
    }
    die "XB_Route::expand_destionation";
} # sub expand_destination($$)

# Description: 
#	Parse function input
# Arguments:
#       @input -- input string command for route
# Returns:
#	$destarg, $destination, $netmask, $gateway, $interface
# Exceptions:
#	"XB_Route::parse_input" on error, nothing to clean up by caller
sub parse_input(@) {
    my @input = @_;
    my $destarg = "-host";
    my $command;
    my ($destination, $netmask, $gateway, $interface) = ("","","","");

    XB_Log::log "info" => "-> XB_Route::parse_input @input";

    $command = shift @input;

    eval {
	# There must be some input to either function
	unless (@input) {
	    XB_Log::log "err" => "There was no input" and die "empty";
	}
	
	# get destination modifier
	if ($input[0] =~ m/^\-/) {
	    if ($input[0] =~ m/^\-host/) {
		shift @input;
	    } elsif ($input[0] =~ m/^\-net/) {
		$destarg = "-net";
		shift @input;
	    } else {
		XB_Log::log "err" => "Input error, destination modifier $input[0] is invalid"
		    and die "destarg";
	    }
	} # if ($input[0] =~ m/^-/)
	
	unless (@input) {
	    XB_Log::log "err" => "Premature end of input" and die "EOI";
	}
	
	# get destination address
	$destination = shift @input;
	if ($destination =~ m/^(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})$/){
	    if (($1 <= 255) && ($2 <= 255) && ($3 <= 255) && ($4 <= 255)) {
		$destination = $1 . "." . $2 . "." . $3 . "." . $4; 
	    } else {
		XB_Log::log "err" => "Input error, destination address $destination is invalid"
		    and die "destination";
	    }
	    
	} elsif ($destination =~ m/^(\d{1,3})(\.\d{1,3})*(\.\d{1,3})*(\.\d{1,3})*$/) {
	    $destination = expand_destination($destarg, $destination) or
	      XB_Log::log "err" => "Could not expand destination: $destination" and die "destination";
	    
	    unless (defined($destination)) {
		XB_Log::log "err" => "Address could not be expanded" 
		    and die "destination";
	    }
	} else {
	    XB_Log::log "err" => "Input error, destination address $destination is invalid"
		and die "destination";
	}
	
	# if end of input and we are parsing for delete or get, can return now
	unless (@input) {
	    if (($command eq "delete") or ($command eq "get")) {
		return;
	    }
	}
	
	# get netmask if applicable
	if ($input[0] =~ m/netmask$/) {
	    shift @input;
	    unless (@input) {
		XB_Log::log "err" => "Premature end of input" and die "EOI";
	    }
	    $netmask = shift @input;
	    
	    if ($netmask =~ m/^(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})$/){
		if (($1 <= 255) && ($2 <= 255) && ($3 <= 255) && ($4 <= 255)) {
		    $netmask = $1 . "." . $2 . "." . $3 . "." . $4; 
		} else {
		    XB_Log::log "err" => "Input error, netmask $netmask is invalid"
			and die "netmask";
		}
	    } elsif ($netmask =~ m/^(0x[0-9a-f]{8})$/) {
		$netmask =~ $1;
	    } else {
		XB_Log::log "err" => "Input error, netmask $netmask is invalid" 
		    and die "netmask";
	    }
	} # if ($input[0] =~ m/^-netmask$/
	
	# if end of input and we are parsing for delete or get, can return now
	unless (@input) {
	    if (($command eq "delete") || ($command eq "get")) {
		return;
	    } else {
		XB_Log::log "err" => "Premature end of input" and die "EOI";
	      }		
	}
	
	# check for interface or get gateway
	if ($input[0] =~ m/^(-interface|dev)$/) {
	    shift @input;
	    unless (@input) {
		XB_Log::log "err" => "Premature end of input" and die "EOI";
	    }
	    $interface = shift @input;
	    if ($interface =~ m/^(([a-zA-Z]+\d+)(:\d+)?)$/) {
		$interface = $1;
	    } else {
		XB_Log::log "err" => "Interface $interface is invalid" 
		    and die "interface";
	    }
	} elsif ($input[0] =~ m/^(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})$/) {
	    if (($1 <= 255) && ($2 <= 255) && ($3 <= 255) && ($4 <= 255)) {
		$gateway = $1 . "." . $2 . "." . $3 . "." . $4; 
		shift @input;
	    } else {
		XB_Log::log "err" => "Input error, gateway $input[0] is invalid" 
		    and die "gateway";
	    }
	} elsif (@input) {
	    my $error = shift @input;
	    XB_Log::log "err" => "Invalid gateway or interface: $error" 
		and die "gateway";
	} 
	
	if (@input) {
	    XB_Log::log "err" => "End of parsing and still input exists: @input" 
		and die "EOP";
	}
    };

  XB_Log::log "info" => "<- XB_Route::parse_input $destarg $destination $netmask $gateway $interface";
    return ($destarg, $destination, $netmask, $gateway, $interface) unless $@;
    unless ($@ =~ /^(empty|destarg|EOI|destination|netmask|interface|gateway|EOP)/) {
	XB_Log::log "warning" => "XB_Route::parse_input caught unexpected exception$@";
    }
    die "XB_Route::parse_input";
} # sub parse_input(@)

# Description: 
#	Removes all gated alterations with empty clauses
# Arguments:
#       None.
# Returns:
#	Nothing.
# Exceptions:
#	"XB_Route::clean_gated" on error, nothing to clean up by caller
sub clean_gated() {
    XB_Log::log "info" => "-> XB_Route::clean_gated";

    my $gated_conf = $XB_Defs::GATED_PATH . "gated.conf";
    my $gated_new = $XB_Defs::GATED_PATH . "gated.conf.new";
    my $gated_bak = $XB_Defs::GATED_PATH . "gated.conf.xbone.bak";
    
    eval {
	open (GATED, "$gated_conf") 
	    or XB_Log::log "err" => "Could not open $gated_conf: $!"
		and die "open";
	my @gated_conf = <GATED>;
	close GATED or XB_Log::log "err" => "Could not close $gated_conf: $!"
	    and die "close";

	open (GATED, "$gated_bak") 
	    or XB_Log::log "err" => "Could not open $gated_bak: $!"
		and die "open";
	my @gated_bak = <GATED>;
	close GATED or XB_Log::log "err" => "Could not close $gated_bak: $!"
	    and die "close";
	
	my $i;
	my $lim = $#gated_conf;
	foreach $i (0..$#gated_conf){
	    if ($i > $lim) {last};
	    if ($gated_conf[$i] =~ m/^\s*rip\s*yes/) {
		# test for empty rip clause
		if ($gated_conf[$i] =~ m/XBONE/) {
		    # if we modified the line, we may have to
		    # remove it or unmodify it.
		    
		    # get next line and check for empty clause
		    if ($gated_conf[$i+1] =~ m/\}/) { 
			# the clause is empty.
			if ($gated_conf[$i] =~ m/ADDED/) {
			    # we added it, splice it out
			    splice (@gated_conf, $i, 2);
			    $lim-=2;
			} elsif ($gated_conf[$i] =~ m/MODIFIED/) {
			    # get the rip yes line from the backup
			    # and substitute it
			    my @rip_lines = grep m/rip\s*yes/, @gated_bak;
			    $gated_conf[$i] = $rip_lines[0];	
			    splice (@gated_conf, $i+1, 1);
			    $lim--;
			}
		    }		    
		}		
	    } elsif ($gated_conf[$i] =~ m/static/) {
		# test for empty static clause
		# or test for empty proto static clause in non-modified
		# export clause
		    
		if ($gated_conf[$i] =~ m/XBONE/) {
		    # if we added it we have to check if it's empty
		    
		    if ($gated_conf[$i+2] =~ m/\}/) {
			#the clause is empty, splice it out
			splice (@gated_conf, $i, 2);
			$lim-=2;
		    }
		}
	    } elsif ($gated_conf[$i] =~ m/^\s*export/) {
		if ($gated_conf[$i] =~ m/XBONE/) {
		    # we have modified this clause
		    
		    if ($gated_conf[$i+1] =~ m/\}/) {
			# empty modified clause 
			# retrieve unmodified clause and restore
			my @line = split('\{', $gated_conf[$i]);
			my $line = $line[0];
			$line =~ s/\s$//g;

			my @exports = grep m/^$line/, @gated_bak;
			$gated_conf[$i] = $exports[0];
		    } elsif ($gated_conf[$i+1] =~ m/\proto\s*static/) {
			# must test for empty proto static clause 
			
			if ($gated_conf[$i+2] =~ m/\}/) {
			    # proto static clause is empty
			    if ($gated_conf[$i+2] =~ m/XBONE/) {
				# empty proto static clause
				# that we added, 
				
				# check for end of export clause
				if ($gated_conf[$i+3] =~ m/\}/) {
				    # the export clause is completely empty
				    
				    # retrieve original line and 
				    # print only that
				    my @line = split('\{', $gated_conf[$i]);
				    my $line = $line[0];
				    $line =~ s/\s$//g;
				    
				    my @exports = grep m/^$line/, @gated_bak;
				    $gated_conf[$i] = $exports[0];
				    splice (@gated_conf, $i+1, 3);
				    $lim -= 3;
				} else {
				    # the export clause has other info
				    splice (@gated_conf, $i+1, 2);
				    $lim -= 2;
				}
			    }
			} 
		    } 
		} 
	    } 
	} # while ($line = <GATED>) 
	
	open (NEW, ">$gated_new") 
	    or XB_Log::log "err" => "Could not open $gated_new: $!"
		and die "open";
	foreach (@gated_conf) {
	    print NEW;
	}
	close NEW or XB_Log::log "err" => "Could not close $gated_new: $!" 
	    and die "close";
	
	# mv gated.conf.new gated.conf
	move("$gated_new", "$gated_conf") or
	  XB_Log::log "err" => "Move failed: $!" and die "move";
    };
    XB_Log::log "info" => "<- XB_Route::clean_gated";
      return 1 unless $@;
      if (-e $gated_new) {
	  # delete the working file
	  unlink $gated_new;
      }
      unless ($@ =~ /^(open|move|close)/) {
	XB_Log::log "warning" => "XB_Route::clean_gated caught unexpected exception$@";
      }
      die "XB_Route::clean_gated";
  } # sub clean_gated()

# Description :
#      Constructs a route string for static mrtd routes
# Arguments :
#      arguments of the route command to the daemon.
# Returns :
#      Full string to send to mrtd or empty on error
# Exceptions :
#      None
sub mrtd_construct($$$$$) {

  XB_Log::log "info" => "->XB_Route::mrtd_construct @_";
    my ($destarg, $destination, $netmask, $gateway, $interface) = @_;
    my $route = "";
    my $bits = "";
    my $block = new Net::Netmask ($destination, $netmask);

    # first we find $bits depending on $netmask if destination is network.
    if (($destarg =~ /net/) && (defined($netmask)) && ($netmask ne "")) {
	$bits = "/" . $block->bits();
    }
    
    if ((defined($gateway)) && ($gateway ne "")) {
	$route = "ip route $destination$bits $gateway\r\n";
    }
    elsif ((defined($interface)) && ($interface ne "")) {
	$route = "ip route $destination$bits $interface\r\n";
    }

    if ($route eq "") {
      XB_Log::log "err", "Something went wrong, couldn't construct command".
	  " for mrtd";
    }

  XB_Log::log "info" => "<-XB_Route::mrtd_construct $route";
    return $route;
}

# Description:
#       Sends the given command to mrtd
# Arguments: 
#       $sock		socket on which to send the command
#       $command	command to send.
# Returns:
#       1 on success
# Exceptions:
#       "XB_Route::mrtd_send" on error, nothing to clean up by caller
sub mrtd_send($$) {
    
  XB_Log::log "info" => "-> XB_Route::mrtd_send @_";
  my ($sock, $command) = @_;
  my ($return_msg, $reply, $delete_flag);
  my $error_flag = 0;

  #eval {  $sock->send($command); };
  eval {  print $sock $command; };
  if ($@) {
    XB_Log::log "err", "mrtd sock->send($command): $@" 
      and die "XB_Route::mrtd_send";
  }else{
    XB_Log::log "info", "<- XB_Route::mrtd_send @_";
    return 1;
  }
} # sub mrtd_send()



# Description: 
#	Receives from the mrtd process
# Arguments:
#       $sock -- socket on which to communicate with mrtd
#	$end  -- [optional] check for this phrase to end receiving
# Returns:
#       Flag -- 1 on error, 0 otherwise
#	String -- read from mrtd
# Exceptions:
#	None
sub mrtd_recv($;$) {
  XB_Log::log "info", "-> XB_Route::mrtd_recv @_";
  my ($sock, $end) = @_;
  my $return_string = "";
  my $flag = 0;
  my $buff;
  my $matched = 0;
  my $prompt = 0;
  my $trailer ="";
  my $msg = "";

  eval {
    my ($in_bitset, $err_bitset, $in_flag, $err_flag) = ("") x 4;
    my ($events, $timer, $timeout) = (0, 0, $XB_Defs::MRTD_CONNECTION_TIMEOUT);
    my $byte_read;
    vec ($in_bitset, $sock->fileno, 1) = 1; 	# set read bitmask
    $err_bitset = $in_bitset;			# set error bitmask
    my $start_time = time();
    fcntl($sock, F_SETFL, O_NDELAY);		# set non-blocking
    while(1){
      ($events, $timer) = select($in_flag=$in_bitset, undef, 
                                 $err_flag=$err_bitset, $timeout);
      if(vec($in_flag, $sock->fileno, 1)){
        # the real receiving part
        if(defined($byte_read = sysread($sock, $buff, 1000))){
	  unless($byte_read == 0){	# check if EOF (=> socket closed)
	    $return_string .= $buff;
          }
        }
	# check the result to see if MRTd has stopped sending (not close)
	if(defined ($return_string)){
	  if($return_string =~ /More\-\-\s*$/){
	    # special case, send a blank character
	    print $sock " ";
            next;
          }
          if((defined $end)&&($end ne "")&&($return_string =~ /$end\s*$/)){
	    $matched = 1;
	  }
	  if($return_string =~ /(Config>|yes\/no\)|MRTd\#|Password\:)\s*$/){
	    # get to common prompt, stop
	    $prompt = 1;
	    $trailer = $1;
	    last;
	  }
	}
	# check for error conditions from sysread
	if(not defined($byte_read)){
	  if((not defined ($!)) or ($! == 0)){
	    # no error, but EOF => doesn't seem to go here at all
	    die "socket closed";
	  }elsif(defined ($!) && ($! == EWOULDBLOCK)){
	    # this probably means there's more to come; check timeout
	    undef $!;
	    my $current_time = time();
            if($current_time > $start_time + $timeout){
	      die "timeout: receiving MRTd message";
            }else{
	      next;
	    }
	  }else{
	    # some other errors
	    my $errno = $! + 0;
	    die "error: sysread: $errno";
	  }
	}elsif ($byte_read == 0) {
	  # need this since select will pass even if socket is closed
	  die "socket closed";
	}
      }elsif(vec($err_flag, $sock->fileno, 1)){
        die "error: select to MRTd: $!";
      }else{
        die "timeout: select to MRTd";
      }
    }
    
#    # vastly simplified version but bad for CPU cycles
#    my $start_time = time();
#    # loop stops when (1) end prompt met, or (2) timeout
#    while(1){
#      #*** what if <$sock> block? ***
#      while($buff = <$sock>){ $return_string .= $buff; }
#      if($return_string){ 
#        # if "--More--", send a space & reset timer
#        if($return_string =~ /More\-\-\s*$/){
#          print $sock " ";
#          $start_time = time();  # reset the timer
#          next;
#        }
#        if (defined $end && $end ne ""){
#          if($return_string =~ /$end\s*$/){ $matched = 1; }
#        } 
#        # if got common prompts, jump out of the loop
#        if($return_string =~ /(Config>|yes\/no\)|MRTd\#|Password\:)\s*$/){   
#          $prompt = 1;
#          $trailer = $1;              
#          last;
#        }
#      }
#      # if we are still here, check for timeout
#      my $current_time = time();
#      if($current_time > $start_time + $XB_Defs::MRTD_CONNECTION_TIMEOUT){
#        die "timeout";
#      }
#    } # while(1)
  };

  # warning if $end is not found
  if((defined $end && $end ne "") && (! $matched)){
    XB_Log::log "warning", "\n**failed to detect expected string $end**\n";
  }
  # check if select times out
  if($prompt){ 
    $msg = "got prompt: $trailer"; 
  } else {
    $msg = "did not get any prompt";
  }
  XB_Log::log "info", "<- XB_Route::mrtd_recv @_";
  XB_Log::log "debug", "\n==MRTd Return String======\n$return_string".
                       "\n--------------------------\n$msg".
                       "\n==========================\n";
  return ($flag, $return_string) unless $@; 
  # exception handling
  unless($@ =~ /(timeout|error)/){
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Route::mrtd_recv: caught unexpected exception".
      " $@";
  }
  die "XB_Route::mrtd_recv: $@";
} # sub mrtd_recv()

# Description:
#	Writes config to /etc/mrtd.conf (or not), then reboots mrtd (or not) 
# Arguments:
#	$sock	mrtd socket
#	$write  1->write to mrtd.conf
#	$reboot 1->reboot mrtd
# Returns:
#	1 on success
# Exceptions:
#	"XB_Route::mrtd_write_reboot" on error, caller might need to close the
#	mrtd socket if reboot fails
sub mrtd_write_reboot($$$) {
  my ($sock, $write, $reboot) = @_;
  XB_Log::log "info" => "-> XB_Route::mrtd_write_reboot $sock, $write, $reboot";
  my $f = 0;		# mrtd_recv error flag
  my $reply = "";
  eval{
    # check if write to /etc/mrtd.conf
      if($write){ 
	  if (!mrtd_send($sock, "write\r\n")) {
	    XB_Log::log "err", "Error writing to mrtd";
	      die "mrtd_send";
	  }
	  ($f, $reply) = mrtd_recv($sock);
	  if($f){ 
	    XB_Log::log "err" => "mrtd_recv error" and 
		die "XB_Route::mrtd_write_reboot mrtd_recv";
	  }elsif($reply !~ /yes\/no/i) {
	    XB_Log::log "err" => "mrtd unexpected response for write" and
		die "mrtd send";
	  }
	  
	  if (!mrtd_send($sock, "yes\r\n"))
	  {
	    XB_Log::log "err","Error confirming write to config file in mrtd";
	      die "mrtd_send";
	  }
	  ($f, $reply) = mrtd_recv($sock);
	  if($f){
	    XB_Log::log "err" => "mrtd_recv error" and 
		die "XB_Route::mrtd_write_reboot mrtd_recv";
	  }elsif($reply !~ /Writing/i){
	    XB_Log::log "err" => "mrtd unexpected response for write-yes" and
		die "mrtd write-yes";
	  }
      } # if($write)
      
      # check if reboot mrtd
      if($reboot){
	  if (!mrtd_send($sock, "reboot\r\n")) {
	    XB_Log::log "err", "Error sending reboot command to mrtd";
	      die "mrtd_send";
	  }
	  ($f, $reply) = mrtd_recv($sock);
	  if($f){
	    XB_Log::log "err" => "mrtd_recv error" and 
		die "XB_Route::mrtd_write_reboot mrtd_recv";
	  }elsif($reply !~ /yes\/no/i) {
	    XB_Log::log "err" => "mrtd unexpected response for reboot" and
		die "mrtd reboot";
	  }
	  if (!mrtd_send($sock, "yes\r\n")) {
	    XB_Log::log "err", "Error sending a yes to mrtd for reboot";
	      die "mrtd_send";
	  }
	  ($f, $reply) = mrtd_recv($sock);
	  if($f){
	    XB_Log::log "err" => "mrtd_recv error" and 
		die "XB_Route::mrtd_write_reboot mrtd_recv";
	  }elsif($reply !~ /Reboot/i){
	    XB_Log::log "err" => "mrtd unexpected response for reboot" and
		die "mrtd reboot-yes";
	  }
      } # if($reboot)
  };
 XB_Log::log "info" => "<- XB_Route::mrtd_write_reboot $sock, $write, $reboot";
  return 1 unless $@;
  die "XB_Route::mrtd_write_reboot $@";
}


# Description: 
#	Adds the given interface to gated.conf
# Arguments:
#       $sock -- socket on which mrtd is listening
#       $dest -- destination
# Returns:
#	1 unless error
# Exceptions:
#	"XB_Route::mrtd_read" on error, nothing to clean up by caller
sub route_exists($$$$) {
	
  XB_Log::log "info" => "->XB_Route::route_exists @_";
    my ($sock, $dest, $gway, $intf) = @_;
    my $return_flag = 0;
    my $flag;
    my ($buff, $line);
    my @reply;	 

    eval {
	if (!mrtd_send($sock, "show\r\n"))
	    {
		XB_Log::log "err","Error sending command to read ".
		    "configuration from mrtd";
		  die "XB_Route::route_exists mrtd_send";
	      }
	($flag, $buff) = mrtd_recv($sock);
	if ($flag) {
	  XB_Log::log "err", "Error reading from mrtd";
	    die "XB_Route::route_exists mrtd_recv";
	}
	
	@reply = split /\n/, $buff;
	while (@reply) {
	    $line = $reply[0];
	    # we first get the address
	    if ($line =~ m/^route\s+(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})/) {
		my $addr =  $1 . "." . $2 . "." . $3 . "." . $4;
		if ($addr =~ /(^|[^0-9])$dest($|[^0-9])/) { 
		  XB_Log::log "info" => "Got this line $line where match".
		      " occured";
		    $return_flag = 1;
		}
	    }
	    shift @reply;
	}	
    }; # eval block
    
  XB_Log::log "info" => "<-XB_Route::route_exists $flag, returns $return_flag";
    return $return_flag unless ($@);
    if ($@ =~ /(mrtd_send|mrtd_recv)/) {
      XB_Log::log "err", "XB_Route::route_exists caught exception $@";
    } else {
      XB_Log::log "err", "XB_Route::route_exists caught unknown exception $@";
    }
    die "XB_Route::route_exists";
} # sub route_exists

# Description: 
#	Adds the given interface to gated.conf
# Arguments:
#       $interface -- interface ID
# Returns:
#	1 unless error
# Exceptions:
#	"XB_Route::add_interface" on error, nothing to clean up by caller
sub add_interface($) {
    my ($interface) = @_;
    my $rip = 0;

    XB_Log::log "info" => "-> XB_Route::add_interface $interface";

    eval {
	# make a master backup of the gated.conf file, if necessary
	make_backup or XB_Log::log "err" => "Could not backup gated.conf" 
	    and die "make_backup";

	# open conf files
	my $gated_conf = $XB_Defs::GATED_PATH . "gated.conf";
	my $gated_new = $XB_Defs::GATED_PATH . "gated.conf.new";
	open (GATED, "$gated_conf") or 
	    XB_Log::log "err" => "$!: Could not open $gated_conf" 
		and die "open";
	open (NEW, ">$gated_new") or 
	    XB_Log::log "err" => "$!: Could not open $gated_new" 
		and die "open";
	
	# step through conf file and make changes
	while (defined(my $line = <GATED>)) {
	    my %ripints;
	    
	    if ($line =~ m/^\s*rip\s*yes/) {	
		# line starts with "rip"
		XB_Log::log "debug" => "adding RIP statement";
		
		if (($line =~ m/\;/) ||
		    ($line =~ m/\}/)) {
		    # no braces at all or 'rip yes {};'
		    print NEW "rip yes { #XBONE MODIFIED\n";
		    print NEW "interface $interface noripin noripout; #XBONE ADDED\n";
		    print NEW "}; #XBONE ADDED\n";
		} else {
		    # assumes there is an open brace but not a closed one.
		    print NEW $line;
		    
		    # locate place for new statement
		    while (defined($line = <GATED>)) {
			if ($line =~ m/interface/) {
			    # make sure definition is not duplicated
			    
			    # remove leading whitespace and 
			    # split line to find interface name
			    $line =~ s/^\s//g;
			    my @line = split(' ', $line);
			    
			    # record interface name
			    $ripints{$line[1]}++;
			    
			    print NEW $line;
			} elsif (($line =~ m/\}/) ||
				 ($line =~ m/trustedgateways/) ||
				 ($line =~ m/sourcegateways/) ||
				 ($line =~ m/traceoptions/)) {
			    
			    # must now print definition
			    # if it is new
			    unless (exists $ripints{$interface}) {
				print NEW "interface $interface noripin noripout; #XBONE ADDED\n";
			    }
			    print NEW $line;
			    
			    # done with all rip checking
			    last;
			} else {
			    # not at the right place for the new
			    # definition
			    print NEW $line;
			}
		    } # while ($line = <GATED>) 
		}
		$rip++;  # set the RIP flag because line has been placed
	    } elsif ($line =~ m/^\s*static/) {
		my %dests;
		print_rip(\*NEW, \$rip, $interface);
		
		print NEW $line;	
		
	    } elsif ($line =~ m/^\s*import/) {
		# we don't need to modify any import statements
		# but we do have to place the rip and static statements 
		# before any import statements.
		
		print_rip(\*NEW, \$rip, $interface);
		print NEW $line;
	    } elsif ($line =~ m/^\s*export/) {
		# we don't need to modify any export statements
		# but we do have to place the rip and static statements 
		# before any import statements.
		print_rip(\*NEW, \$rip, $interface);
		print NEW $line;
		
	    } else {
		print NEW $line;
	    }
	} # while ($line = <GATED>)
	
	# if we get here and rip and static haven't been found,
	# we must add them.
	print_rip(\*NEW, \$rip, $interface);
	
	# close conf files
	close GATED or XB_Log::log "err" => "Could not close $gated_conf: $!" 
	    and die "close"; 
	close NEW or XB_Log::log "err" => "Could not close $gated_new: $!" 
	    and die "close"; 
	
	# mv gated.conf.new gated.conf
	move("$gated_new", "$gated_conf") or
	    XB_Log::log "err" => "$!: Move failed" and die "move";
	
	# send reboot signal to gated process
	restart_gated or 
	    XB_Log::log "err" => "restart failed" and die "restart";
	
	# remove temporary working file
	my $gated_temp = $XB_Defs::GATED_PATH . "gated.conf.bak";
	unlink $gated_temp;
    };
    XB_Log::log "info" => "<- XB_Route::add_interface $interface";
    return 1 unless $@;
    if ($@ =~ /^(make_backup|open|move|close)/) {
	# remove working file if it still exists
	eval {
	    my $gated_new = $XB_Defs::GATED_PATH . "gated.new";
	    unlink $gated_new;
	};
    } elsif ($@ =~ /^restart/) {
	# undo changes to gated.conf
	eval {
	    my $gated_temp = $XB_Defs::GATED_PATH . "gated.conf.bak";
	    my $gated_conf = $XB_Defs::GATED_PATH . "gated.conf";
	    move ("$gated_temp", "$gated_conf") 
		or "$!: Cleanup move failed" and die "move";
	};
    } else {
      XB_Log::log "warning" => "XB_Route::add_interface caught unexpected exception $@";
    }
    die "XB_Route::add_interface";
} # sub add_interface()

# Description: 
#	Deletes given interface from gated.conf
# Arguments:
#       $interface -- interface ID
# Returns:
#	1 unless error
# Exceptions:
#	"XB_Route::delete_interface" on error, nothing to clean up by caller
sub delete_interface($) {
    my ($interface) = @_;
    XB_Log::log "info" => "-> XB_Route::delete_interface $interface";

    eval {
	# make a local backup in case update fails
	temp_backup or XB_Log::log "err" => "Could not make temp backup of gated.conf" and die "temp_backup";

	#open conf files
	my $gated_conf = $XB_Defs::GATED_PATH . "gated.conf";
	my $gated_new = $XB_Defs::GATED_PATH . "gated.conf.new";
	open (GATED, "$gated_conf")
	    or XB_Log::log "err" => "$!: Could not open $gated_conf"
		and die "open";
	open (NEW, ">$gated_new") 
	    or XB_Log::log "err" => "$!: Could not open $gated_new"
		and die "open";
	
	# first pass at gated.conf
	# remove lines containing the interface
	my $line;
	while (defined($line = <GATED>)) {
	    if ($line =~ m/$interface/) {
		# don't print lines containing destination
		next;
	    } else {
		# default is to print the line
		print NEW $line;
	    }
	} # while ($line = <GATED>)
	close GATED;
	close NEW;
	
	# mv gated.conf.new gated.conf
	move("$gated_new", "$gated_conf") or
	    XB_Log::log "err" => "$!: Move failed" and die "move";
	
	# remove empty clauses that we added/altered
	clean_gated or 
	    XB_Log::log "err" => "Gated cleanup failed" and die "clean";
	
	# send reboot signal to gated process
	restart_gated or
	    XB_Log::log "err" => "Restart failed" and die "restart";

	# remove temporary working file
	my $gated_temp = $XB_Defs::GATED_PATH . "gated.conf.bak";
	unlink $gated_temp;
    };
    XB_Log::log "info" => "<- XB_Route::delete_interface $interface";
    return 1 unless $@;
    if ($@ =~ /^(open|move|clean|temp_backup)/) {
	# remove working file if it still exists
	eval {
	    my $gated_new = $XB_Defs::GATED_PATH . "gated.new";
	    unlink $gated_new;
	};
    } elsif ($@ =~ /^restart/) {
	# undo changes to gated.conf
	eval {
	    my $gated_temp = $XB_Defs::GATED_PATH . "gated.conf.bak";
	    my $gated_conf = $XB_Defs::GATED_PATH . "gated.conf";
	    move ("$gated_temp", "$gated_conf") 
		or "$!: Cleanup move failed" and die "move";
	};
    } else {
      XB_Log::log "warning" => "XB_Route::add_interface caught unexpected exception $@";
    }
    die "XB_Route::delete_interface";
} # sub delete_interface()

###########################################################
# Exported API
###########################################################

# Description: 
#	Checks to see if gated is running
# Arguments:
#       None.
# Returns:
#	0 if gated is not running
#	gated_pid otherwise
# Exceptions:
#	"XB_Route::check_gated" on error, nothing to clean up by caller
sub check_gated() {
    my $pid = 0;
    my $gated_pid = $XB_Defs::GATED_PID_PATH . "/gated.pid";    

    XB_Log::log "info" => "-> XB_Route::check_gated $gated_pid";

    eval {
	if (-e $gated_pid) {
	    open (PID, "$gated_pid")
		or XB_Log::log "err" => "Could not open $gated_pid: $!" 
		    and die "open";
	    $pid = <PID>;
	    close PID or XB_Log::log "err" => "Could not close $gated_pid" and die "close";
	    $pid =~ s/\s//g;
	    
	    if ($pid =~ m/^(\d+)$/) {
		$pid = $1;
		my $ps = "ps";
		open (PS, "$ps $pid |") 
		    or XB_Log::log "err" => "couldn't exec $ps" and die "ps";
	    } else {
		# pid is corrupt, gated must not be running
		XB_Log::log "debug" => "XB_Route::check_gated GATED not running";
		  return 0;
	    }
	    
	    my $line;
	    while (defined($line = <PS>)) {
		my @line = split(" ", $line);
		
		if ($line[0] =~ m/$pid/) {
			# gated is running
		    if ($line[4] =~ m/gated/) {
		      XB_Log::log "debug" => "XB_Route::check_gated GATED running:  pid = $pid";
			return $pid;
		    } else {
			# some other process, gated is not running
			XB_Log::log "debug" => "XB_Route::check_gated GATED not running (no process)";
			  $pid = 0;
			  return $pid;
		    }
		}
	    }
	    close PS or XB_Log::log "err" => "PS closed: $!" and die "ps";
	    
	    # couldn't get info from PS, assume the worst
	    XB_Log::log "debug" => "XB_Route::check_gated GATED not running (couldn't get info from ps)";
	    $pid = 0;
	    return $pid;
	} else {
	    # file doesn't exist or we can't access it
	    XB_Log::log "debug" => "XB_Route::check_gated GATED not running (no file or file inaccessible";
	      $pid = 0;
	      return $pid;
	  }
    };
    XB_Log::log "info" => "<- XB_Route::check_gated $gated_pid";
    return $pid unless $@;
    unless ($@ =~ /^(open|ps)/) {
	XB_Log::log "warning" => "XB_Route::check_gated caught unexpected exception $@";
    }
    die "XB_Route::check_gated";
} # sub check_gated()


# Description: 
#	Restarts the GATED process
# Arguments:
#       None.
# Returns:
#	1 unless error
# Exceptions:
#	"XB_Route::restart_gated" on error, nothing to clean up by caller
sub restart_gated(){
    XB_Log::log "info" => "-> XB_Route::restart_gated";
    
    eval {
	defined(my $gated_pid = check_gated) or 
	    XB_Log::log "err" => "check_gated failed" and die "check_gated";
	
	if ($gated_pid) {
	    kill("HUP", $gated_pid) or 
		XB_Log::log "err" => "kill failed" and die "kill";
	}
    };
    XB_Log::log "info" => "<- XB_Route::restart_gated";
    return 1 unless $@;
    unless ($@ =~ /^kill/) {
	XB_Log::log "warning" => "XB_Route::restart_gated caught unexpected exception $@";
    }
    die "XB_Route::restart_gated";
} # sub restart_gated();

# Description: 
#	Configures the given route
# Arguments:
#       @args -- ifconfig command
# Returns:
#	1 if successful
#       0 if not
# Exceptions:
#	"XB_Route::config_route" on error, nothing to clean up by caller
sub config_route(@) {
    my (@args) = @_;
    my $interface = $args[1];
    my $action = $args[$#args];

  XB_Log::log "info" => "-> XB_Route::config_route @args";
    
    eval {
	# determine if we edit gated.conf
	defined(my $gated_pid = check_gated) or 
	  XB_Log::log "err" => "check_gated error" and die "check_gated";

	# check if we need to alter gated.conf
	if ($gated_pid) {
	    my $gated_temp = $XB_Defs::GATED_PATH . "gated.conf.bak.config";
	    my $gated_conf = $XB_Defs::GATED_PATH . "gated.conf";
	    copy("$gated_conf", "$gated_temp") or
	      XB_Log::log "err" => "Copy $gated_conf -> $gated_temp failed: $!"
		  and die "copy";
		    
	    # check if we need to modify rip (only on rip yes)
	    my $ripstatus = check_rip or XB_Log::log "err" => "Could not check RIP status" and die "check_rip";
	    
	    if ($ripstatus) {
		# rip yes;
		if ($action eq "up") {
		    # check return status, etc
		    add_interface($interface) or 
		      XB_Log::log "err" => "cannot add interface" 
			  and die "add";
		} elsif (($action eq "down") ||
			 ($action eq "delete")) {
		    # check return status, etc
		    delete_interface($interface) or 
		      XB_Log::log "err" => "cannot delete interface" 
			  and die "delete";
		} else {
		  XB_Log::log "err" => "bad rip status" and die "rip";
		}
	    }
	} # if ($gated_pid)
	
	# issue ifconfig command Linux and FreeBSD
	system (@args) == 0 or
	    XB_Log::log "err" => "@args failed with $?" and die "ifconfig";
	    # remove temporary working file
	
	if ($gated_pid) {
	    # remove unnecessary temp file
	    my $gated_temp = $XB_Defs::GATED_PATH . "gated.conf.bak";
	    unlink $gated_temp;
	}
    };
    XB_Log::log "info" => "<- XB_Route::config_route";
    return 1 unless $@;
    if ($@ =~ /^(ifconfig)/) {
	# check for gated and undo gated edits
	eval {
	    defined(my $gated_pid = check_gated) or 
	      XB_Log::log "err" => "check_gated error" and die "check_gated";
	    if ($gated_pid) {
		# restore temp file
		my $gated_temp = $XB_Defs::GATED_PATH . "gated.conf.bak.config";
		my $gated_conf = $XB_Defs::GATED_PATH . "gated.conf";
		move ("$gated_temp", "$gated_conf") 
		    or "$!: Cleanup move failed" and die "move";
		
		# restart gated
		restart_gated;
	    }
	};
    } elsif ($@ =~ /^(check_gated|copy|exists)/) {
	# nothing to clean up
    } elsif ($@ =~ /^(rip|add|delete|check_rip)/) {
	# remove temp backup
	eval {
	    my $gated_temp = $XB_Defs::GATED_PATH . "gated.conf.bak.config";
	    unlink $gated_temp;
	};
    } else {
	XB_Log::log "warning" => "XB_Route::config_route caught unexpected error $@";
    }
    die "XB_Route::config_route"
} # sub config_route(@);

# Description: 
#	Add the given route
# Arguments:
#       @args -- the route command
# Returns:
#	1 on success
# Exceptions:
#	"XB_Route::add_route" on error, nothing to clean up by caller
sub add_route(@) {
    my(@args) = @_;
    my $gated_pid;
    my $mrtd_sock;
    my $static = 0;  # corresponding lines
    my ($destarg, $destination, $netmask, $gateway, $interface);
    my $line;

  XB_Log::log "info" => "-> XB_Route::add_route @args";
    
    eval {
	# determine if we edit gated.conf|mrtd.conf or issue a route command
	defined($gated_pid = check_gated) or 
	    XB_Log::log "err" => "XB_Route::check_gated could not be completed" and
		die "check_gated";

	defined ($mrtd_sock = check_mrtd(1)) or
	  XB_Log::log "err", "Xb_Route::check_mrtd could not be completed" and
	      die "check_mrtd";

	($destarg, $destination, $netmask, $gateway, $interface) = parse_input("add", @args) 
	    or XB_Log::log "err" => "Input could not be parsed" and die "parse";
	
      XB_Log::log "debug" => "add_route: 1: $destarg, 2: $destination, 3: $netmask, 4: $gateway, 5: $interface";
	
	unless ($gated_pid or $mrtd_sock) {
	    # gated and mrtd not running, must issue route command
	    my $cpid;
	    XB_Log::log "debug" => "add_route: Issuing explicit route add";
	    
	  XB_Log::log "info" => "add_route: OS is $XB_Defs::NODEOS";
	    if ($XB_Defs::NODEOS =~ /linux|nist/i) {	
		# Linux OS 
		if ((defined($gateway)) && ($gateway ne "")) {
		    if ((defined($netmask)) && ($netmask ne "")) {
		      XB_Log::log "debug" => "add_route(L1): route -n add $destarg $destination netmask $netmask gw $gateway";
			$cpid = open3(\*WTR, \*RDR, \*ERR, "route", "-n", "add",
				      $destarg, $destination, "netmask", $netmask,
				      "gw", $gateway) or
					XB_Log::log "err" => "$!: Route command failed" and die "route";
		    } else {
		      XB_Log::log "debug" => "add_route(L2): route -n add $destarg $destination gw $gateway";
			$cpid = open3(\*WTR, \*RDR, \*ERR, "route", "-n", "add",
				      $destarg, $destination, "gw", $gateway) or
					XB_Log::log "err" => "$!: Route command failed" and die "route";
		    }
		} elsif ((defined($interface)) && ($interface ne "")) {
		    if ((defined($netmask)) && ($netmask ne "")) {
		      XB_Log::log "debug" => "add_route(L3): route -n add $destarg $destination netmask $netmask dev $interface";
			$cpid = open3(\*WTR, \*RDR, \*ERR, "route", "-n", "add",
				      $destarg, $destination, "netmask", $netmask,
				      "dev", $interface) or
					XB_Log::log "err" => "$!: Route command failed" and die "route";
		    } else {
		      XB_Log::log "debug" => "add_route(L4): route -n add $destarg $destination dev $interface";
			$cpid = open3(\*WTR, \*RDR, \*ERR, "route", "-n", "add",
				      $destarg, $destination, "dev", $interface) or
					XB_Log::log "err" => "$!: Route command failed" and die "route";
		    }
		} else {
		   XB_Log::log "err" => "No destination specified for route command" and 
		      die "route";
		}
	    } else {
		# Default OS (FreeBSD)
		if ((defined($gateway)) && ($gateway ne "")) {
		    if ((defined($netmask)) && ($netmask ne "")) {
		      XB_Log::log "debug" => "add_route(D1): route -n -q add $destarg $destination -netmask $netmask $gateway";
			$cpid = open3(\*WTR, \*RDR, \*ERR, "route", "-n", "-q", "add", 
				      $destarg, $destination, "-netmask", $netmask, 
				      $gateway) or
					XB_Log::log "err" => "$!: Route command failed" and die "route";
			
		    } else {
		      XB_Log::log "debug" => "add_route(D2): route -n -q add $destarg $destination $gateway";
			$cpid = open3(\*WTR, \*RDR, \*ERR, "route", "-n", "-q", "add", 
				      $destarg, $destination, $gateway) or
					XB_Log::log "err" => "$!: Route command failed" and die "route";
		    }
		} elsif ((defined($interface)) && ($interface ne "")) {
		    if ((defined($netmask)) && ($netmask ne "")) {
		      XB_Log::log "debug" => "add_route(D3): route -n -q add $destarg $destination -netmask -interface $interface";
			$cpid = open3(\*WTR, \*RDR, \*ERR, "route", "-n", "-q", "add", 
				      $destarg, $destination, "-netmask", $netmask, 
				      "-interface", $interface) or
					XB_Log::log "err" => "$!: Route command failed" and die "route";
		    } else {
		      XB_Log::log "debug" => "add_route(D4): route -n -q add $destarg $destination -interface $interface";
			$cpid = open3(\*WTR, \*RDR, \*ERR, "route", "-n", "-q", "add", 
				      $destarg, $destination, "-interface", $interface) or
					XB_Log::log "err" => "$!: Route command failed" and die "route";
		    }
		} else {
		  XB_Log::log "err" => "No destination specified for route command" and 
		      die "destination";
		}
	    }  
	    
	    # check output from route and warn if not successful
	    my @route = <ERR>;
	    
	    close WTR or XB_Log::log "err" => "Could not close: $!" and die "close_route";
	    close RDR or XB_Log::log "err" => "Could not close: $!" and die "close_route";
	    close ERR or XB_Log::log "err" => "Could not close: $!" and die "close_route";
	    
	    if (defined $cpid) {
		waitpid $cpid, 0 == $cpid or
		    XB_Log::log "err" => "No child $cpid" and die "wait";
	    }
	    
	    if (defined(@route)) {
		# error
		XB_Log::log "err" => "route command returned an error: @route" 
		    and die "route";
	    } else {
		# no error
		return;
	    }
	    
	    return;
	} # unless ($gated_pid or $mrtd_flag)
	
	if ($gated_pid) {
	
	    # make a master backup of the gated.conf file, if necessary
	    my $result = make_backup or XB_Log::log "err" => "Could not backup gated.conf"
		and die "make_backup";
	    
	    # open conf files
	    my $gated_conf = $XB_Defs::GATED_PATH . "gated.conf";
	    my $gated_new = $XB_Defs::GATED_PATH . "gated.conf.new";
	    open (GATED, "$gated_conf") or 
	      XB_Log::log "err" => "$!: Could not open $gated_conf" 
		  and die "gopen";
	    open (NEW, ">$gated_new") 
		or XB_Log::log "err" => "$!: Could not open $gated_new" 
		    and die "gopen";
	    
	    # step through conf file and make changes
	    while (defined(my $line = <GATED>)) {
		if ($line =~ m/^\s*static/) {
		    my %dests;
		    print NEW $line;	
		    
		  XB_Log::log "debug" => "route_add: Adding STATIC statement\n";
		    
		    # make sure destination isn't already configured
		    # and configure it.
		    while (defined($line = <GATED>)) {
			if ($line =~ m/\}/) {
			    # found end of static clause 
			    if (exists $dests{$destination}) {
				close GATED or XB_Log::log "err" => "Could not close $gated_conf" and die "gclose";
				close NEW or XB_Log::log "err" => "Could not close $gated_new" and die "gclose";
			      XB_Log::log "err" => "Destination $destination already configured"
				  and die "gduplicate";
			    } else {			    
				if (defined($gateway)) {
				    if ($destarg eq "-host") {
					print NEW "host $destination gateway $gateway; #XBONE ADDED\n";
				    } elsif ($destarg eq "-net") {
					print NEW "$destination mask $netmask gateway $gateway; #XBONE ADDED\n";
				    }
				    print NEW $line;
				    last;
				} elsif (defined ($interface)) {
				    if ($destarg eq "-host") {
					print NEW "host $destination interface $interface; #XBONE ADDED\n";
				    } elsif ($destarg eq "-net") {
					print NEW "$destination mask $netmask interface $interface; #XBONE ADDED\n";
				    }
				    print NEW $line;
				    last;
				} else {
				    # error of some sort
				}
			    }
			} else {
			    # record next configured destination
			    my $wholeline;
			    my $currdest;
			    
			    $wholeline = $line;
			    $line =~ s/^\s//g; # chop leading whitespace
			    my @line = split(' ', $line);
			    
			    # record if host or net
			    if ($line[0] eq "host") {
				$currdest = $line[1];
			    } else {
				$currdest = $line[0];
			    }
			    
			    $dests{$currdest}++;
			    print NEW $wholeline;
			}
		    } # while ($line = <GATED>) 
		    $static++; # set static flag
		    
		} elsif ($line =~ m/^\s*import/) {
		    # we don't need to modify any import statements
		    # but we do have to place the rip and static statements 
		    # before any import statements.
		    print_static(\*NEW, \$static, $destarg, $destination, $netmask, $gateway);
		    print NEW $line;
		} elsif ($line =~ m/^\s*export/) {
		    print_static(\*NEW, \$static, $destarg, $destination, $netmask, $gateway);
		    if ($line =~ m/restrict/) {
			# don't alter export...restrict statements
			print NEW $line;
		    } else {
			if ($line =~ m/\;/) {
			    # have to add a subclause
			    
			    # add opening brace
			    $line =~ s/\;//g;
			    $line =~ s/\n//g;
			    $line = $line . " { #XBONE MODIFIED\n";
			    
			    # add proto static clause
			    print NEW $line;
			    print NEW "proto static { #XBONE ADDED\n";
			    print NEW "$destination mask $netmask restrict; #XBONE ADDED\n";
			    print NEW "}; #XBONE ADDED\n";
			    
			    # add closing brace
			    print NEW "}; #XBONE ADDED\n";
			    
			} else {
			    # already has braces
			    # have to find and/or add static statement
			    my $braces = 1; 
			    print NEW $line;
			    
			    while (defined($line = <GATED>)) {
				# look for "proto static"
				if ($line =~ m/proto static/) {
				    print NEW $line;
				    print NEW "$destination mask $netmask restrict; #XBONE ADDED\n";
				    last;
				} else {
				    # keep track of braces so it is added before
				    # end of export statement
				    if ($line =~ /\{/) {
					$braces++;
				    } elsif ($line =~ /\}/) {
					$braces--;
				    }
				    if ($braces == 0) {
					# we're looking at the closing brace 
					# for export clause
					print NEW "proto static { #XBONE ADDED\n";
					print NEW "$destination mask $netmask restrict; #XBONE ADDED\n";
					print NEW "}; #XBONE ADDED\n";
					print NEW $line;
					last;
				    }
				    print NEW $line;
				}
			    }
			}
		    } 		
		} else {
		    print NEW $line;
		}
		
	    } # while ($line = <GATED>)
	    
	    # if we get here and rip and static haven't been found,
	    # we must add them.
	    print_static(\*NEW, \$static, $destarg, $destination, $netmask, $gateway);
	    
	    # close conf files
	    close GATED or XB_Log::log "err" => "Could not close $gated_conf: $!" 
		and die "gclose"; 
	    close NEW or XB_Log::log "err" => "Could not close $gated_new: $!" 
		and die "gclose";
	    
	    # mv gated.conf.new gated.conf
	    move("$gated_new", "$gated_conf") or
	      XB_Log::log "err" => "$!: Move failed" and die "gmove";
	    
	    # send reboot signal to gated process
	    restart_gated or XB_Log::log "err" => "could not restart gated" 
		and die "grestart";
	    
	    # remove local backup
	    my $gated_bak = $XB_Defs::GATED_PATH . "gated.conf.bak";
	    unlink $gated_new;
	} # if ($gated_pid)

	#--------------------------------------------------------------------#
        #  mrtd                                                              #
	#--------------------------------------------------------------------#
        if($mrtd_sock) {

	  # check if special case on linux interface host route
	  if((defined ($interface)) && ($interface ne "") &&
	     ($XB_Defs::NODEOS =~ /(linux|nist)/i)){
	    # special case, use route add
	    XB_Log::log "debug" => "add_route(L4): route -n add $destarg ".
	    		           "$destination dev $interface";
	    my $cpid = open3(\*WTR, \*RDR, \*ERR, "route", "-n", "add",
			     $destarg, $destination, "dev", $interface) or
	    XB_Log::log "err" => "$!: Route command failed" and die "route";
	    my @route = <ERR>;
	    close WTR or XB_Log::log "err" => 
			 "Could not close: $!" and die "close_route";
	    close RDR or XB_Log::log "err" =>
			 "Could not close: $!" and die "close_route";
	    close ERR or XB_Log::log "err" =>
			 "Could not close: $!" and die "close_route";
	    if (defined $cpid) {
	      waitpid $cpid, 0 == $cpid or
	      XB_Log::log "err" => "No child $cpid" and die "wait";
	    }
	    if (defined(@route)) {
	      # error
	      XB_Log::log "err" => "route command returned an error: @route"
	        and die "route";
	    } 
	  } else { 
	    # normal case, use mrtd
	    my $flag = 0;
	    my $reply = "";
	    # first enter config mode.

	    if (!mrtd_send($mrtd_sock, "config\r\n")) {
	      XB_Log::log "err", "Error sending command to enter config mode";
		die "mrtd_send";
	    }
	    ($flag, $reply) = mrtd_recv($mrtd_sock, "Config>");
	    if($flag){ 
	      XB_Log::log "err" => "mrtd_recv error" and 
		  die "XB_Route::add_route mrtd_recv";
	    }elsif($reply !~ /Config>/) {
	      XB_Log::log "err" => "mrtd unexpected response for config" and
		die "mrtd \"config\"";
	    }

	    $flag = 0;
	    my $command = mrtd_construct($destarg, $destination, $netmask, 
                                         $gateway, $interface);
	    if ($command eq "") {
	      XB_Log::log "err", "Command is empty!";
		die "mrtd_command";
	    }

	    if (!mrtd_send($mrtd_sock, $command)) {
	      XB_Log::log "err", "Error sending route command to mrtd";
		die "mrtd_send";
	    }
	    $reply = ""; $flag=0;
	    ($flag, $reply) = mrtd_recv($mrtd_sock, "Config>");
	    if($flag){
	      XB_Log::log "err" => "mrtd_recv error" and 
		  die "XB_Route::add_route mrtd_recv";
	    }elsif($reply !~ /static\s*route/) {
	      # not enough to debug, just sanity check
	      XB_Log::log "err" => "mrtd unexpected response for config" and
	        die "mrtd \"config\"";
	    }

	    if (!mrtd_send($mrtd_sock, "exit\r\n")) {
	      XB_Log::log "err", "Error sending command to exit config mode";
		die "XB_Route::add_route mrtd_send";
	      }
	    ($flag, $reply ) = mrtd_recv($mrtd_sock, "MRTd\#");
	    if ($flag) {
	      XB_Log::log "err", "Error reading from mrtd";
		die "XB_Route::add_route mrtd_recv";
	    }
          }
	  if (!mrtd_send($mrtd_sock, "exit\r\n")) {
	    XB_Log::log "err", "Error sending command to exit mrtd";
	      die "XB_Route::add_route mrtd_send";
	    }
	  $mrtd_sock->close();
	} # if ($mrtd_sock)

    }; # eval block
    XB_Log::log "info" => "<- XB_Route::add_route @args";
    return 1 unless $@;
    if ($@ =~ /^(check_gated|parse|close_route|route|wait|make_backup|check_mrtd|route_exists|mrtd_send|mrtd_command|mrtd_recv)/) {
	# no cleanup
    } elsif ($@ =~ /^(gopen|gduplicate|gclose)/) {
	# remove $gated_new and $gated_bak;
	eval {
	    my $gated_new = $XB_Defs::GATED_PATH . "gated.conf.new";
	    unlink $gated_new;
	    my $gated_bak = $XB_Defs::GATED_PATH . "gated.conf.bak";
	    unlink $gated_new;
  	};
    } elsif ($@ =~ /^(gmove|grestart)/) {
	# restore gated.conf to original and remove backups
	eval {
	    my $gated_new = $XB_Defs::GATED_PATH . "gated.conf.new";
	    unlink $gated_new;
	    my $gated_bak = $XB_Defs::GATED_PATH . "gated.conf.bak";
	    my $gated_conf = $XB_Defs::GATED_PATH . "gated.conf";
	    move ("$gated_bak", "$gated_conf");
	};
    } else {
      XB_Log::log "warning" => "XB_Route::add_route caught unexpected exception $@";
    }
    die "XB_Route::add_route";
} # sub add_route($)

# Description: 
#	Deletes given route
# Arguments:
#       @args -- route delete command
# Returns:
#	1 on success
# Exceptions:
#	"XB_Route::delete_route" on error, nothing to clean up by caller
sub delete_route(@) {
    my(@args) = @_;
    my $gated_pid;
    my ($destarg, $destination, $netmask, $gateway, $interface);
    my $targetgway;
    my $targetiface;
    my $mrtd_sock;
    my %interfaces;
    my %gateways;

    XB_Log::log "info" => "-> XB_Route::delete_route @_";

    eval {
	# determine if we edit gated.conf or issue a route command
	defined($gated_pid = check_gated) or 
	    XB_Log::log "err" => "check_gated() could not be completed" 
		and die "check_gated";

	defined ($mrtd_sock = check_mrtd(1)) or
	  XB_Log::log "err", "Xb_Route::check_mrtd could not be completed" 
	    and die "check_mrtd";
	
	($destarg, $destination, $netmask, $gateway, $interface) = parse_input("delete", @args) or 
	    XB_Log::log "err" => "Input could not be parsed" and die "parse";
	
	# This is the ugly hack for the Linux problem!!!!
	# Basic route command will be issued if gated and mrtd are not 
	# running or mrtd is running but no interface/gateway has been
	# given.
	unless (($gated_pid) or 
		($mrtd_sock and ($interface ne "" or $gateway ne ""))) {
	    my $cpid;
	    # gated not running, must issue route command
	  XB_Log::log "info" => "OS is $XB_Defs::NODEOS";
	    if ($XB_Defs::NODEOS =~ /linux|nist/i) {
		# Linux OS
		if ((defined($netmask)) && ($netmask ne "")) {
		  XB_Log::log "debug" => "delete_route(L1): route -n del $destarg $destination netmask $netmask";
		    $cpid = open3(\*WTR, \*RDR, \*ERR, "route", "-n", "del",
				  $destarg, $destination, "netmask", $netmask) or
				    XB_Log::log "err" => "$!: Route command failed" and die "route";
		} else {
		  XB_Log::log "debug" => "delete_route(L2): route -n del $destarg $destination";
		    $cpid = open3(\*WTR, \*RDR, \*ERR, "route", "-n", "del",
				  $destarg, $destination) or
				    XB_Log::log "err" => "$!: Route command failed" and die "route";
		}
	    } else {	
		# Default OS (FreeBSD)
		if ((defined($netmask)) && ($netmask ne "")) {
		  XB_Log::log "debug" => "delete_route(D1): route -n -q delete $destarg $destination -netmask $netmask";
		    $cpid = open3(\*WTR, \*RDR, \*ERR, "route", "-n", "-q", "delete", 
				  $destarg, $destination, "-netmask", $netmask) or
				    XB_Log::log "err" => "$!: Route command failed" and die "route";
		} else {
		  XB_Log::log "debug" => "delete_route(D2): route -n -q delete $destarg $destination";
		    $cpid = open3(\*WTR, \*RDR, \*ERR, "route", "-n", "-q", "delete", 
				  $destarg, $destination) or
				    XB_Log::log "err" => "$!: Route command failed" and die "route";
		}
	    }
	    
	    # check output from route and warn if not successful
	    my @route = <ERR>;
	    
	    close WTR or XB_Log::log "err" => "Close failed: $!" and die "close_route";
	    close RDR or XB_Log::log "err" => "Close failed: $!" and die "close_route";
	    close ERR or XB_Log::log "err" => "Close failed: $!" and die "close_route"; 
	    
	    if (defined $cpid) {
		waitpid $cpid, 0 == $cpid or
		    XB_Log::log "err" => "No child $cpid" and die "wait";
	    }
	   
	    # NOTE: NO ERROR WHEN DELETING A NON-EXISTING ROUTE. NEED REVIEW!
	    #if (defined(@route)) {
	    ### This is an ugly hack to avoid giving an error if the route 
	    ### didn't exist.
	    if ((defined(@route)) and ($route[0] !~ m/No such process/i)) {    
		# error
		XB_Log::log "err" => "route command returned an error: @route" 
		    and die "route";
	    } else {
		# no error
		return;
	    }

	    # close socket if mrtd is running and we are here
	    if($mrtd_sock){
	      if (!mrtd_send($mrtd_sock, "exit\r\n")) {
		XB_Log::log "Error sending exit command";
		  die "mrtd_send";
	      }
	      $mrtd_sock->close;
	    }

	    return;
	} # unless ($gated_pid) or ($mrtd_sock 
	
	if ($gated_pid) {
	    my $gated_conf = $XB_Defs::GATED_PATH . "gated.conf";
	    my $gated_new = $XB_Defs::GATED_PATH . "gated.conf.new";
	    open (GATED, "$gated_conf")
		or XB_Log::log "err" => "$!: Could not open $gated_conf"
		    and die "open";
	    open (NEW, ">$gated_new") 
		or XB_Log::log "err" => "$!: Could not open $gated_new"
		    and die "open";
	    
	    # remove lines containing the destination
	    my $line;
	    while (defined($line = <GATED>)) {
		if ($line =~ m/$destination/) {
		    # don't print lines containing destination
		    next;
		} else {
		    # default is to print the line
		    print NEW $line;
		}
	    } # while ($line = <GATED>)
	    close GATED or XB_Log::log "err" => "Could not close $gated_conf" and die "close";
	    close NEW or XB_Log::log "err" => "Could not close $gated_conf" and die "close";
	    
	    # mv gated.conf.new gated.conf
	    move("$gated_new", "$gated_conf") or
	      XB_Log::log "err" => "$!: Move failed" and die "move";
	    
	    # remove empty clauses that we added/altered
	    clean_gated or XB_Log::log "err" => "Could not clean_gated" and die "clean_gated";
	    
	    # send reboot signal to gated process
	    restart_gated or XB_Log::log "err" => "Could not restart_gated" and die "restart";
	    
	    # remove local backup
	    my $gated_bak = $XB_Defs::GATED_PATH . "gated.conf.bak";
	    unlink $gated_new;
	    
	} # if ($gated_pid)

	# To enter into mrtd code, mrtd must be running and either a gateway
	# or an interface must be given.
	if (($mrtd_sock) and (($interface ne "") or ($gateway ne ""))) {
	  # first we enter config mode.
	  my $flag;
	  my $reply = "";
	  if (!mrtd_send($mrtd_sock, "config\r\n")) {
	    XB_Log::log "err", "Error sending command to enter config mode";
	      die "XB_Route::delete_route mrtd_send";
	  }
	  ($flag, $reply) = mrtd_recv($mrtd_sock, "Config>");
	  if($flag) {
	    XB_Log::log "err" => "mrtd_recv error" and 
		die "XB_Route::delete_route mrtd_recv";
	  }elsif ($reply !~ /Config/){
	    XB_Log::log "err" => "mrtd unexpected response for config" and
	      die "mrtd config";
	  }
	  my $command = mrtd_construct($destarg, $destination, $netmask, 
				       $gateway, $interface);
	  if ($command eq "") {
	    XB_Log::log "err", "Couldn't construct command for delete";
	      die "XB_Route::delete_route mrtd_command";
	  }
	  $command = "no " . $command;
	  if (!mrtd_send($mrtd_sock, $command)) {
	    XB_Log::log "err", "Error sending command to delete route";
	      die "XB_Route::delete_route mrtd_send";
	  }
	  ($flag, $reply) = mrtd_recv($mrtd_sock, "Config>");
	  if ($flag) {
	    XB_Log::log "err" => "mrtd_recv error" and 
	      die "XB_Route::delete_route mrtd_recv";
	  } elsif ($reply !~ /being deleted/i){
	    XB_Log::log "err" => "mrtd unexpected response for $command" and
	      die "mrtd $command";
	  }
	  if (!mrtd_send($mrtd_sock, "exit\r\n")) {
	    XB_Log::log "err", "Error sending command to exit config mode";
	      die "mrtd_send";
	  }
	  ($flag, $reply) = mrtd_recv($mrtd_sock, "MRTd\#");
	  if ($flag) {
	    XB_Log::log "err", "Error reading from mrtd" and 
	      die "XB_Route::delete_route mrtd_recv";
	  }
	  # write to mrtd.conf if linux to clean up /etc/mrtd.conf
	  # if($XB_Defs::NODEOS =~ /(linux|nist)/i){
	  #   mrtd_write_reboot($mrtd_sock, 1, 0);  
   	  # }
	  if (!mrtd_send($mrtd_sock, "exit\r\n")) {
	    XB_Log::log "err", "Error sending command to exit mrtd";
	      die "XB_Route::delete_route mrtd_send";
	  }
	  $mrtd_sock->close();
	} # if ($mrtd_flag)
    }; # eval block


  XB_Log::log "info" => "<- XB_Route::delete_route @args";
    return 1 unless $@;
    if ($@ =~ /^(check_gated|check_mrtd|parse|close_route|route|wait|make_backup|mrtd_send|mrtd_command)/) {
	# no cleanup
    } elsif ($@ =~ /^(open|close)/) {
	# remove $gated_new and $gated_bak;
	eval {
	    my $gated_new = $XB_Defs::GATED_PATH . "gated.conf.new";
	    unlink $gated_new;
	    my $gated_bak = $XB_Defs::GATED_PATH . "gated.conf.bak";
	    unlink $gated_new;
  	};
    } elsif ($@ =~ /^(move|clean_gated|restart)/) {
	# restore gated.conf to original and remove backups
	eval {
	    my $gated_new = $XB_Defs::GATED_PATH . "gated.conf.new";
	    unlink $gated_new;
	    my $gated_bak = $XB_Defs::GATED_PATH . "gated.conf.bak";
	    my $gated_conf = $XB_Defs::GATED_PATH . "gated.conf";
	    move ("$gated_bak", "$gated_conf");
	};
    } else {
      XB_Log::log "warning" => "XB_Route::delete_route caught unexpected exception $@";
    }
    die "XB_Route::delete_route $@";
} # sub delete_route(@)

# Description: 
#	Retrieve configuration for given route
# Arguments:
#       @args -- route command
# Returns:
#	1 if route is configured
#	0 otherwise
# Exceptions:
#	"XB_Route::get_route" on error, nothing to clean up by caller
sub get_route(@) {
    my (@args) = @_;
    my $gated_pid;
    my $mrtd_sock;
    my ($destarg, $destination, $netmask, $gateway, $interface);
   
    XB_Log::log "info" => "-> XB_Route::get_route @args";

    my $get_route = 0;
    $get_route = eval {
	defined($gated_pid = check_gated) or XB_Log::log "err" => "check_gated failed" and die "check_gated";

	defined($mrtd_sock = check_mrtd(1)) or XB_Log::log "err" => "check_mrtd failed" and die "check_mrtd";

	($destarg, $destination, $netmask, $gateway, $interface) = parse_input("get", @args);
	
	if ($gated_pid) {
	    my $gated_conf = $XB_Defs::GATED_PATH . "gated.conf";
	    open (GATED, "$gated_conf") or 
		XB_Log::log "err" => "$!: Could not open $gated_conf" 
		    and die "open";
	    
	    while (defined(my $line = <GATED>)) {
		if ($line =~ m/^\s*static/) {
		    my %dests;
		    
		    # make sure destination isn't already configured
		    while (defined($line = <GATED>)) {
			if ($line =~ m/\}/) {
			    # found end of static clause 
			    if (exists $dests{$destination}) {
				close GATED;
				return 1;
			    } else {
				close GATED;
				return 0;
			    }
			} else {
			    # record next configured destination
			    my $wholeline;
			    my $currdest;
			    
			    $wholeline = $line;
			    $line =~ s/^\s//g; # chop leading whitespace
			    my @line = split(' ', $line);
			    
			    # record if host or net
			    if ($line[0] eq "host") {
				$currdest = $line[1];
			    } else {
				$currdest = $line[0];
			    }
			    
			    $dests{$currdest}++;
			}
		    } # while ($line = <GATED>) 
		    last;
		} # if ($line =~ m/^\s*static/) 
	    } # while (defined(my $line = <GATED>))
	    close GATED;
	    
	    return 0;
	} # if ($gated_pid)

	elsif ($mrtd_sock) {
	    mrtd_send($mrtd_sock, "config\r\n");
	    mrtd_recv($mrtd_sock, "Config>");
	    my $reply = "";
	    my $flag = 0;
	    my $route_flag = route_exists($mrtd_sock, $destination, 
				    $gateway, $interface);
            if (!mrtd_send($mrtd_sock, "exit\r\n")) {
              XB_Log::log "err", "Error sending command to exit config mode";
                die "mrtd_send";
            }
            ($flag, $reply) = mrtd_recv($mrtd_sock, "MRTd\#");
            if ($flag) {
              XB_Log::log "err", "Error reading from mrtd" and
                die "XB_Route::delete_route mrtd_recv";
            }
            if (!mrtd_send($mrtd_sock, "exit\r\n")) {
              XB_Log::log "err", "Error sending command to exit mrtd";
                die "XB_Route::delete_route mrtd_send";
            }
            $mrtd_sock->close();

	    if ($route_flag) {
		return 1;
	    } else {
		return 0;
	    }
	} # if ($mrtd_sock)

	else {
	  XB_Log::log "info" => "OS is $XB_Defs::NODEOS";
	    if ($XB_Defs::NODEOS =~ /linux|nist/i) {
		# Linux OS (No "route get")
		my $pipe = "route -n |";
		open ROUTE, $pipe or
		  XB_Log::log "err" => "$!: Could not open pipe $pipe" 
		      and die "route";
		my $line;
		while (defined($line = <ROUTE>)) {
		    my @line = split(" ", $line);
		    
		    if ($line[0] eq $destination) {
			if ($netmask) {  # neccessary? I doubt.
			    $netmask = inet_ntoa(inet_aton($netmask)); # maybe 0x????
			    if ($line[2] eq $netmask) { return 1; }
			    else { return 0; }
			} else { return 1; }
		    }  
		}      
		close ROUTE;
		return 0;	   
	    } else {
		# Default OS (FreeBSD)
		my $cpid; 
		
		if ((defined($netmask)) && ($netmask ne "")) {
		  XB_Log::log "debug" => "get_route(D1): route -n get $destarg $destination -netmask $netmask";
		    $cpid = open3(\*WTR, \*RDR, \*ERR, "route", "-n", "get", $destarg, 
				  $destination, "-netmask", $netmask) or
				      XB_Log::log "err" => "$!: Route command failed" and die "route";
		} else {
		  XB_Log::log "debug" => "get_route(D2): route -n get $destarg $destination";
		    $cpid = open3(\*WTR, \*RDR, \*ERR, "route", "-n", "get", $destarg, 
				  $destination) or
				      XB_Log::log "err" => "$!: Route command failed" and die "route";
		}
		
		if (defined(my $line = <ERR>)) {
		    close WTR or XB_Log::log "err" => "Close failed: $!" 
			and die "close";
		    close RDR or XB_Log::log "err" => "Close failed: $!"
			and die "close";
		    close ERR or XB_Log::log "err" => "Close failed: $!"
			and die "close";
		    
		    if (defined $cpid) {
			waitpid $cpid, 0 == $cpid or
			    XB_Log::log "err" => "No child $cpid" and die "wait";
		    }
		    return 0;
		} else {
		    while (defined(my $line = <RDR>)) {
			if ($line =~ m/route to:/) {
			    # get "route to:" line and extract IP
			    $line =~ s/\s$//g;
			    my @line = split(':', $line);
			    my $goal = $line[1];
			    $goal =~ s/^\s*//;
			    
			    # get "destination:" line and extract IP
			    $line = <RDR>;
			    $line =~ s/\s$//g;
			    @line = split(':', $line);
			    my $dest = $line[1];
			    $dest =~ s/^\s*//;
			    
			    close WTR;
			    close RDR;
			    close ERR;
			    
			    if (defined $cpid) {
				waitpid $cpid, 0 == $cpid or 
				    XB_Log::log "err" => "No child $cpid"
					and die "wait"
			    }
			    if ($goal eq $dest) {
				# explicit route defined
				return 1;
			    } else {
				# uses default route
				return 0;
			    }
			    
			    last;
			}
		    } #while (defined(my $line = <ROUTE>))
		}
	    
	    }	
	}
    };
    XB_Log::log "info" => "<- XB_Route::get_route @args"; 
    return $get_route unless $@;
    unless ($@ =~ /^(open|route|wait|close|check_gated|check_mrtd|mrtd_send|mrtd_recv)/){
	XB_Log::log "warning" => "XB_Route::get_route caught unexpected exception $@";
    }
    die "XB_Route::get_route";
} # sub get_route(@)


# Description: 
#	Removes all routes matching given prefix/netmask pairs
# Arguments:
#       @args -- list of prefix/netmask pairs
# Returns:
#	1 unless error
# Exceptions:
#	"XB_Route::flush_route" on error, command may finish incomplete with error
#       must check to see if all routes were deleted
# input is in list form, (prefix, netmask, [prefix, netmask, ...])
sub flush_route(@) {
    my (@args) = @_;
    my (@prefix, @netmask, @gated_conf);
    my (@routes);

    XB_Log::log "info" => "-> XB_Route::flush_route";
    
    eval {
	# test that input has an even number of args.
	# we must assume that they are in the correct order

	if (($#args + 1) % 2) {
	    XB_Log::log "err" => "incomplete input pairs" and die "bad_input";
	} 
	
	# make a list of Net::Netmask objects that match routes
	# to be deleted
	while (defined(my $dest = shift(@args))) {
	    my (@addr, $newaddr, $binaddr);
	    my $netmask = shift(@args);
	    
	    my $new_block = new Net::Netmask($dest, $netmask);
	    push @routes, $new_block;
	}	    

	my $gated_new = $XB_Defs::GATED_PATH. "gated.conf.new";
	my $gated_conf = $XB_Defs::GATED_PATH. "gated.conf";

	defined(my $gated_pid = check_gated) or 
	  XB_Log::log "err" => "check_gated failed" and die "check_gated";

	defined(my $mrtd_sock = check_mrtd(1)) or 
	  XB_Log::log "err" => "check_mrtd failed" and die "check_mrtd";

	if ($gated_pid) {
	    make_backup or XB_Log::log "err" => "could not backup gated.conf" 
		and die "make_backup";
	    open (INPUT, "$gated_conf") or XB_Log::log "err" => "Could not open $gated_conf: $!" and die "open";
	    @gated_conf = <INPUT>;
	    close INPUT or XB_Log::log "err" => "Could not close $gated_conf: $!"
		and die "close"
	    # INPUT = open gated.conf
	} elsif (!defined($mrtd_sock)) {
	    open (INPUT, "netstat -nr |") or 
		XB_Log::log "err" => "Couldn't netstat" and die "netstat";
	}
	
	my $destination;
	if ($gated_pid) {
	    my $i = 0;
	    my $static = 0;
	    my $lim = $#gated_conf;
	    while ($i < $lim) {
		if ($i > $lim) {last;}
		# find static section
		if ($gated_conf[$i] =~ /^(\s)*static \{/) {
		    $static++;
		  XB_Log::log "debug" => "Found static line";
		    $i++;
		} elsif (($static == 1) and 
		    ($gated_conf[$i] =~ /^(\s)*host/) and
		    ($gated_conf[$i] =~ /XBONE/)) {
		    # might be a line we added, check if should be flushed
		  XB_Log::log "debug" => "Found host line: $gated_conf[$i]";
		    my @line = split(" ", $gated_conf[$i]);
		    my $destination = $line[1];
		    
		    # host destinations
		    my $route;
		    foreach $route (@routes) {
			if ($route->match($destination)) {
			    splice(@gated_conf, $i, 1);
			    $lim--;
			}
		    }
		    # check against patterns
		    # if host and IP matches, splice out line
		} elsif (($static == 1) and
		    ($gated_conf[$i] =~ /\}/)) {
		    # done with static entries, can finish.
		    last;
		} else {
		    $i++;
		}
	    }
	} elsif ($mrtd_sock){

	    # mrtd code.
	    # enter config mode.
	    $mrtd_sock->send("config\r\n");
	    $mrtd_sock->send("show\r\n");
	    my ($flag, $reply) = mrtd_recv($mrtd_sock);
	    if ($flag) {
	      XB_Log::log "err", "Error reading from mrtd";
		die "XB_Route::flush_route mrtd";
	    }

	    # read configuration into a list.
	    my @config = split /\n/, $reply;
	    @config = grep /^route/, @config;

	    # go thru the list and find any route that you need to delete
	    # save any such route in another list
	    my @deleted_routes;
	    my ($route, $command);
	    
	    while (@config) {
		### verify the pattern match.
		my @line = split /\s+/, $config[0];
		$destination = $line[1];
		foreach $route (@routes) {
		    if ($route->match($destination)) {
			chomp($config[0]);
			push @deleted_routes, $config[0];
		    }
		}
		shift @config;
	    }
	    
	    # send saved routes one at a time.
	    foreach $command (@deleted_routes) {
		if (!mrtd_send($mrtd_sock, "no " . $command . "\r\n")) {
		  XB_Log::log "err", "Error sending route delete command";
		    die "mrtd_send";
		}
	    }

	    # close socket.
	    $mrtd_sock->send("exit\r\n");
	    $mrtd_sock->send("exit\r\n");
	    #$mrtd_sock->close();
	}
	else {  # neither gated nor mrtd running.
	  XB_Log::log "info" => "OS is $XB_Defs::NODEOS";
	  if ($XB_Defs::NODEOS =~ /linux|nist/i) {	
	     # Linux netstat command 
	     while (defined(my $line = <INPUT>)) {
		# need to check if first item is defined and 
		# an IP addr.
		my @line = split (" ", $line); 
		$destination = $line[0];
		
		# skip lines that don't list routes
		# skips "default" too
		unless (defined($destination)) {
		    next;
		}
		unless ($destination eq "0.0.0.0") {
		    next;
		}	
		if ($line[2] ne "255.255.255.255") {
		    # destination is a net destination
		    # make a new Net::Netmask block and compare 
		    # base and masks
		    my $dest_block = new Net::Netmask($destination);
		    my $route;
		    foreach $route (@routes) {
			if (($route->base() eq $dest_block->base()) and
			    ($route->mask() eq $dest_block->mask())) {
                          XB_Log::log "debug" => "flush_route(D1): route del -net $dest_block->base() netmask $dest_block->mask()";
			    system("route", "del", "-net", $dest_block->base(),
				   "netmask", $dest_block->mask()) == 0 or
				     XB_Log::log "err" => "Could not delete $destination"
					 and die "route";
			    last;
			}
		    }
		} else {
		    # host destinations
		    my $route;
		    foreach $route (@routes) {
			if ($route->match($destination)) {
                          XB_Log::log "debug" => "flush_route(D2): route del -host $destination";
			    system("route", "del", "-host", $destination) == 0 or
			      XB_Log::log "err" => "Could not delete $destination"
				  and die "route";
			    last;
			}
		    }
		}
	     } # while		
          } else {  # not Linux 			
	      while (defined(my $line = <INPUT>)) {
		# need to check if first item is defined and 
		# an IP addr.
		my @line = split (" ", $line); 
		$destination = $line[0];
		
		# skip lines that don't list routes
		# skips "default" too
		unless (defined($destination)) {
		    next;
		}
		unless ($destination =~ /^\d/) {
		    next;
		}
		
		if ($destination =~ /\//) {
		    # destination is a net destination
		    # make a new Net::Netmask block and compare 
		    # base and masks
		    my $dest_block = new Net::Netmask($destination);
		    my $route;
		    foreach $route (@routes) {
			if (($route->base() eq $dest_block->base()) and
			    ($route->mask() eq $dest_block->mask())) {
			  XB_Log::log "debug" => "flush_route(D1): route delete $dest_block->base() -netmask $dest_block->mask()";
			    system("route", "delete", $dest_block->base(),
				   "-netmask", $dest_block->mask()) == 0 or
				     XB_Log::log "err" => "Could not delete $destination"
					 and die "route";
			    last;
			}
		    }
		} else {
		    # host destinations
		    my $route;
		    foreach $route (@routes) {
			if ($route->match($destination)) {
			  XB_Log::log "debug" => "flush_route(D2): route delete $destination";
			    system("route", "delete", $destination) == 0 or
			      XB_Log::log "err" => "Could not delete $destination"
				  and die "route";
			    last;
			}
		    }
		}
	    } #  while (defined(my $line = <INPUT>)) 
	  }  # not Linux
	} # gated not running
	
	if ($gated_pid) {
	    open (NEW, ">$gated_new") 
		or XB_Log::log "err" => "Could not open $gated_new: $!"
		    and die "open";
	    foreach (@gated_conf) {
		print NEW;
	    }
	    close NEW or XB_Log::log "err" => "Could not close $gated_new: $!" 
		and die "close";
	    
	    # mv gated.conf.new gated.conf
	    move("$gated_new", "$gated_conf") or
	      XB_Log::log "err" => "Move failed: $!" and die "move";
	    
	    clean_gated or XB_Log::log "err" => "Could not clean_gated"
		and die "clean_gated";
	    restart_gated or XB_Log::log "err" => "Could not restart_gated"
		and die "restart_gated";
	    # remove local backup
	    my $gated_bak = $XB_Defs::GATED_PATH . "gated.conf.bak";
	    unlink $gated_new;
	} # if ($gated_pid)
    }; # eval {
    XB_Log::log "info" => "<- XB_Route::flush_route @args";
    return 1 unless $@;
    if ($@ =~ /^(bad_input|make_backup|check_gated|netstat|route|check_mrtd|mrtd_send|mrtd_recv)/){
	# nothing to cleanup
    } elsif ($@ =~ /^(open|close)/) {
	# remove local backup|
	eval {
	    my $gated_bak = $XB_Defs::GATED_PATH . "gated.conf.bak";
	    unlink $gated_bak;
	};
    } elsif ($@ =~ /^(move|clean_gated|restart_gated)/) {
	# must restore gated.conf
	eval {
	    my $gated_new = $XB_Defs::GATED_PATH . "gated.conf.new";
	    unlink $gated_new;
	    my $gated_bak = $XB_Defs::GATED_PATH . "gated.conf.bak";
	    my $gated_conf = $XB_Defs::GATED_PATH . "gated.conf";
	    move ("$gated_bak", "$gated_conf");
	};
    } else {
	XB_Log::log "warning" => "XB_Route::flush_route caught unexpected exceptions $@";
    }
    die"XB_Route::flush_route";
} # sub flush_route(@)

###########################################################################
# flush_xbone_route
#
# Description:
#	Flush ALL the xbone-related routes in the blocks defined by
#	$XB_Defs::XB_IP_EXTERNAL_BLOCK, $XB_Defs::XB_IP_INTERNAL_BLOCK,
#	with both class C netmask & $XB_Defs::XB_OVL_SUBNET_BITS.
# Arguments:
#	None
# Constants:
#	$XB_Defs::XB_IP_EXTERNAL_BLOCK, $XB_Defs::XB_IP_INTERNAL_BLOCK
#	$XB_Defs::XB_OVL_SUBNET_BITS
# Return:
#	1 on success
# Limits:
#	Assume IPv4 & class C overlay addresses.
###########################################################################
sub flush_xbone_route()
{
# needs exception handling

    my $eblock = new Net::Netmask($XB_Defs::XB_IP_EXTERNAL_BLOCK);
    my $iblock = new Net::Netmask($XB_Defs::XB_IP_INTERNAL_BLOCK);
    my $subnet_block= $XB_Defs::XB_IP_EXTERNAL_BLOCK;
    my $subnet_bits=$XB_Defs::XB_OVL_SUBNET_BITS;
    my $subnet_mask_bits = $subnet_bits + $eblock->bits();
    my $host_bits = 32 - $subnet_mask_bits;  # Not IPv6-aware!

    $subnet_block =~ s/(\d+\.\d+\.\d+\.\d+)\/\d+/$1\/$subnet_mask_bits/;
    my $sblock = new Net::Netmask($subnet_block);

    push my @net_mask_pairs, $eblock->base(), $eblock->mask(),
	    $iblock->base(), $iblock->mask();

    foreach my $base_ip ($eblock->base(),$iblock->base()) {
        for (my $i=0; $i<(2**$subnet_bits); $i++) {
	    my $forth_byte = $i << $host_bits;
            $base_ip =~ s/(\d+\.\d+\.\d+\.)\d+/$1$forth_byte/;
	    push @net_mask_pairs, $base_ip, $sblock->mask(); 
        }
    }
    flush_route(@net_mask_pairs);

    return 1;
} # sub flush_xbone_route(@)




1;














































