# -*- perl -*- 
#
# DO NOT MOVE THE FIRST LINE
# It identifies the rest of the file as PERL for EMACS autoformatting
# put perl options at the end of that line, e.g., -p
#
# ----------------------------------------------------------------------------
#				    X-BONE
#
#		 USC Information Sciences Institute (USC/ISI)
#		    Marina del Rey, California 90292, USA
#			   Copyright (c) 1998-2000
# ----------------------------------------------------------------------------
#
# Copyright (c) 1998-2000 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
#
# -------------------------------------------------------------------------
#
# Effort sponsored by the Defense Advanced Research Projects Agency (DARPA)
# and Air Force Research Laboratory, Air Force Materiel Command, USAF, under
# agreement number F30602-98-1-0200. The U.S. Government is authorized to 
# reproduce and distribute reprints for Governmental purposes not withstanding
# any copyright annotation therein.
# 
# The views and conclusions contained herein are those of the authors and
# should not be interpreted as necessarily representing the official policies
# or endorsements, either expressed or implied, of the Defense Advanced
# Research Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
# 
# -------------------------------------------------------------------------
# 
#
# ----------------------------------------------------------------------------
#  $RCSfile: XB_Tunnel.pm,v $
# 
# $Revision: 1.28 $
#   $Author: touch $
#     $Date: 2000/10/24 17:28:36 $
#    $State: Exp $
# ----------------------------------------------------------------------------
#
# Primary Author: Lars Eggert

package XB_Tunnel;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(up down init reset);

use strict;
use sigtrap;

use POSIX;
use FindBin;
use Socket; #inet_aton

use XB_Defs;
use XB_Route;
use XB_Log;
use XB_Utils;


###############################################################################
# UTILITY FUNCTIONS
###############################################################################

# Description: 
#	Return interface type of passed interface name.
# Arguments:
#	$name	name of interface (from ifconfig or similar)
# Returns:
#	type of interface (strip interface number)
#	undef if string did not looks like an interface
# Exceptions:
#	None.
sub if_type ($) {
  my $if = shift;
  if($if =~ /^(\D+)\d+/) { return $1; }
  return undef;
}  


# Description: 
#	Try to demand-create new interface of type $type.
# Arguments:
#	$type	interface type to create a new instance of
# Returns:
#	1 on success
#	undef on failure to create (but no critical error)
# Exceptions:
#	"XB_Tunnel::add" on error, nothing to clean up by caller
sub add ($) {
  my $if_type = shift;
  my $created = undef;
  # print trace line
  XB_Log::log "info", "-> XB_Tunnel::add $if_type";
  eval {
    if($if_type eq "tunl" and $XB_Defs::NODEOS =~ /linux|nist/i) {
      # only install ipip if not there yet
      my $pipe = "lsmod |";
      # pipe to lsmod
      open PIPE, $pipe or XB_Log::log "err", "cannot open pipe $pipe: $!" 
	and die "open";
      # see if ipip module already there
      my $found = 0;
      while(<PIPE>) { if(/ipip/) { $found = 1; last; } }
      # done with pipe
      close PIPE 
	#or not $! 
	or XB_Log::log "err", "cannot close pipe $pipe: $!" 
	  and die "close";
      # if we didn't find it, load ipip
      unless($found) {
	# no ipip loaded yet, load it
	my @cmd = ("insmod", "ipip");
	XB_Log::log "debug3", "@cmd";
	system(@cmd) == 0 or XB_Log::log "err", "@cmd failed: $!" 
	  and die "insmod";
      }  	
      # add a new tunnel (arbitrarily assign remote and local to make it
      # P-to-P; change them later)
      my @cmd = ("ip", "tunnel", "add", "mode", "ipip", "remote",
		 "192.168.0.1", "local", "192.168.0.1", "ttl", "30");
      XB_Log::log "debug3", "@cmd";
      system(@cmd) == 0 or XB_Log::log "err", "@cmd failed: $!" and die "ip";
      $created = 1;
    } elsif($if_type eq "iti"and $XB_Defs::NODEOS =~ /cairn/i) {
      # create a new iti if via sysctl
      my $sysctl_var = "net.inet.ip.iti_count";
      my $msg = XB_Utils::sysctl_read $sysctl_var;
      if(defined $msg and $msg ne "") { 
	XB_Utils::sysctl_write $sysctl_var, $msg+1;
	$created = 1;
      }
    }
  };
  # print trace line
  XB_Log::log "info", "<- XB_Tunnel::add $if_type";
  return $created unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(open|close|insmod|ip|sysctl_(read|write))/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::add: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_Tunnel::add";
  }
  

# Description: 
#	Checks if an interface is owned by xbone, by looking if configured 
#	addresses lie in xbone networks.
# Arguments:
#	$if	interface to check
# Returns:
#	echoes argument ($if) if at least one of the interface's addresses 
#	lies in an xbone net
#	undef if the interface does not belong to xbone
# Exceptions:
#	"XB_Tunnel::is_xbone_if" on error, nothing to clean up by caller
sub is_xbone_if ($) {
  my $if = shift;
  # print trace line
  XB_Log::log "info", "-> XB_Tunnel::is_xbone_if $if";
  my $result = undef;
  eval {
    # untaint $if
    if($if =~ /^(\D+\d+(:\d+)?)/) { $if = $1; }
    else { die "ifformat"; }
    # interface belongs to xbone is either of its configure addresses lies
    # within one of the xbone address spaces, check by looking at ifconfig
    # (xbone interfaces are also ones which have no IP set at all)
    my $pipe = "ifconfig $if |";
    open PIPE, $pipe or XB_Log::log "err", "cannot open pipe $pipe: $!"
      and die "open";
    # look for assigned virtual addresses
    while(<PIPE>) { 
      if(/((\d{1,3}\.){3}\d{1,3})\s+-->\s+((\d{1,3}\.){3}\d{1,3})/g or
	 /\s+inet\saddr:((\d{1,3}\.){3}\d{1,3})\s+
	 P-t-P:((\d{1,3}\.){3}\d{1,3})/gx) {
	# found interface addresses, do check
	$result |= XB_Utils::XB_is_xbone_address $1 | 
	  XB_Utils::XB_is_xbone_address $3;
      }
    }
    # done with pipe
    close PIPE 
      #or not $! 
      or XB_Log::log "err", "cannot close pipe $pipe: $!"
      and die "close";
  };
  # print trace line
  XB_Log::log "info", "<- XB_Tunnel::is_xbone_if $if";
  return ($result ? $if : undef) unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(open|close|ifformat)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::is_xbone_if: caught unexpected " .
      "exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_Tunnel::is_xbone_if";
}


# Description: 
#	Return a list of used or unused interfaces of a certain type.
# Arguments:
#	$used	1 to return used ( = configured or up), 0 to look for unused
# Returns:
#	array of interace names
# Exceptions:
#	"XB_Tunnel::if_list" on error, nothing to clean up by caller
sub if_list ($) {
  my $used = shift;
  my @ifs;
  my $if="";
  my $configured;
  my $pipe;

  eval {
    if($XB_Defs::NODEOS =~ /freebsd|cairn|kame/i) {
      # CAIRN/KAME way of doing it
      $pipe = sprintf "ifconfig -l%s |", ($used ? "u" : "d");
      # look at ifconfig output
      open PIPE, $pipe or XB_Log::log "err", "cannnot open pipe $pipe: $!"
	and die "open";
      # grab results
      $if = <PIPE>;
      # done with pipe
      close PIPE 
	#or not $! 
	or XB_Log::log "err", "cannot close pipe $pipe: $!"
	and die "close";
      @ifs =  split " ", $if;
    } else {
      # Linux way of doing it
      $pipe = "ifconfig -a |";
      # look at ifconfig output
      open PIPE, $pipe or XB_Log::log "err", "cannnot open pipe $pipe: $!"
        and die "open";
      while(<PIPE>) {
        # if a new interface follows in the next lines
        if(/^(\w+\d+(:\d+)?)\s+/) {
          my $new_if = $1;
	  # remember previous interface (it's one we're looking for)
	  if($if and $configured == $used and
	     # on Linux, tunl0 cannot be made pointopoint, so skip it
	     ($XB_Defs::NODEOS !~ /linux|nist/i or $if ne "tunl0")) { 
	    push @ifs, $if; 
	  }
          # reset state for scanning the next interface
          $configured = 0;
          $if = $new_if;
        } elsif(/\s+inet\s(addr:)?
		(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})/x) {
	  # if current interface has an inet address, mark configured
	  if($2 or $3 or $4 or $5) { $configured = 1; }
	}
      }
      if($if and $configured == $used and $if ne "tunl0") { push @ifs, $if; }
      close PIPE
        or XB_Log::log "err", "cannot close pipe $pipe: $!"
          and die "close";
    }
  };
  return @ifs unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(open|close)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::if_list: caught unexpected " .
      "exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_Tunnel::if_list";
}


# Description: 
#	Return a list of used or unused interfaces of a certain type.
# Arguments:
#	$type	type of interface to check for
#	$used	1 to return used ( = configured or up), 0 to look for unused
#	$all	1 to return xbone and non-xbone ifs (optional: default :0)
# Returns:
#	array of interace names
# Exceptions:
#	"XB_Tunnel::grab" on error, nothing to clean up by caller
sub grab ($$;$$) {
  my ($type, $used, $all, $recursion) = @_;
  my @ifs;
  my $trace_str = "XB_Tunnel::grab $type, $used";
  if(defined $all) { $trace_str .= ", $all"; }
  elsif(defined $recursion) { $trace_str .= ", undef, $recursion"; }
  XB_Log::log "info", "-> $trace_str";
  eval {
    # get a list of used (up) or unused (down) interfaces
    # process output and remove interfaces of types we don't want
    @ifs = grep /$type\d+/, if_list $used;
    # get rid of unused interface with active alias (for linux tunl)
    # linux can have tunlx with no IP address but still UP for it's aliases
    # tunlx:y, tunlx:z, etc.
    if(not $used){
      # get a list of active interfaces & aliases of the given type
      my @used_ifs = grep /$type\d+/, if_list 1;   
      my @new_list;
      foreach my $i (@ifs){
        my @dup = grep /$i:\d+/, @used_ifs;
        if((not defined $dup[0]) || ($dup[0] eq "")){
          # no aliases found, save to use it
          push @new_list, $i;    
        }
      }
      @ifs = @new_list;
    }
    # post-process results
    if(not $used and $#ifs == -1 and not defined $recursion) {
      # if we're looking for an unused interface, try creating one if we could
      # not find one and we're not currently in the retry phase
      add $type;
      @ifs = grab($type, $used, $all, "retry after add");
    } elsif($used and $#ifs != -1 and not $all) {
      # looking for used interfaces, make sure we only return xbone ones
      # map { my $u = $_; my $r = is_xbone_if $_; $_ = $u; return $r; } @ifs;
      # I think there's a bug in the "map" block above. It only checks the 
      # first one, returns & exits eval. => Rewrite with the stupid approach.
      my @new_ifs;
      foreach my $j (@ifs){
	if( defined is_xbone_if $j ) { # keep only xbone interfaces
	  push @new_ifs, $j;
	}
      }
      @ifs = @new_ifs;
    }
  };
  # print trace line
  XB_Log::log "info", "<- $trace_str";
  return @ifs unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(open|close|XB_Tunnel::if_list)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::grab: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_Tunnel::grab";
}


###############################################################################
# CAIRN-2.5 ITI TUNNELS / KAME GIF TUNNELS
###############################################################################

# flag set by init indicating if a KAME host supports "gifconfig delete",
# ca. 7/2000 and later
my $have_gifconfig_delete = 0;

# Description: 
#	Wrapper for iticonfig/gifconfig.
# Arguments:
#	$if	interface to configure
#	$l	local address (outer header source)
#	$r	remote address (outer header destination)		    
# Returns:
#	nothing
# Exceptions:
#	"XB_Tunnel::iti_or_gif_config" on error, nothing to clean up by caller
sub iti_or_gif_config ($$$) {
  my ($if, $l, $r) = @_;
  XB_Log::log "info", "-> XB_Tunnel::iti_or_gif_config $if, $l $r";
  my @cmd = (if_type $if eq "iti" ? "iticonfig" : "gifconfig", $if, $l, $r);
  eval {
    XB_Log::log "debug3", "@cmd";
    my $cpid = open PIPE, "-|";
    if($cpid == 0) { 
      exec @cmd or XB_Log::log "err", "@cmd failed: $!" and exit; 
    }
    # if we don't read the output, we cannot close the pipe
    my @output = <PIPE>;
    close PIPE 
      #or not $! 
      or XB_Log::log "err", "cannot close pipe: $!" and die "close";
    # reap zombie
    waitpid $cpid, 0 == $cpid or XB_Log::log "err", "waitpid $cpid, 0: $!"
      and die "waitpid";
  };
  XB_Log::log "info", "<- XB_Tunnel::iti_or_gif_config $if, $l $r";
  return $if unless $@; # success if no exception
  # exception handling
  if($@ =~ /^close/) {
    # command failed, don't know how to undo (XXX restore state from before)
    XB_Log::log "warning", "XB_Tunnel::iti_or_gif_config: @cmd failed, but ".
      "cannot rollback to previous state";
  } else {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::iti_or_gif_config: caught unexpected " .
      "exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_Tunnel::iti_or_gif_config";
}


# Description:
#	Check if a configured gif with the same outer (local, remote) address
#	pair as ($l, $r) exists.
# Arguments:
#       $l      local address
#       $r      remote address
# Returns:
#       tunnel interface handle of the matched gif interface; or    
#       undef if none matched
# Exceptions:
#       "XB_Tunnel::gif_find_outer" on error, nothing to clean up by caller
sub gif_find_outer ($$) {
  my ($l, $r) = @_;
  XB_Log::log "info", "-> XB_Tunnel::gif_find_outer $l, $r";
  my $result = undef;
  my $p = "physical";
  my $a = "address";
  my $n = "inet";
  my @gifs = (grab("gif", 1, 1), grab("gif", 0, 1)); # grab all gif interfaces
  eval {
    foreach my $if (@gifs) {
      my $pipe = "gifconfig $if |";
      open GPIPE, $pipe or XB_Log::log "err", "cannont open pipe $pipe: $!"
	and die "open";
      # look for assigned "physical" ("outer" is more appropriate) addresses
      while(<GPIPE>) {
        if(/\s+$p\s$a\s$n\s((\d+\.){3}\d+)\s-->\s((\d+\.){3}\d+)/) {
	  # check if end points match $l & $r
	  if( ($1 eq $l) && ($3 eq $r) ){ 
	    if( defined is_xbone_if $if ) {
	      # has xbone alias, save to use it
	      $result = $if;
	      last;
	    } elsif ($XB_Defs::SHARE_TUN_IF) {
	      # has non-xbone alias, but we share gifs with others, so use it
	      $result = $if;
	      XB_Log::log "warning", "NON-XBONE tunnels exist between $l->$r. ".
		"$if is used by other applications and could be deleted ".
		"without warning!";
	      last;
	    } else {
	      # has non-xbone alias & not shared, die
	      XB_Log::log "err", "NON-XBONE tunnels exist between $l->$r" 
		and die "[gif/tunl sharing violation: non-xbone tunnels found,"
		." set XB_Defs::SHARE_TUN_IF=1 to enable gif/tunl sharing]";
	    }
          } #<if
	} #<if
      } #<while
      close GPIPE or XB_Log::log "err", "cannot close pipe $pipe: $!"
        and die "close";
    } #<foreach 
  }; #<eval
  XB_Log::log "info", "<- XB_Tunnel::gif_find_outer $l $r";
  return $result unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(open|close)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::gif_find_outer: caught unexpected " .
      "exception $@";
  } #<unless
  die "XB_Tunnel::gif_find_outer";
}


# Description: 
#	Tear down iti/gif tunnel $if.
# Arguments:
#	$if	interface to tear down
# Returns:
#	1 on success
# Exceptions:
#	"XB_Tunnel::iti_or_gif_down" on error, nothing to clean up by caller
sub iti_or_gif_down ($) {
  my $if = shift;
  XB_Log::log "info", "-> XB_Tunnel::iti_or_gif_down $if";
  if( not $XB_Defs::TUN_ALIASING ){
    # do not touch the original code if we don't use gif aliasing, the format
    # of the first argument is different
    eval {
      unless(if_type($if) =~ /^(iti|gif)/) { 
        XB_Log::log "err", "incompatible tag $if" and die "tag";
      }
      if(if_type($if) =~ /^gif/ and $have_gifconfig_delete) {
        # delete the gif interface
        my @cmd = ("gifconfig", $if, "delete");
        XB_Log::log "debug3", "@cmd";
        system(@cmd) == 0 or XB_Log::log "err", "@cmd failed: $!" 
	  and die "gifconfig";
      } else {
        # XXX This should do the equivalent of "gifconfig down delete",
        # but neither gifconfig nor iticonfig support that (gifconfig just
        # started to though). Instead, we configure the outer headers to
        # 255.255.255.255 -> 255.255.255.255 hoping this will be equivalent
        # (FreeBSD-4.0 does not like the address 0.0.0.0, which we'd
        # like to use here.)
        #iti_or_gif_config $if, "255.255.255.255", "255.255.255.255";
	# read comments below
      }
      # the old syntax "ifconfig ifX down delete" no longer works in 3.4
      foreach my $cmd (qw(down delete)) {
        XB_Route::config_route "ifconfig", $if, $cmd;
      }
    };
  } else { # block for TUN_ALIASING ==========================================
    eval {
      my $lv;
      my $rv;
      my $found = 0;
      if( $if =~ /^(\D+\d+)%%((\d+\.){3}\d+)-->((\d+\.){3}\d+)/ ) {
        $if = $1;  $lv = $2;  $rv = $4;
      } else {
        XB_Log::log "err", "unknown tag format $if" and die "tag";
      }
      unless(if_type($if) =~ /^(iti|gif)/) {
        XB_Log::log "err", "incompatible tag $if" and die "tag";
      }
      my $alias_count = 0;
      my $result = 0;
      my $pipe = "ifconfig $if |";
      open PIPE, $pipe or XB_Log::log "err", "cannot open pipe $pipe: $!"
        and die "open";
      # look for assigned virtual addresses 
      while(<PIPE>) {
        if(/((\d{1,3}\.){3}\d{1,3})\s+-->\s+((\d{1,3}\.){3}\d{1,3})/g or
           /\s+inet\saddr:((\d{1,3}\.){3}\d{1,3})\s+
           P-t-P:((\d{1,3}\.){3}\d{1,3})/gx) {
          # found interface addresses, do check
	  # $alias_count = no. of remaining alias, 
	  # $result = true => still have xbone alias
	  if( ($1 eq $lv) && ($3 eq $rv) ){
	    $found = 1;		# found the alias we want to delete
				# don't count this one
          } else {
            $alias_count++;
            $result |= XB_Utils::XB_is_xbone_address $1 |
                       XB_Utils::XB_is_xbone_address $3;
          }
          if ( $alias_count==0 && $result ){
            XB_Log::log "err", "conflict alias count" and die "impossible";
          }
        } #<if
      } #<while
      # done with pipe
      close PIPE or XB_Log::log "err", "cannot close pipe $pipe: $!"
        and die "close";
      # delete the given alias with "ifconfig delete" first, don't do ifconfig
      # down and gifconfig delete yet, need to know if it's the last one
      if( $found ){
	# found the alias, safe to delete
        XB_Route::config_route "ifconfig", $if, $lv, $rv, "delete";
      } else {
	# didn't find the alias, warn and proceed
	XB_Log::log "warning", "couldn't find $lv->$rv to delete, either ".
			    "doesn't exist or is doing crash recovery";
      }
      if (($alias_count == 0) or (not $result and not $XB_Defs::SHARE_TUN_IF)){
	# delete & down if (1) no other alias exists (2) no other xbone alias
	# exists and we aren't sharing gifs with non-xbone applications
        if(if_type($if) =~ /^gif/ and $have_gifconfig_delete) {
          # delete the gif interface
          my @cmd = ("gifconfig", $if, "delete");
          XB_Log::log "debug3", "@cmd";
          system(@cmd) == 0 or XB_Log::log "err", "@cmd failed: $!"
            and die "gifconfig";
        } else {   
          # can NOT do this since it trips over the "multiple tunnnel btn two
	  # points" bug! Arrrgh! This really sucks. upgrade to 4.1 if you can,
	  # it has "gifconfig delete".
          # iti_or_gif_config $if, "255.255.255.255", "255.255.255.255";
        }
	# don't do "ifconfig delete" here, it doesn't exist and would cause
	# "ioctl (SIOCDIFADDR): Can't assign requested address" error
        # bring down the interface 
        XB_Route::config_route "ifconfig", $if, "down";
      } 
    }
  }
  # print trace line
  XB_Log::log "info", "<- XB_Tunnel::iti_or_gif_down $if";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(tag|gifconfig|XB_Tunnel::iti_or_gif_config|XB_Route::config_route)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::iti_or_gif_down: caught unexpected " .
      "exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_Tunnel::iti_or_gif_down";
}
		     

# Description: 
#	Configure iti/gif tunnel on interface $if.
# Arguments:
#	$if	interface to configure
#	$lv	local virtual address (inner header source)
#	$rv	remote virtual address (inner header destination)
#	$nm	netmask for tunnel (on inner addresses)
#	$l	local address (outer header source)
#	$r	remote address (outer header destination)		    
#	$alias	1 to add as an alias (optional: default :0)
# Returns:
#	$if	on success (tunnel handle)
# Exceptions:
#	"XB_Tunnel::iti_or_gif_up" on error, nothing to clean up by caller
sub iti_or_gif_up ($$$$$$;$) {
  my ($if, $lv, $rv, $nm, $l, $r, $alias) = @_;
  my $trace_str = "XB_Tunnel::iti_or_gif_up $if, $lv, $rv, $nm, $l, $r";
  # $alias could only be set if $XB_Defs::TUN_ALIASING=1
  # did not completely separate the two because we use the old way if no
  # aliasing is needed
  if(defined $alias) { $trace_str .= ", $alias"; }
  else { $alias = 0; }
  XB_Log::log "info", "-> $trace_str";
  eval {
    if(not $alias) {
      XB_Route::config_route "ifconfig", $if, $lv, $rv, "netmask", $nm, "up";
      iti_or_gif_config $if, $l, $r
    } elsif ( $alias == 1 ) {
      # if add as an alias, only need to do ifconfig w/ "alias", not gifconfig
      # since outer header is already there
      XB_Route::config_route "ifconfig", $if, $lv, $rv, "netmask", $nm,
			     "alias", "up";
    } else {
      XB_Log::log "err", "unknown value: \$alias=$alias" 
	and die "unknown \$alias value";
    }
  };
  # print trace line
  XB_Log::log "info", "<- $trace_str";
  # attach $lv-->$rv to the tag so we know which one to delete when down
  if( $XB_Defs::TUN_ALIASING == 1) { $if .= "%%$lv-->$rv"; }
  return $if unless $@; # success if no exception
  # exception handling
  if($@ =~ /^(XB_Tunnel::iti_or_gif_config|XB_Route::config_route)/) {
    # undo iti/gifconfig
    eval { iti_or_gif_down $if; }
  } else {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::iti_or_gif_up: caught unexpected " .
      "exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_Tunnel::iti_or_gif_up";
}


###############################################################################
# LINUX TUNL TUNNELS
###############################################################################

# atomicity helper undoing ifconfig
sub tunl_undo_ifconfig ($) {
  my $if = shift;
  XB_Log::log "info", "-> XB_Tunnel::tunl_undo_ifconfig $if";
  eval {
    unless(if_type($if) =~ /^tunl/) { 
      XB_Log::log "err", "incompatible tag $if" and die "tag";
    }
    my @cmd = ("ifconfig", $if, "0.0.0.0", "down");   
    XB_Route::config_route(@cmd);
  };
  # print trace line
  XB_Log::log "info", "<- XB_Tunnel::tunl_undo_ifconfig $if";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^tag/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::tunl_undo_ifconfig: caught " .
      "unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_Tunnel::tunl_undo_ifconfig";
}


# Description:
#       Check if a configured tunl with the same outer (local, remote) address
#       pair as ($l, $r) exists.
# Arguments:
#       $l      local address
#       $r      remote address
# Returns:
#       tunnel interface handle of the matched tunl interface; or
#       undef if none matched
# Exceptions:
#       "XB_Tunnel::tunl_find_outer" on error, nothing to clean up by caller
sub tunl_find_outer ($$) {
  my ($l, $r) = @_;
  XB_Log::log "info", "-> XB_Tunnel::tunl_find_outer $l, $r";
  my $err_msg = "[gif/tunl sharing violation: non-xbone tunnels found,"
		  ." set XB_Defs::SHARE_TUN_IF=1 to enable gif/tunl sharing]";
  my $result = undef;
  my $match = undef;
  my $has_xbone_alias=0;
  my $has_non_xbone_alias=0;
  my @uuifs;
  my @uifs;
  my $m = "remote";
  my $o = "local";
  eval {
    my $pipe = "ip tunnel show |";
    # open a pipe to get the outer addresses of existing tunnels
    open LPIPE, $pipe or XB_Log::log "err", "cannot open pipe $pipe: $!"
      and die "open";
    while (<LPIPE>) {
      if(/^(\D+\d+):\s+ip\/ip\s+$m\s+((\d+\.){3}\d+)\s+$o\s+((\d+\.){3}\d+)/){
	if(($4 eq $l) && ($2 eq $r)){
	  $match = $1;
	  last;
        }
      }
    }
    close LPIPE or XB_Log::log "err", "cannot close pipe $pipe: $!"
      and die "close";
    # do the following only if we found a match (tunlx)
    if(defined $match){
      # search the unused ones since the primary interface tunlx might be here
      @uuifs = grep /^$match$/, if_list 0;
      # search the used ones to determine the first available alais tunlx:y
      @uifs = grep /^$match(:\d+)?$/, if_list 1;
      # determine the first available alias name
      if(defined $uuifs[0]) { 
	$result = $uuifs[0];
      }else{
	my $alias_no = 0;
	foreach my $i (@uifs){
	  if($i =~ /$match:(\d+)/){
	    if($1 == $alias_no){
	      $alias_no++;
	    }else{
	      $result = $match.":$alias_no";
	      last;
	    }
	  }
        }
	if(! defined $result) { $result = $match.":$alias_no"; }
      } 
      # check if any xbone alias
      foreach my $i (@uifs){
        if(defined is_xbone_if $i){
          $has_xbone_alias = 1;
        }else{
	  $has_non_xbone_alias = 1;
        }
      }
      # die if non-xbone alias exists && no sharing of tun interfaces
      if(! $has_xbone_alias && $has_non_xbone_alias &&  
	 ! $XB_Defs::SHARE_TUN_IF){
        XB_Log::log "err", "Non-XBone tunnel alias exist between $l->$r"
	  and die "$err_msg";
      }
    } # if(defined $match)
  }; #<eval
  XB_Log::log "info", "<- XB_Tunnel::tunl_find_outer $l, $r";
  return $result unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(open|close)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::tunl_find_outer: caught unexpected " .
      "exception $@";
  } #<unless
  die "XB_Tunnel::tunl_find_outer";   
}

# Description:
#       Configure a tunl type on $if
# Arguments:
#       $if     interface to configure
#       $lv     local virtual address (inner header source)
#       $rv     remote virtual address (inner header destination)
#       $nm     netmask for tunnel (on inner addresses)
#       $l      local address (outer header source)
#       $r      remote address (outer header destination)
#       $alias  1 to add as an alias (optional: default :0)
# Returns:
#       $if     on success (tunnel handle)
# Exceptions:
#       "XB_Tunnel::tunl_up" on error, nothing to clean up by caller
sub tunl_up ($$$$$$;$) {
  my ($if, $lv, $rv, $nm, $l, $r, $alias) = @_;
  my $trace_str = "XB_Tunnel::tunl_up $if, $lv, $rv, $nm, $l, $r";
  # $alias could only be set if $XB_Defs::TUN_ALIASING=1
  # did not completely separate the two because we use the old way if no
  # aliasing is needed
  if(defined $alias) { $trace_str .= ", $alias"; }
  else { $alias = 0; }
  XB_Log::log "info", "-> $trace_str";
  eval {
    # ip tunnel change (the remote and local we arbitrarily chose)
    my $pif=$if;
    my @cmd = ("ip", "tunnel", "change", $if, "mode", "ipip", "remote",$r,
	 "local", $l);
    if(! $alias){
      # don't do this if it's an alias since it's outer header is already up
      XB_Log::log "debug3", "@cmd";
      system(@cmd) == 0 or XB_Log::log "err", "@cmd failed: $!" and die "ip";
    }
    # ifconfig
    @cmd = ("ifconfig", $if, $lv, "pointopoint", $rv, "netmask", $nm, "up");
    XB_Route::config_route(@cmd);
    # route command must be used on the primary interface name, not alias
    if($if =~ /^(\D+\d+)(:\d+)/){	# get rid of trailing ":\d+" for alias
      $pif = $1;
    }
    # When gated is not running, we need to issue route commands
    # Linux-2.2: ifconfig will automatically add a net route if netmask exists
    #            This route will collide with each other; replace it
    if(!XB_Route::check_gated()) {  # gated not running
      if($nm ne "255.255.255.255" and $nm ne "0xffffffff"){ # a net route added
        my $subnet = inet_ntoa(inet_aton($rv) & inet_aton($nm));
        XB_Route::delete_route("-net", $subnet, "-netmask", $nm)
          or XB_Log::log "err", "can't delete route!" 
            and die "route";
        XB_Route::add_route($rv, "-interface", $if)
          or XB_Log::log "err", "can't add route!"
            and die "route";
      }
    }
  };
  # print trace line
  XB_Log::log "info", "<- $trace_str";
  return $if unless $@; # success if no exception
  # exception handling
  if($@ =~ /^(route|XB_Route::config_route)/) {
    # undo ifconfig
    eval { tunl_undo_ifconfig $if; }
  }
  else 
  {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::tunl_up: caught unexpected " .
      "exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_Tunnel::tunl_up";
}


# tear down a tunl-type tunnel
sub tunl_down ($) {
  my $if = shift;
  my $rv;
  # extrace the primary interface name (without :\d+)
  my $pif = ($if =~ /^(\D+\d+):\d+/ ? $1 : $if);
  XB_Log::log "info", "-> XB_Tunnel::tunl_down $if";
  my $xbone_alias=-1;		# don't count self
  my $non_xbone_alias=0;
  my $total_alias=0;
  my @cmd;
  my @uifs;
  eval {
    unless(if_type($if) =~ /^tunl/) { 
      XB_Log::log "err", "incompatible tag $if" and die "tag";
    }
    if(! $XB_Defs::TUN_ALIASING){
      # good old days without alias business
      tunl_undo_ifconfig $if;
    }else{
      # need to figure out which route belongs to this tunnel
      my $pipe = "ifconfig $if |";
      open PIPE, $pipe
        or XB_Log::log "err", "can not open pipe $pipe" and die "open";
      # look for address of remote virtual interface in output
      while(<PIPE>) { if(/\s+P-t-P:((\d+\.){3}\d+)/gx) { $rv = $1; last; } }
      close PIPE or XB_Log::log "err", "cannot close pipe $pipe" 
	and die "close";
      unless(defined $rv) { die "rv"; }
      # count the number of xbone / non-xbone aliases
      @uifs = grep /^$pif(:\d+)?$/, if_list 1;
      my $alias_no = 0;
      # check if any xbone alias
      foreach my $i (@uifs){
        if(defined is_xbone_if $i){
          $xbone_alias++;
        }else{
          $non_xbone_alias++;
        }
      }
      $total_alias = $xbone_alias + $non_xbone_alias;
      my $route_del = 0;
      my $delete_primary = 0;
      if($pif eq $if){
	# this means we are the primary interface, don't bring it down unless
	# there are NO other aliases
	if(! $total_alias){	# no other alias, bring it down
	  @cmd = ("ifconfig", $if, "0.0.0.0", "down");
	}else{			# leave it up since other aliases exist
	  @cmd = ("ifconfig", $if, "0.0.0.0");
	  # need to issue a explicit route_delete here since the host route
	  # we added will NOT be deleted automatically in this case
	  $route_del = 1;
        }
      }else{
	# just an alias, bring it down
	@cmd = ("ifconfig", $if, "down");
	# check if last one, if yes, bring down the primary as well
	if(! $total_alias){ $delete_primary = 1; }
      }
      XB_Route::config_route(@cmd);
      if($delete_primary){
	@cmd = ("ifconfig", $pif, "0.0.0.0", "down");
	XB_Route::config_route(@cmd);
      }
      if($route_del == 1) {
	# don't use "-interface" or "dev" for now
	XB_Route::delete_route $rv;
      }
      # determine if we should delete the whole interface
      # delete if (1) we are the last one; or (2) we are the last XBone alias
      # and we don't share
      if($total_alias==0 || ($xbone_alias=0 && !$XB_Defs::SHARE_TUN_IF)){
	# don't delete it, but change it to any to any for future use, and
	# this might solve the mrtd problem
	# this doesn't seem to work, but no error message as well! :-)
	my $any= "0.0.0.0";
	@cmd = ("ip", "tunnel", "change", $pif, "mode", "ipip", "remote", $any,
		   "local", $any);
	# @cmd = ("ip", "tunnel", "del", $pif);
	system(@cmd) == 0 or XB_Log::log "err", "@cmd failed: $!";
	#----------------------------------------------------------------------
	# no need for the following if we don't actually delete tunl interfaces
	# by ip tun del
	#----------------------------------------------------------------------
	# if mrtd is running, we need to reboot mrtd to get rid of the old
	# interface index, otherwise mrtd will generate error next time we use
	# the same interface name but ip tun or ifconfig uses different index
	#my $mrtd_sock = XB_Route::check_mrtd(1);
	#if( $mrtd_sock != 0 ) {
	#  unless(XB_Route::mrtd_write_reboot($mrtd_sock, 1, 1)){
	#    $mrtd_sock->send("exit\r\n");
	#    XB_Route::mrtd_recv($mrtd_sock);
	#    XB_Log::log "warning", "XB_Route::config_route: ".
	#      "cannot reboot mrtd" and die "mrtd reboot";
	#  }
	#  $mrtd_sock->close;
	#}
      }
      # disable the check since mrtd wouldn't find it if it doesn't know
      # safety check since ifconfig down an alias will NOT delete the route
      #if(not XB_Route::get_route $rv) {
      #  XB_Log::log "warning", "Route to $rv is already gone";
      #} else {
      #remove route to virtual remote interface (used to be in xb_tunnel)
      #  XB_Route::delete_route $rv, "dev", $pif;
      #}
    }
  };
  # print trace line
  XB_Log::log "info", "<- XB_Tunnel::tunl_down $if";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(tag|XB_Route::config_route|XB_Route::delete_route)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::tunl_down: caught unexpected " .
      "exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_Tunnel::tunl_down";
}


###############################################################################
# TUN/DIVERT TUNNELS (XB_TUNNEL)
###############################################################################

# atomicity helper killing xb_tunnel
sub tun_undo_fork ($) {
  my $pid = shift;
  XB_Log::log "info", "-> XB_Tunnel::tun_undo_fork $pid";
  eval {
    # kill xb_tunnel daemon
    XB_Log::log "debug3", "Sending INT to daemon child $pid.\n";
    kill INT => $pid or
      XB_Log::log "err", "cannot send INT to child $pid: $!" and die "kill";
    # reap zombie
    XB_Log::log "debug3", "Sent INT to $pid, waiting...\n"; 
    waitpid $pid, 0 == $pid 
      or XB_Log::log "err", "no child $pid" and die "waitpid";
    XB_Log::log "debug3", "Child $pid came home.\n";
  };
  # print trace line
  XB_Log::log "info", "<- XB_Tunnel::tun_undo_fork $pid";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(kill|waitpid)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::tun_undo_fork: caught unexpected " .
      "exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_Tunnel::tun_undo_fork";
}

  
# atomicity helper uninstalling route to remote virtual interface
sub tun_undo_route ($) {
  my $rv = shift;
  XB_Log::log "info", "-> XB_Tunnel::tun_undo_route $rv";
  eval {
    # Check if we route is still there before deleting it. While this is not 
    # necessary for this code to work, it helps troubleshooting a lot.
    if(not XB_Route::get_route $rv) { 
      XB_Log::log "warning", "Route to $rv is already gone"; 
    } else {
      # remove route to virtual remote interface (used to be in xb_tunnel)
      XB_Route::delete_route $rv;
    }
  };
  # print trace line
  XB_Log::log "info", "<- XB_Tunnel::tun_undo_route $rv";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^XB_Route::(get|delete)_route/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::tun_undo_route: caught unexpected " .
      "exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_Tunnel::tun_undo_route";
}
   

# tear down a tun-type tunnel
sub tun_down ($) {
  # extract tunnel information from tag
  my $if = shift;
  XB_Log::log "info", "-> XB_Tunnel::tun_down $if";
  my $rv;
  # first, figure out which route belongs to this tunnel
  eval {
    if($if =~ /(tun\d+)/) { $if = $1; } else { die "tag"; }
    # route delete (figure out which to delete by looking at ifconfig output)
    my $pipe = "ifconfig $if |";
    open PIPE, $pipe 
      or XB_Log::log "err", "Could not open pipe $pipe" and die "open";
    # look for address of remote virtual interface in output
    while(<PIPE>) { if(/\s-->\s((\d{1,3}\.){3}\d{1,3})/) { $rv = $1; last; } }
    close PIPE or XB_Log::log "err", "Cannot close pipe $pipe" and die "close";
    unless(defined $rv) { die "rv"; }
  };
  my $save_at = $@;
  # second, delete the route we found
  eval { if(defined $rv) { tun_undo_route $rv; } };
  unless($@) { $@ = $save_at; }
  # third, find out which kid process handles this tunnel
  my $pid;
  eval {
    if(defined $rv) {
      my $pipe = "ps -ax |";
      open PIPE, $pipe 
	or XB_Log::log "err", "Could not open pipe $pipe" and die "open";
      # look for address of remote virtual interface in output
      while(<PIPE>) { 
	if(/(\d+).*$XB_Defs::XB_TUNNEL_CMD.*$rv/) { $pid = $1; last; } 
      }
      close PIPE 
	or XB_Log::log "err", "Cannot close pipe $pipe" and die "close";
    }
    unless(defined $pid) { die "pid"; }
  };
  unless($@) { $@ = $save_at; }
  # fourth, kill that kid
  eval { if(defined $pid) { tun_undo_fork $pid; } };
  unless($@) { $@ = $save_at; }
  # print trace line
  XB_Log::log "info", "<- XB_Tunnel::tun_down $if";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(rv|tag|open|close|pid|XB_Tunnel::tun_undo_(route|fork))/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::tun_down: caught unexpected " .
      "exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_Tunnel::tun_down";
}


sub tun_up ($$$$$$;$) {
  my ($if, $lv, $rv, $nm, $l, $r, $alias) = @_;
  XB_Log::log "info", "-> XB_Tunnel::tun_up $if, $lv, $rv, $nm, $l, $r";
  my $pid;
  eval {
    # First, see if we already have a route to the default virtual
    # address.  This is a strong indication that another tunnel
    # instance is already up, so bail out before we do
    # damage. (Starting a second tunnel process and then killing it
    # will remove the route and loopback alias from the first!)
    if(XB_Route::get_route $rv) {
      XB_Log::log "warning", "Have route to $rv - another tunnel already up?";
    }
    my $pipe = "$XB_Defs::XB_TUNNEL_CMD -r -g -P -d -l $lv $rv $nm $l $r |";
    open PIPE, $pipe or XB_Log::log "err", "cannot open pipe $pipe: $!" 
      and die "open";
    while(defined(my $line = <PIPE>)) { if($line =~ /^(\d+)/) { $pid = $1; } }
    # done with pipe
    close PIPE or XB_Log::log "err", "cannot close pipe $pipe: $!" 
      and die "close";
    if(defined $pid) { XB_Log::log "debug3", "tunnel running with pid $pid"; }
    else { XB_Log::log "debug3", "cannot get tunnel child pid" and die "pid"; }
    # add route to virtual remote interface (used to be in xb_tunnel)
    # if anything fails, undo the fork before returning
    XB_Route::add_route $rv, "-interface", $if;
  };
  # print trace line
  XB_Log::log "info", "<- XB_Tunnel::tun_up $if, $lv, $rv, $nm, $l, $r";
  return $if unless $@; # success if no exception
  # unknown exception caught, log and pass up a defined one
  if($@ =~ /^(open|close|XB_Route::(get|add)_route)/) {
    if($@ =~ /^(close|XB_Route::add_route)/) {
      eval { 
#	if($pid) { tun_undo_fork $pid; }
#	else{ 
#	  XB_Log::log "warning", "need to kill ip-tun daemon, but have no pid";
#	}
      };
    }
  } else {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::tun_up: caught unexpected " .
      "exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_Tunnel::tun_up";
}

###############################################################################
# GENERIC EXPORTED API
###############################################################################


# dispatch hashes for each of those interfaces
my %up_fn = ( gif => \&iti_or_gif_up, iti => \&iti_or_gif_up, 
	      tunl => \&tunl_up, tun => \&tun_up );
my %down_fn = (gif => \&iti_or_gif_down, iti => \&iti_or_gif_down, 
	       tunl => \&tunl_down, tun => \&tun_down );

# Description: 
#	Configure tunnel interface.
# Arguments:
#	$lv	local virtual address (inner header source)
#	$rv	remote virtual address (inner header destination)
#	$nm	netmask for tunnel (on inner addresses)
#	$l	local address (outer header source)
#	$r	remote address (outer header destination)		    
# Returns:
#	tunnel handle
# Exceptions:
#	"XB_Tunnel::up" on error, nothing to clean up by caller
sub up ($$$$$) {
  my ($lv, $rv, $nm, $l, $r) = @_;
  # print trace line
  XB_Log::log "info", "-> XB_Tunnel::up $lv, $rv, $nm, $l, $r";
  my $tag = eval {
    foreach my $addr ($lv, $rv, $l, $r) {
      if($addr !~ /^(\d{1,3}\.){3}\d{1,3}/) {
	XB_Log::log "err", "illegal IP address $addr" and die "addr";
      }
    }
    my $if = undef;
    my $alias = 0;
    # TUN ALIASING: 
    #   - check gif/tunl interfaces for the same outer end points
    #   - an interface found here would precede any other types of tunnels
    #   * die if a non-xbone alias exists on the interface found and sharing
    #     is disabled ($XB_Defs::SHARE_TUN_IF=0)
    # check if a used (gif/tunl) interface w/ the same outer end points exists
    if( $XB_Defs::TUN_ALIASING == 1) {
      if($XB_Defs::NODEOS =~ /freebsd|kame/i){
	$if = gif_find_outer($l, $r);
      }elsif($XB_Defs::NODEOS=~ /linux|nist/i){
	$if = tunl_find_outer($l, $r);
      } 
    } # $if = undef if TUN_ALIASING=0 or no alias found
    if(defined $if) {
      # found one, log and set alias flag
      XB_Log::log "info", "   Found $if with the same outer end point: $l->$r.".
		        "\n   Add $lv->$rv as an alias to $if."; 
      $alias = 1;
    } else {
      # no gif/tunl interface between $l->$r found, or not using aliasing;
      # try to find an unused interface of a kind we know how to tunnel with
      foreach my $type (@XB_Defs::TUNNELING_IFS) { 
        my @ifs = grab $type, 0;
        if($if = shift @ifs) { last; } 
      }
      if(not $if) { 
        XB_Log::log "err", "no unused tunneling interfaces" and die "none";
      }
    }
    # do tunnel config, depending on the interface type we could grab
    my $type = if_type $if;
    if(my $fn = $up_fn{$type}) { 
      return ($alias ? &$fn($if, $lv, $rv, $nm, $l, $r, $alias) :
		       &$fn($if, $lv, $rv, $nm, $l, $r) ); 
    } else { 
      XB_Log::log "err", "have $if, cannot tunnel with it" and die "unknown"; 
    }
  };
  # print trace line
  XB_Log::log "info", "<- XB_Tunnel::up $lv, $rv, $nm, $l, $r";
  return $tag unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(addr|none|unknown|XB_Tunnel::.*_up)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::up: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_Tunnel::up";
}


# Description: 
#	Tear down tunnel $tag.
# Arguments:
#	$if	tunnel tag (return value from corresponding XB_Tunnel::up)
# Returns:
#	1 on success
# Exceptions:
#	"XB_Tunnel::down" on error, nothing to clean up by caller
sub down ($) {
  my $tag = shift;
  # print trace line
  XB_Log::log "info", "-> XB_Tunnel::down $tag";
  eval {
    my $type = if_type $tag;
    if(my $fn = $down_fn{$type}) { &$fn($tag); } 
    else { XB_Log::log "err", "unknown $type tunnel $tag" and die "unknown"; }
  };
  # print trace line
  XB_Log::log "info", "<- XB_Tunnel::down $tag";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^(unknown|XB_Tunnel::(iti_or_gif|tun)_down)/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::down: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_Tunnel::down";
}


# Description: 
#	Initialize tunneling module. Must be called once prior to using
#       any other funtion in this module.
# Arguments:
#       none 
# Returns:
#	1 on success
# Exceptions:
#	"XB_Tunnel::init" on error, nothing to clean up by caller
sub init () {
  XB_Log::log "info", "-> XB_Tunnel::init";
  # check if we are on a CAIRN host and if so, turn on recursive tunneling
  eval {
    # enable recursive tunnels on CAIRN (requires our CAIRN patch)
    if($XB_Defs::NODEOS =~ /cairn/i) {
      my $sysctl_var = "net.inet.ip.iti_recursion";
      my $val = XB_Utils::sysctl_read $sysctl_var;
      if(defined $val and $val eq "0") {
	# ok, this seems to be a CAIRN host that needs it turned on
	XB_Log::log "info", "turning on recursive tunneling";
	XB_Utils::sysctl_write $sysctl_var, 1;
      }
    }
    # if there are tun devices, make sure we can execute ip-tun
    my $have_tuns = 0;
    foreach my $used (qw(0 1)) { $have_tuns = grab "tun", $used, 1 and last; }
    if($have_tuns) { 
      # this machine has tun interfaces, make sure we can find ip-tun
      my $cmd;
      foreach my $dir (qw(.. .)) {
	$cmd = "$FindBin::RealBin/$dir/$XB_Defs::XB_TUNNEL_CMD";
	if(-x $cmd) { 
	  XB_Log::log "info", "found ip-tun binary at $cmd";
	  # overwrite relative path $XB_Defs::XB_TUNNEL_CMD with found one
	  $XB_Defs::XB_TUNNEL_CMD = $cmd;
	  last;
	}
	$cmd = "";
      }
      unless($cmd) {
	# fix later - do this only in certain cases
	# XB_Log::log "err", "cannot find ip-tun binary" and die "ip-tun";
      }
    }
    # if there are gif devices, check if "gifconfig delete" is supported
    my $have_gifs = 0;
    foreach my $used (qw(0 1)) { $have_gifs = grab "gif", $used, 1 and last; }
    if($have_gifs) { 
      my @test_gifs = grab "gif", 0;
      if($#test_gifs != -1) {
	# got free gif interface, test gifconfig delete
	my @cmd = ("gifconfig", shift @test_gifs, "delete");

	my $pipe = "gifconfig " .(shift @test_gifs). " delete 2>&1 |";
	local $SIG{PIPE} = sub { 
	  XB_Log::log "err", "Pipe $pipe broke"; return 
	};
	open PIPE, $pipe 
	  or XB_Log::log "err", "cannot open pipe $pipe: $!" and die "open";
	# if we don't read the output, we cannot close the pipe
	my @output = <PIPE>;
	
	close PIPE or not $!
	  or XB_Log::log "err", "cannot close pipe $pipe: $!" and die "close";
	$have_gifconfig_delete = ($? ? 0 : 1);
	XB_Log::log "notice", "gifconfig(8) on this KAME host ".
	  ($have_gifconfig_delete ? "seems" : "does not yet seem") .
	    " to support deleting gifs (functionality added ~ 7/2000)";
      } else {
	XB_Log::log "err", "XB_Tunnel::init: no free gif interfaces"
	  and die "XB_Tunnel::init";
      }
      
    }
  };
  # print trace line
  XB_Log::log "info", "<- XB_Tunnel::init";
  return 1 unless $@; # success if no exception
  # exception handling
  unless($@ =~ /^ip-tun/) {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", "XB_Tunnel::init: caught unexpected exception $@";
  }
  # pass defined exceptions up to caller
  die "XB_Tunnel::init";
}


# Description: 
#	Reset all xbone tunnels.
# Arguments:
#	none
# Returns:
#	1 on success
# Exceptions:
#	"XB_Tunnel::reset" on error, nothing to clean up by caller
sub reset () {
  XB_Log::log "info", "-> XB_Tunnel::reset";
  # for all tunneling interface types
  my $fail;
  foreach my $type (@XB_Defs::TUNNELING_IFS) { 
    # for all configured interfaces of the current type
    my @ifs;
    eval { @ifs = grab $type, 1 };
    if($@) { 
      $fail = $@;
      XB_Log::log "warning", "cannot grab used $type tunnels: $@, continuing"; 
    } else {
      foreach my $if (@ifs) {
	eval { down $if };
	if($@) { 
	  $fail = $@;
	  XB_Log::log "warning", "cannot init $if: $@, continuing"; 
	}
      }
    }
  }
  # print trace line
  XB_Log::log "info", "<- XB_Tunnel::reset";
  return 1 unless $fail; # success if no exception
  # exception handling
  unless($@ eq "") {
    # unknown exception caught, log and pass up a defined one
    XB_Log::log "warning", 
      "XB_Tunnel::reset: caught unexpected exception $fail";
  }
  # pass defined exceptions up to caller
  die "XB_Tunnel::reset";
}


1;

